
#ifndef _LOC_H
#define _LOC_H

#include "Global.H"


namespace blazi {


/** Location representation.
 *
 *  There is a lexicographical ordering among them,
 *  where two undefined locations are considered equal.
 */
struct Loc
{
    static const sint4 UNDEF = -1000000;

    Loc( sint4 x_ = UNDEF, sint4 y_ = UNDEF ) : x(x_), y(y_) {}
    Loc( const Loc &loc_ ) { *this = loc_; }

    bool valid() const { return UNDEF/2 <= x && UNDEF/2 <= y; }

    Loc& operator+=( const Loc &loc_ ) { x += loc_.x; y += loc_.y; return *this; }
    Loc& operator-=( const Loc &loc_ ) { x -= loc_.x; y -= loc_.y; return *this; }
    Loc& operator*=( const Loc &loc_ ) { x *= loc_.x; y *= loc_.y; return *this; }
    Loc& operator/=( const Loc &loc_ ) { x /= loc_.x; y /= loc_.y; return *this; }

    Loc& operator=( const Loc &loc_ ) { x = loc_.x; y = loc_.y; return *this; }

    static Loc avg( const Loc &loc_a_, const Loc &loc_b_ );

    /* vector operators (LinAlg.H should be used instead) */
    sint4 length() const { return distance( *this, Loc(0,0) ); }
    void normalize() { sint4 len = length(); x /= len; y /= len; }
    void multiply( sint4 m_ ) { x *= m_; y *= m_; }

    void rotate_l() { sint4 t = x; x = -y; y = t; }
    void rotate_r() { sint4 t = x; x = y; y = -t; }

    static sint4 distance( const Loc &loc_a_, const Loc &loc_b_ );
    static sint4 hamilton_distance( const Loc &loc_a_, const Loc &loc_b_ );
    static Loc away( const Loc &away_, const Loc &now_, sint4 with_ = 1 );

    sint4 x, y;
};


inline bool operator==( const Loc &loc_a_, const Loc &loc_b_ )
{
    return loc_a_.x == loc_b_.x && loc_a_.y == loc_b_.y;
}

inline bool operator!=( const Loc &loc_a_, const Loc &loc_b_ )
{
    return !(loc_a_ == loc_b_);
}

inline bool operator<( const Loc &loc_a_, const Loc &loc_b_ )
{
    return loc_a_.x < loc_b_.x ||
           ( loc_a_.x == loc_b_.x && loc_a_.y < loc_b_.y );
}

inline bool operator>( const Loc &loc_a_, const Loc &loc_b_ )
{
    return loc_b_ < loc_a_;
}

inline bool operator<=( const Loc &loc_a_, const Loc &loc_b_ )
{
    return !(loc_a_ > loc_b_);
}

inline bool operator>=( const Loc &loc_a_, const Loc &loc_b_ )
{
    return !(loc_a_ < loc_b_);
}

inline std::ostream& operator<<( std::ostream &os, const Loc &loc_ )
{
    return os << "(" << loc_.x << "," << loc_.y << ")";
}


} /* namespace blazi */


#endif /* _LOC_H */

