
#include <set>

#include "GameObj.H"

#include "_Global.H"
#include "_PathFind.H"


/* implementation specific addons */
namespace blazi {


/** Additional position information for path finding. */
class PFPosition
{
public:

    typedef PathFind::Distance Distance;

public:

    PFPosition( const Loc &loc_,
                Distance g_ = 0,
                Distance h_ = 0,
                PFPosition *parent_ = NULL )
    {
        loc = loc_;
        set_f( g_, h_ );
        parent = parent_;
    }

    Distance get_g() const { return _g; }
    Distance get_h() const { return _h; }
    Distance get_f() const { return _f; }

    void set_g( Distance g_ ) { assert( 0 <= g_ ); _g = g_; update_f(); }
    void set_h( Distance h_ ) { assert( 0 <= h_ ); _h = h_; update_f(); }

    void set_f( Distance g_, Distance h_ )
    {
        assert( 0 <= g_ );
        assert( 0 <= h_ );
        _g = g_;
        _h = h_;
        update_f();
    }

    Loc loc;
    PFPosition *parent;

private:

    void update_f() { _f = _g + _h; }

    Distance _g, _h, _f;
};


/** A common set representation for the open and closed sets. */
class PFSet
{
public:

    ~PFSet()
    {
        while( !_hset.empty() )
        {
            PFHashSetIt it = _hset.begin();
            PFPosition *pos = *it;
            _hset.erase( it );
            delete pos;
        }
    }

    bool empty() const { return _hset.empty(); }
    size_t size() const { return _hset.size(); }

    PFPosition* find( const Loc &loc_ ) // O( hash )
    {
        PFPosition pos( loc_ );
        PFHashSetIt it = _hset.find( &pos );
        if( _hset.end() == it ) return NULL;
        return *it;
    }

    PFPosition* pop_min_f() // O( n ) + O( hash )
    {
        PFPosition *pos = NULL;

        FORALL( _hset, it )
        {
            if( !pos || (*it)->get_f() < pos->get_f() )
            {
                pos = *it;
            }
        }

        if( pos ) erase( pos );
        return pos;
    }

    void insert( PFPosition *pos_ ) // O( hash )
    {
        assert( pos_ );
        assert( pos_->loc.valid() );
        _hset.insert( pos_ );
    }

    void erase( PFPosition *pos_ ) // O( hash )
    {
        assert( pos_ );
        _hset.erase( pos_ );
    }

private:

    struct Hash
    {
        size_t operator()( PFPosition *pos_ ) const
        { return pos_->loc.y << (4*sizeof(pos_->loc.x)) + pos_->loc.x; }
    };

    struct Eq
    {
        bool operator()( PFPosition *pos_a_, PFPosition *pos_b_ ) const
        { return pos_a_->loc == pos_b_->loc; }
    };

    typedef Sgi::hash_set<PFPosition*, Hash, Eq> PFHashSet;
    typedef PFHashSet::iterator PFHashSetIt;

    PFHashSet _hset;
};


/** The open set representation for the A* algorithm. */
class PFOpenSet : public PFSet
{
public:

    ~PFOpenSet()
    {
        _oset.clear();
        // entries are deleted by the super destructor
    }

    PFPosition* pop() // O( hash )
    {
        assert( 0 < _oset.size() );

        OrderedSetIt it = _oset.begin();
        PFPosition *pos = *it;

        PFSet::erase( pos );
        _oset.erase( it );

        assert( size() == _oset.size() );
        return pos;
    }

    void insert( PFPosition *pos_ ) // O( hash ) + O( log(n) )
    {
        assert( pos_ );
        PFSet::insert( pos_ );
        _oset.insert( pos_ );
        assert( size() == _oset.size() );
    }

    void erase( PFPosition *pos_ ) // O( log(n) )
    {
        assert( pos_ );
        PFSet::erase( pos_ );
        _oset.erase( pos_ );
        assert( size() == _oset.size() );
    }

private:

    struct Cmp
    {
        bool operator()( const PFPosition *pos_a_,
                         const PFPosition *pos_b_ ) const
        {
            return pos_a_->get_f() < pos_b_->get_f() ||
                   ( pos_a_->get_f() == pos_b_->get_f() &&
                     pos_a_->loc < pos_b_->loc );
        }
    };

    typedef std::set<PFPosition*, Cmp> OrderedSet;
    typedef OrderedSet::iterator OrderedSetIt;

    OrderedSet _oset;
};


/** The closed set representation for the A* algorithm. */
typedef PFSet PFClosedSet;


/** Creating the path described by the parent links
 *  starting from the given position and returns
 *  it as a result. It also deletes the given position object. */
PathFind::Result create_result( const Terrain &terrain_,
                                sint4 counter_,
                                bool success_,
                                PFPosition *pos_ )
{
    assert( pos_ );

    PFPosition *pos = pos_;

    DEBUG( "create result(" << success_
                            << "," << counter_
                            << ") from " << pos->loc );
    ++counter_; // just not to get warning (eliminate this)

    Path *path = new Path();

    sint4 len = 0;
    while( pos->parent )
    {
        ++len;

        assert( terrain_.valid(pos->loc) );
        DEBUG( "   + " << pos->loc );
        path->push_front( terrain_.pos2real(pos->loc) );

        pos = pos->parent;
    }

    DEBUG( "path len: " << len );

    delete pos_;
    return PathFind::Result( success_, path );
}


} /* namespace blazi */


blazi::PathFind::PathFind( const Terrain &terrain_ )
    : _terrain(terrain_)
{
}


blazi::PathFind::Result
blazi::PathFind::search_path( const Obj *obj_,
                              const Loc &from_,
                              const Loc &to_,
                              Distance freeCheckRadius_ ) const
{
    DEBUG( "search: " << from_ << " -> " << to_ );

    assert( obj_ );
    assert( terrain().valid(from_) );
    assert( terrain().valid(to_) );

    LocSeq fromSeq;
    fromSeq.insert( from_ );

    LocSeq toSeq;
    toSeq.insert( to_ );

    return search_engine( obj_, fromSeq, toSeq, freeCheckRadius_ );
}


blazi::PathFind::Result
blazi::PathFind::search_path( const Obj *from_,
                              const Obj *to_,
                              Distance freeCheckRadius_,
                              const Obj *checkedObj_ ) const
{
    assert( from_ );
    assert( to_ );

//    Terrain::ObjLocationsPtr fromP = terrain().obj_locs( from_ );
    LocSeq fromP;
    fromP.insert( terrain().posloc( from_ ) );
    Terrain::ObjLocationsPtr toP = terrain().obj_locs( to_ );

//    assert( !fromP->empty() );
    assert( !toP->empty() );

    DEBUG( "fromP: " << *(fromP.begin()) );
    DEBUG( "toP: " << *(toP->begin()) );

    const Obj *who = checkedObj_ ? checkedObj_ : from_;
    return search_engine( who, fromP, *toP, freeCheckRadius_ );
}


blazi::PathFind::Result
blazi::PathFind::search_engine( const Obj *obj_,
                                const LocSeq &from_,
                                const LocSeq &to_,
                                Distance freeCheckRadius_ ) const
{
    assert( obj_ );
    assert( !from_.empty() );
    assert( !to_.empty() );

//    DEBUG( "pathfind, freeCheckRadius = " << freeCheckRadius_ );

    PFOpenSet open;
    PFClosedSet closed;

    // create the starting position and push it into the open set
    FORALL( from_, it )
    {
        assert( terrain().valid(*it) );
        open.insert( new PFPosition( *it, 0, terrain().distance(*it, to_) ) );
    }

    // the search
    sint4 c = 0;
    Loc new_loc;
    PFPosition *pos, *new_pos;
    Distance new_g;

    while( !open.empty() )
    {
        ++c;

        // get a position with minimal f value
        pos = open.pop();
        DEBUG( "  POSloc = " << pos->loc );

        // check whether it is a goal, if so the optimal path is found
        if( to_.end() != to_.find(pos->loc) )
        {
            DEBUG( "  optimal path has been found" );
            return create_result( terrain(), c, true, pos );
        }

        // check the neighbouring positions
        FORALLCOMPASS( compass )
        {
            // looking for an expandable boundary
            new_loc = terrain().compass2loc( pos->loc, compass );
            if( !new_loc.valid() ) continue; // out of map is ignored

 //           DEBUG( "  " << pos->loc << " -> " << new_loc );

            if( !terrain().check_pos_pass( obj_,
                                           pos->loc,
                                           new_loc,
                                           terrain().distance( new_loc, from_ ) <
                                                                    freeCheckRadius_ ) )
            {
                // if the ignored one is a goal, we consider it a success
                if( to_.end() != to_.find( new_loc ) )
                {
                    DEBUG( "  goal is unwalkable, but reachable" );
                    return create_result( terrain(), c, true, pos );
                }

  //              DEBUG( "  ignored: " << new_loc );
                continue;
            }

            // skip if exist in the closed list
            if( NULL != closed.find( new_loc ) ) continue;

            // computing the new distance from the starting position
            new_g = pos->get_g() + terrain().distance( pos->loc, new_loc );

            if( (new_pos = open.find( new_loc )) )
            {
                if( new_pos->get_g() <= new_g ) continue;
                open.erase( new_pos );
            }
            else
            {
                // position has never been explored
                new_pos = new PFPosition( new_loc );
            }

            new_pos->parent = pos;
            new_pos->set_f( new_g, terrain().distance( new_pos->loc, to_ ) );
            open.insert( new_pos );
        }

        // put the actual position into the closed set
        closed.insert( pos );
    }

    DEBUG( "  goal is unreachable" );

    // no path has been found (failure), return with the closest path
    return create_result( terrain(), c, false, closed.pop_min_f() );
}

