
#include <cmath>
#include <algorithm>

#include "Game.H"
#include "GameTile.H"
#include "LinAlg.H"

#include "_Terrain.H"


blazi::Position::Position()
{
    _tile = NULL;
    _has_nomob_obj = false;
}


void blazi::Position::init( const GameTile *tile_, sint4 x_, sint4 y_ )
{
    assert( tile_ );

    _tile = tile_;
    _loc = Loc( x_, y_ );
}


void blazi::Position::add_obj( const Obj *obj_ )
{
    assert( obj_ );

    _objs.insert( obj_ );
    _has_nomob_obj |= !obj_->mobile();
}


void blazi::Position::rem_obj( const Obj *obj_ )
{
    assert( obj_ );

    _objs.erase( obj_ );

    if( !obj_->mobile() )
    {
        _has_nomob_obj = false;
        FORALL( objs(), it )
        {
            if( (*it)->mobile() )
            {
                _has_nomob_obj = true;
                break;
            }
        }
    }
}


void blazi::Path::push_front( const Loc &loc_ )
{
    if( !empty() && loc_ == front() ) return;

    if( !empty() ) _len += Loc::distance( loc_, front() );
    _path.push_front( loc_ );
}


void blazi::Path::push_back( const Loc &loc_ )
{
    if( !empty() && loc_ == back() ) return;

    if( !empty() ) _len += Loc::distance( loc_, back() );
    _path.push_back( loc_ );
}


void blazi::Path::pop_front()
{
    assert( !empty() );
    
    Loc rem;
    if( 3 <= size() ) rem = front();

    _path.pop_front();

    if( 3 <= size() )
        _len -= Loc::distance( rem, front() );
    else
        _len = 0;
}


void blazi::Path::pop_back()
{
    assert( !empty() );
    
    Loc rem;
    if( 3 <= size() ) rem = back();

    _path.pop_back();

    if( 3 <= size() )
        _len -= Loc::distance( back(), rem );
    else
        _len = 0;
}


blazi::Path& blazi::Path::concat( const Path *path_ )
{
    assert( path_ );

    if( !path_->empty() )
    {
        LocSeq::const_iterator cit = path_->path().begin();
        if( !empty() && *cit == back() ) ++cit;

        while( path_->path().end() != cit )
        {
            push_back( *cit );
            ++cit;
        }
    }

    return *this;
}


blazi::Terrain::Terrain( const Map<GameTile> &map_, sint4 tile_points_ )
    : _map( map_ ),
      _tile_points( tile_points_ ),
      _tile_points_per_4( (sint4) tile_points_ / 4 ),
      _straight_step( (Distance)( get_tile_points() ) ),
      _diagonal_step( (Distance)( M_SQRT2 * get_tile_points() ) )

{
    _terrain = new Position[ size() ];

    DEBUG( "straight step = " << _straight_step );
    DEBUG( "diagonal step = " << _diagonal_step );

    FORU( i, size() )
    {
        _terrain[i].init( &(get_map()(i)),
                          get_map().ind2x(i),
                          get_map().ind2y(i) );
    }
}


blazi::Terrain::~Terrain()
{
    while( !_objTerr.empty() )
    {
        ObjTerritoriesIt it = _objTerr.begin();
        ObjTerritory *terr = it->second;
        _objTerr.erase( it );
        delete terr;
    }

    delete []_terrain;
}


blazi::Loc blazi::Terrain::pos2real( const Loc &loc_ ) const
{
    static sint4 tile_points_per_2 = get_tile_points() / 2;
    return Loc( loc_.x * get_tile_points() + tile_points_per_2,
                loc_.y * get_tile_points() + tile_points_per_2 );
}


void blazi::Terrain::update_pos( const Obj *obj_, ObjTerritory *terr_, const Loc &loc_ )
{
    assert( obj_ );
    assert( terr_ );

//    DEBUG( "    update_pos: " << loc_ );
    if( !valid( loc_ ) ) return;

    Position *pos = get_pos( loc_ );
    pos->add_obj( obj_ );
    terr_->push_back( pos );
}


void blazi::Terrain::update_obj( Obj *obj_ )
{
    assert( obj_ );

    const GameObj *gob = obj_->obj();
    const sint4 &x = *gob->sod.x;
    const sint4 &y = *gob->sod.y;

    ObjTerritory *terr = NULL;
    ObjTerritoriesIt it = _objTerr.find( obj_ );

    if( _objTerr.end() == it )
    {
        // the object has not been refreshed
        terr = new ObjTerritory;
        _objTerr.insert( std::make_pair( obj_, terr ) );
    }
    else
    {
        // skip if the object has not moved
        if( x == gob->sod.prev.x && y == gob->sod.prev.y ) return;

        terr = it->second;

        // cleaning the old territory
        while( !terr->empty() )
        {
            ObjTerritoryIt it2 = terr->begin();

            (*it2)->rem_obj(obj_);
            terr->erase( it2 );
        }
    }

    // creating the new territory
    switch( gob->get_shape() )
    {
        case Object::CIRCLE :
        {
            Loc l = loc( obj_ );
            sint4 r = gob->get_radius();

            update_realpos( obj_, terr, Loc( l.x, l.y ) );
            update_realpos( obj_, terr, Loc( l.x, l.y - r ) );          // N
            update_realpos( obj_, terr, Loc( l.x + r/2, l.y - r/2 ) );  // NE
            update_realpos( obj_, terr, Loc( l.x + r, l.y ) );          // E
            update_realpos( obj_, terr, Loc( l.x + r/2, l.y + r/2 ) );  // SE
            update_realpos( obj_, terr, Loc( l.x, l.y + r ) );          // S
            update_realpos( obj_, terr, Loc( l.x - r/2, l.y + r/2 ) );  // SW
            update_realpos( obj_, terr, Loc( l.x - r, l.y ) );          // W
            update_realpos( obj_, terr, Loc( l.x - r/2, l.y - r/2 ) );  // NW

            break;
        }

        case Object::RECTANGLE :
        {
            Loc l1, l2, l;
            gob->get_p1( l1.x, l1.y, NULL, NULL );
            gob->get_p2( l2.x, l2.y, NULL, NULL );
            l1 = real2pos( l1 );
            l2 = real2pos( l2 );

            l = l1;
            do
            {
                update_pos( obj_, terr, l );

                if( ++l.x > l2.x ) { l.x = l1.x; ++l.y; }
            }
            while( l.y <= l2.y );

            break;
        }

        default :
        {
            ERR( "not supported shape: " + gob->get_shape() );
        }
    }
}


blazi::Loc
blazi::Terrain::compass2loc( const Loc &loc_, Tile::Compass compass_ ) const
{
    assert( valid( loc_ ) );

    const sint4 &x = loc_.x;
    const sint4 &y = loc_.y;

    switch( compass_ )
    {
        case Tile::N  : if(                  0 < y          ) return Loc( x  , y-1 );
        case Tile::NE : if( x < width()-1 && 0 < y          ) return Loc( x+1, y-1 );
        case Tile::E  : if( x < width()-1                   ) return Loc( x+1, y   );
        case Tile::SE : if( x < width()-1 && y < height()-1 ) return Loc( x+1, y+1 );
        case Tile::S  : if(                  y < height()-1 ) return Loc( x  , y+1 );
        case Tile::SW : if( 0 < x         && y < height()-1 ) return Loc( x-1, y+1 );
        case Tile::W  : if( 0 < x                           ) return Loc( x-1, y   );
        case Tile::NW : if( 0 < x         && 0 < y          ) return Loc( x-1, y-1 );
    }

    return Loc(); // returns undefined location
}


bool blazi::Terrain::check_pos_pass( const Obj *obj_,
                                     const Loc &from_,
                                     const Loc &to_,
                                     bool checkFree_ ) const
{
    assert( obj_ );
    assert( valid( from_ ) );
    assert( valid( to_ ) );

    const Position *to = get_pos( to_ );
    if( !walkable( to ) ) return false;

    // check whether the pass is diagonal or not
    sint4 diff_x = to_.x - from_.x;
    sint4 diff_y = to_.y - from_.y;

    if( 0 != diff_x && 0 != diff_y )
    {
        // diagonal pass

        const Position *e, *w;
        Tile::Split s;

        // check according to the compass type
        if( 0 == diff_x + diff_y )
        {
            // default is NE setting
            s = Tile::BT_SPLIT; // |/|
            e = get_pos( from_.x, from_.y + diff_y );
            w = get_pos( from_.x + diff_x, from_.y );

            // switch to SW if necessary
            if( from_.y < to_.y ) std::swap( e, w );
        }
        else
        {
            // default is NW setting
            s = Tile::TB_SPLIT; // |\|
            e = get_pos( from_.x + diff_x, from_.y );
            w = get_pos( from_.x, from_.y + diff_y );

            // switch to SE if necessary
            if( from_.y < to_.y ) std::swap( e, w );
        }

        // if the object was too wide, a different check would be necessary

        if( !walkableE(e) ||
            ( e->get_tile()->get_split() != s && !walkableW(e) ) ) return false;

        if( !walkableW(w) ||
            ( w->get_tile()->get_split() != s && !walkableE(w) ) ) return false;
    }

    Loc from;
    if( from_ == real2pos(obj_->loc()) )
    {
        from = obj_->loc();
    }
    else
    {
        from = pos2real(from_);
    }

    return check_free_pass( obj_, from, pos2real(to_), checkFree_ );
}


bool blazi::Terrain::check_free_pass( const Obj *obj_,
                                      const Loc &from_,
                                      const Loc &to_,
                                      bool consMob_ ) const
{
    DEBUG( "  check_free_pass: " << from_ << " " << real2pos(from_) << " -> "
                                 << to_ << " " << real2pos(to_) );

    assert( obj_ );
    assert( realvalid( from_ ) );
    assert( realvalid( to_ ) );

    if( from_ == to_ ) return true;

    Vec2<real8> offvec( to_.x - from_.x, to_.y - from_.y );

    sint4 step = obj_->obj()->get_radius();

    sint4 c = 1;
    sint4 max_c = (sint4) offvec.length() / step;

    offvec.normalize();
    offvec *= step;
    Vec2<real8> sumvec = offvec;    // = c * offvec

    Loc prop;
    do
    {
        prop.x = from_.x + (sint4) sumvec.x;
        prop.y = from_.y + (sint4) sumvec.y;
//        DEBUG( "prop: " << prop );

        if( !check_free( obj_, prop, consMob_ ) )
        {
            DEBUG( "    !free, prop: " << prop );
            return false;
        }

        sumvec += offvec;
    }
    while( ++c <= max_c );

    return true;
}


blazi::Loc blazi::Terrain::dodge_point( const Worker *obj_,
                                        Loc to_,
                                        Loc to2_ ) const
{
    assert( obj_ );
    assert( to_ != to2_ );

    Loc objloc = obj_->loc();

    if( objloc == to_ ) to_ = to2_;
    DEBUG( "dodge: " << objloc << " <> " << to_ );

    Vec2<real8> step( to_.x - objloc.x, to_.y - objloc.y );
    step.normalize();
    DEBUG( "dodge, step = " << step.x << "," << step.y );

    Loc hit;
    {
        Vec2<real8> temp = step * obj_->obj()->get_radius();
        hit = Loc( (sint4)temp.x, (sint4)temp.y );
    }
    hit += objloc;
    DEBUG( "dodge, hit = " << hit );

    PosSet ps;
    fill_pos_set( ps, obj_->obj()->get_radius(), hit );

    const Obj *other = NULL;
    real8 min_d = 0, d;

    FORALL( ps, it )
    {
        FORALL( (*it)->objs(), oit )
        {
            if( Object::CIRCLE != (*oit)->obj()->get_shape() ||
                *oit == obj_ ) continue;

            d = Game::distance( *obj_->obj(), *(*oit)->obj() );

            if( !other || d < min_d )
            {
                other = *oit;
                min_d = d;
            }
        }
    }

    if( !other )
    {
        DEBUG( "dodge but WHY?" );
        return hit;
    }

    Loc result;
    bool left = false;

    const Worker *worker = dynamic_cast<const Worker*>( other );
    if( !worker || !obj_->dodge_wait || ( !worker->target() && obj_->target() ) || 
        ( worker->target() && obj_->target() &&
          Game::distance( *worker->obj(), *worker->target()->obj() ) <
          Game::distance( *obj_->obj(), *obj_->target()->obj() ) ) )
    {
        Loc dodge_l;
        {
            Loc ol = other->loc();
            step.x = ol.x - objloc.x;
            step.y = ol.y - objloc.y;
        DEBUG( "dodge, step = " << step.x << "," << step.y );
            step.normalize();
        DEBUG( "dodge, step = " << step.x << "," << step.y );
            step *= 1.25 * other->obj()->get_radius();
        DEBUG( "dodge, step = " << step.x << "," << step.y );
            dodge_l = Loc( (sint4)step.x, (sint4)step.y );
        }
        DEBUG( "dodge straight: " << dodge_l );
        Loc dodge_r = dodge_l;

        dodge_l.rotate_l();
        dodge_l += objloc;
        DEBUG( "dodge_l: " << dodge_l );

        if( realvalid( dodge_l ) &&
            !obj_->dodge_right &&
            ( !worker || !worker->dodge_right ) &&
            check_free( obj_, dodge_l, true ) )
        {
            left = true;
            result = dodge_l;
        }

        dodge_r.rotate_r();
        dodge_r += objloc;

        DEBUG( "dodge_r: " << dodge_r );
        if( realvalid( dodge_r ) &&
            !obj_->dodge_left &&
            ( !worker || !worker->dodge_left ) &&
            ( !result.valid() ||
              Loc::distance(to_, dodge_r) + Loc::distance(dodge_r, to2_) <
              Loc::distance(to_, dodge_l) + Loc::distance(dodge_l, to2_) ) &&
            check_free( obj_, dodge_r, true ) )
        {
            left = false;
            result = dodge_r;
        }

        if( !result.valid() )
        {
            DEBUG( "dodge AWAY" );
/*
            if( !obj_->dodge_wait )
            {
                Loc aw = Loc::away( other->loc(),
                                    obj_->loc(),
                                    obj_->obj()->get_radius() );

                if( realvalid( aw ) / *&&
                    check_free( obj_, aw, true )* / )
                {
                    result = aw;
                }

            }
*/
            obj_->dodge_left = false;
            obj_->dodge_right = false;
        }
        else
        {
            if( left )
            {
                obj_->dodge_left = true;
                obj_->dodge_right = false;
            }
            else
            {
                obj_->dodge_left = false;
                obj_->dodge_right = true;
            }
            result = Loc::away( other->loc(), result );
        }
    }

    DEBUG( "dodge result: " << result );
    return result;
}


blazi::Terrain::Distance
blazi::Terrain::distance( const Loc &from_, const Loc &to_ ) const
{
    assert( valid( from_ ) );
    assert( valid( to_ ) );

    Distance h_min = std::abs( from_.x - to_.x );
    Distance h_max = std::abs( from_.y - to_.y );

    if( h_max < h_min ) std::swap( h_min, h_max );

    return _straight_step * ( h_max - h_min ) + _diagonal_step * h_min;
}


blazi::Terrain::Distance
blazi::Terrain::distance( const Loc &from_, const ObjLocations &to_ ) const
{
    assert( valid( from_ ) );
    assert( !to_.empty() );

    ObjLocations::const_iterator cit = to_.begin();
    Distance act, min = distance( from_, *cit );
    while( to_.end() != ++cit )
    {
        act = distance( from_, *cit );
        if( act < min ) min = act;
    }

    return min;
}


size_t blazi::Terrain::obj_locs_size( const Obj *obj_ ) const
{
    assert( obj_ );

    ObjTerritoriesCIt cit = _objTerr.find( obj_ );
    if( _objTerr.end() == cit ) return 0;

    assert( cit->second );
    return cit->second->size();
}


blazi::Terrain::ObjLocationsPtr blazi::Terrain::obj_locs( const Obj *obj_ ) const
{
    assert( obj_ );

    ObjTerritoriesCIt cit = _objTerr.find( obj_ );
    if( _objTerr.end() == cit ) return ObjLocationsPtr();
    assert( cit->second );

    ObjLocationsPtr locs( new ObjLocations );
    for( ObjTerritoryIt it2 = cit->second->begin(); it2 != cit->second->end(); ++it2 )
    {
        locs->insert( (*it2)->loc() );
    }

    return locs;
}


/*
blazi::Loc blazi::Terrain::min_objloc( const Loc &loc_, const Obj *obj_ ) const
{
    assert( valid( loc_ ) );
    assert( obj_ );

    ObjTerritoriesCIt cit = _objTerr.find( obj_ );
    if( _objTerr.end() == cit ) return Loc();

    assert( cit->second );
    assert( !cit->second->empty() );

    ObjTerritoryIt it = cit->second->begin();
    Loc min_l = (*it)->loc();
    Distance act_d, min_d = distance( loc_, min_l );
    while( cit->second->end() != ++it )
    {
        act_d = distance( loc_, (*it)->loc() );
        if( act_d < min_d )
        {
            min_l = (*it)->loc();
            min_d = act_d;
        }
    }

    return min_l;
}
*/
blazi::Loc blazi::Terrain::min_realobjloc( const Loc &loc_, const Obj *obj_ ) const
{
    assert( realvalid( loc_ ) );
    assert( obj_ );

    DEBUG( "    realobjloc, loc = " << loc_ );
    Loc res;
    switch( obj_->obj()->get_shape() )
    {
        case Object::CIRCLE :
        {
            Loc ol = obj_->loc();

            Vec2<real8> vect( loc_.x - ol.x, loc_.y - ol.y );
            vect.normalize();
            vect *= obj_->obj()->get_radius();
            DEBUG( "    vect = " << vect.x << "," << vect.y );

            res = Loc( ol.x + (sint4)vect.x, ol.y + (sint4)vect.y );
            DEBUG( "    res = " << res );

            break;
        }

        case Object::RECTANGLE :
        {
            Loc tp = obj_->locTP();
            Loc br = obj_->locBR();

            res = loc_;

            if( res.x < tp.x ) res.x = tp.x;
            else
            if( res.x > br.x ) res.x = br.x;

            if( res.y < tp.y ) res.y = tp.y;
            else
            if( res.y > br.y ) res.y = br.y;

            break;
        }

        default : ERR( "min_realobjloc: invalid shape" );
    }

    assert( realvalid( res ) );
    return res;
}


bool less( sint4 a_, sint4 b_ ) { return a_ < b_; }
bool less_equal( sint4 a_, sint4 b_ ) { return a_ <= b_; }

bool blazi::Terrain::fit( const Obj *obj_,
                          const Loc &proposal_,
                          const Obj *other_ ) const
{
    assert( obj_ );
    assert( realvalid( proposal_ ) );
    assert( other_ );

    if( obj_ == other_ ) return true;

    bool (*cmp)( sint4, sint4 ) = less;

    if( real2pos( obj_->loc() ) == real2pos( proposal_ ) )
    {
        cmp = less_equal;
    }

    // hack to support the opposite case (should be refactored)
    /*
    if( Object::CIRCLE == other_->obj()->get_shape() &&
        Object::RECTANGLE == obj_->obj()->get_shape() )
    {
        std::swap( obj_, other_ );
    }*/

    if( Object::CIRCLE == obj_->obj()->get_shape() )
    {
        switch( other_->obj()->get_shape() )
        {
            case Object::CIRCLE :
            {
                Loc oL = other_->loc();
                sint4 diffX = proposal_.x - oL.x;
                sint4 diffY = proposal_.y - oL.y;
                sint4 d = (sint4) sqrt( diffX * diffX + diffY * diffY );

                sint4 r = obj_->obj()->get_radius();
                sint4 oR = other_->obj()->get_radius();

//                DEBUG( "    fit(C) " << proposal_
//                                     << " [" << real2pos( proposal_ ) << "]"
//                                     << " ? " << other_->loc()
//                                     << " [" << real2pos( other_->loc() ) << "]"
//                                     << " , r = " << (r + oR < d)
//                                     << " , ["
//                                     << r << "," << oR << "," << d << "]" );
                return cmp(r + oR, (int)d);
            }

            case Object::RECTANGLE :
            {
                Loc tp = other_->locTP();
                Loc br = other_->locBR();

                sint4 r = obj_->obj()->get_radius();

                return cmp(proposal_.x + r, tp.x) || cmp(br.x, proposal_.x - r) ||
                       cmp(proposal_.y + r, tp.y) || cmp(br.y, proposal_.y - r);
            }

            default : ERR( "can_fit: not supp case(not circle or rect shape)" );
        }
    }
    else ERR( "can_fit: not supp case(not circle shape)" );
    return false;
}


bool blazi::Terrain::check_free( const Obj *obj_,
                                 const Loc &proposal_,
                                 bool consMob_ ) const
{
    assert( obj_ );
    assert( realvalid( proposal_ ) );

    PosSet ps;
    fill_pos_set( ps, obj_->obj()->get_radius(), proposal_ );

    FORALL( ps, pit )
    {
        FORALL( (*pit)->objs(), oit )
        {
            if( ( consMob_ || !(*oit)->mobile() ) &&
                *oit != obj_ && !fit( obj_, proposal_, *oit ) )
            {
                DEBUG( "    !fit: " << (*oit)->obj()->bp_name() << "  " << *oit
                       << "  " << Game::distance( *obj_->obj(), *(*oit)->obj() ) );
                return false;
            }
        }
    }

    return true;
}


void blazi::Terrain::fill_pos_set( PosSet &set_,
                                   sint4 radius_,
                                   const Loc &proposal_ ) const
{
    // inserting an existing position does not have any effect

    Loc l = real2pos( proposal_ );
    set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x, proposal_.y - radius_ ) );              // N
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x + radius_, proposal_.y - radius_ ) );    // NE
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x + radius_, proposal_.y ) );              // E
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x + radius_, proposal_.y + radius_ ) );    // SE
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x, proposal_.y + radius_ ) );              // S
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x - radius_, proposal_.y + radius_ ) );    // SW
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x - radius_, proposal_.y ) );              // W
    if( valid(l) ) set_.insert( get_pos(l) );

    l = real2pos( Loc( proposal_.x - radius_, proposal_.y - radius_ ) );    // NW
    if( valid(l) ) set_.insert( get_pos(l) );
}

