
#ifndef _TERRAIN_H
#define _TERRAIN_H

#include <set>
#include <list>
#include <iostream>

#include "Map.H"
#include "Tile.H"
#include "GameObj.H"
#include "GameTile.H"

#include "_Global.H"
#include "_Loc.H"
#include "_Obj.H"


#define FORALLCOMPASS(var)  \
        FORALLENUM(Tile::Compass,var,Tile::N,Tile::NW,1)


namespace blazi {


/** Sequence of locations which represents a path. */
class Path
{
public:

    typedef std::list<Loc> LocSeq;

public:

    Path() { _len = 0; }

    bool empty() const { return _path.empty(); }
    size_t size() const { return _path.size(); }
    const LocSeq& path() const { return _path; }

    void push_front( const Loc &loc_ );
    void push_back( const Loc &loc_ );

    Loc front() const { assert( !empty() ); return _path.front(); }
    void pop_front();

    Loc back() const { assert( !empty() ); return _path.back(); }
    void pop_back();

    sint4 length() const { return _len; }
    Path& concat( const Path *path_ );

private:

    LocSeq _path;
    sint4 _len;
};


/** Represents a position in the terrain. */
class Position
{
public:

    friend class Terrain;

    typedef std::set<const Obj*> ObjCont;

public:

    Position();

    const GameTile* get_tile() const { return _tile; }

    Loc loc() const { return _loc; }

    bool has_nomobile_obj() const { return _has_nomob_obj; }

    const ObjCont& objs() const { return _objs; }
    void add_obj( const Obj *obj_ );
    void rem_obj( const Obj *obj_ );

protected:

    void init( const GameTile *tile_, sint4 x_, sint4 y_ );

private:

    const GameTile *_tile;

    Loc _loc;
    bool _has_nomob_obj; // true if has a not mobile object
    ObjCont _objs;
};


/** Represents the terrain for a player. */
class Terrain
{
public:

    static const uint4 GROUND = 2; // fixme: read it somehow from the blueprint file

    typedef sint4 Distance;

    typedef Loc ObjLocation;
    typedef std::set<ObjLocation> ObjLocations;
    typedef boost::shared_ptr<ObjLocations> ObjLocationsPtr;

public:

    Terrain( const Map<GameTile> &map_, sint4 tile_points_ );
    ~Terrain();

    // Position obtaining functions

    Position* get_pos( sint4 x_, sint4 y_ )
              { return &_terrain[xy2ind(x_, y_)]; }
    const Position* get_pos( sint4 x_, sint4 y_ ) const
                    { return &_terrain[xy2ind(x_, y_)]; }

    Position* get_pos( const Loc &loc_ )
              { return &_terrain[loc2ind(loc_)]; }
    const Position* get_pos( const Loc &loc_ ) const
                    { return &_terrain[loc2ind(loc_)]; }

    Position* get_realpos( const Loc &loc_ )
              { return get_pos( real2pos(loc_) ); }
    const Position* get_realpos( const Loc &loc_ ) const
              { return get_pos( real2pos(loc_) ); }

    Position* operator()( sint4 x_, sint4 y_ ) { return get_pos(x_, y_); }
    const Position* operator()( sint4 x_, sint4 y_ ) const { return get_pos(x_, y_); }

    Position* operator()( const Loc &loc_ ) { return get_pos(loc_); }
    const Position* operator()( const Loc &loc_ ) const { return get_pos(loc_); }

    // Map related functions

    sint4 size() const { return get_map().get_num(); }
    sint4 width() const { return get_map().get_width(); }
    sint4 height() const { return get_map().get_height(); }

    sint4 xy2ind( sint4 x_, sint4 y_ ) const { return get_map().xy2ind(x_, y_); }
    sint4 loc2ind( const Loc &loc_ ) const { return get_map().xy2ind(loc_.x, loc_.y); }
    sint4 get_tile_points() const { return _tile_points; }
    Loc real2pos( const Loc &loc_ ) const; // sod(x,y) -> pos(x,y)
    Loc pos2real( const Loc &loc_ ) const; // pos(x,y) -> sod(x,y)

    // Object related functions

    Loc loc( const Obj *obj_ ) const;
    Loc posloc( const Obj *obj_ ) const;

    Loc prev_loc( const Obj *obj_ ) const;
    Loc prev_posloc( const Obj *obj_ ) const;

    /** Updates the position info of the terrain related to the given object. */
    void update_obj( Obj *obj_ );

    // Position and location related functions

    bool valid( const Loc &loc_ ) const;
    bool realvalid( const Loc &loc_ ) const;

    bool walkableW( const Position *pos_ ) const
         { return GROUND == pos_->get_tile()->get_typeW(); }
    bool walkableE( const Position *pos_ ) const
         { return GROUND == pos_->get_tile()->get_typeE(); }
    bool walkable( const Position *pos_ ) const
         { return walkableW(pos_) && walkableE(pos_); }

    /** Returns a location reached from the given one by using the
     *  specified compass. If it does not exist on the terrain, the
     *  function returns the undefined location. */
    Loc compass2loc( const Loc &loc_, Tile::Compass compass_ ) const;

    /** Check whether passing is possible between the two specified position locations.
     *  If so, a free location is given inside the position, otherwise an
     *  invalid location is returned.
     *  Note: from and to parameters have to identify adjacent positions now. */
    bool check_pos_pass( const Obj *obj_,
                         const Loc &from_,  // position loc
                         const Loc &to_,    // position loc
                         bool checkFree_ ) const;

    /** Check the possible obstacles (not included the given object which is
     *  considered to pass) between the two locations. If consMob parameter is
     *  true, not only the static objects are checked, but the mobile ones as well. */
    bool check_free_pass( const Obj *obj_,
                          const Loc &from_,
                          const Loc &to_,
                          bool consMob_ ) const;

    /** Returns a location through which the given destination can be reached.
     *  If such a location does not exist, returns an invalid one. */
    Loc dodge_point( const Worker *obj_, Loc to_, Loc to2_ ) const;

    /** Diagonal distance computations. */
    Distance distance( const Loc &from_, const Loc &to_ ) const;
    Distance distance( const Loc &from_, const ObjLocations &to_ ) const;
    Distance distance( const Obj *from_, const Obj *to_ ) const
             { return distance( posloc(from_), posloc(to_) ); }

    size_t obj_locs_size( const Obj *obj_ ) const;
    ObjLocationsPtr obj_locs( const Obj *obj_ ) const;
    Loc min_realobjloc( const Loc &loc_, const Obj *obj_ ) const;
    Loc min_objloc( const Loc &loc_, const Obj *obj_ ) const
        { return real2pos( min_realobjloc( pos2real(loc_), obj_ ) ); }

    typedef std::set<const Position*> PosSet;
    void fill_pos_set( PosSet &set_, sint4 radius_, const Loc &proposal_ ) const;

protected:

    typedef std::list<Position*> ObjTerritory;
    typedef ObjTerritory::iterator ObjTerritoryIt;

    void update_pos( const Obj *obj_, ObjTerritory *terr_, const Loc &loc_ );
    void update_realpos( const Obj *obj_, ObjTerritory *terr_, const Loc &loc_ )
         { return update_pos( obj_, terr_, real2pos(loc_) ); }

    bool fit( const Obj *obj_, const Loc &proposal_, const Obj *other_ ) const;
    bool check_free( const Obj *obj_, const Loc &proposal_, bool consMob_ ) const;

private:

    const Map<GameTile>& get_map() const { return _map; }

private:

    const Map<GameTile> &_map;
    const sint4 _tile_points;
    const sint4 _tile_points_per_4;

    const Distance _straight_step;
    const Distance _diagonal_step;

    Position *_terrain;

    typedef Sgi::hash_map< const Obj*,ObjTerritory*,PtrHash<Obj> > ObjTerritories;
    typedef ObjTerritories::iterator ObjTerritoriesIt;
    typedef ObjTerritories::const_iterator ObjTerritoriesCIt;

    ObjTerritories _objTerr;
};


inline bool Terrain::valid( const Loc &loc_ ) const
{
    return 0 <= loc_.x && loc_.x < width() &&
           0 <= loc_.y && loc_.y < height();
}


inline bool Terrain::realvalid( const Loc &loc_ ) const
{
    return 0 <= loc_.x && loc_.x < width() * get_tile_points() &&
           0 <= loc_.y && loc_.y < height() * get_tile_points();
}


inline Loc Terrain::real2pos( const Loc &loc_ ) const
{
    return Loc( loc_.x / get_tile_points(),
                loc_.y / get_tile_points() );
}


inline Loc Terrain::loc( const Obj *obj_ ) const
{
    return Loc( *obj_->obj()->sod.x,
                *obj_->obj()->sod.y );
}

inline Loc Terrain::posloc( const Obj *obj_ ) const
{
    return real2pos( loc(obj_) );
}

inline Loc Terrain::prev_loc( const Obj *obj_ ) const
{
    return Loc( obj_->obj()->sod.prev.x,
                obj_->obj()->sod.prev.y );
}

inline Loc Terrain::prev_posloc( const Obj *obj_ ) const
{
    return real2pos( prev_loc(obj_) );
}


} /* namespace blazi */


#endif /* _TERRAIN_H */

