#include "Global.H"
#include "Collision2D.H"
#include "Options.H"

#include "Game1_ST_BasicMove.H"
#include <boost/smart_ptr.hpp>

using namespace std;
using boost::scoped_ptr;

#define TEST(x)
//#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x

namespace Game1_SimpleTerrain
{
  const sint4 ST_BasicMove::ANTICIPATE_COLLISION_TICKS = 16; // 2 seconds * 8 ticks_per_second
  const sint4 ST_BasicMove::FINE_GRID_RES_BITS = 6;
  const uint4 ST_BasicMove::TILES_PER_MOTION_SECTOR = 3;

  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }

  static ScalarPoint get_center_sp(const Object *obj)
  {
    ScalarPoint l;
    obj->get_center(l.x, l.y);
    return l;
  }

  //-----------------------------------------------------------------------------
  void ST_BasicMove::add_options()
  {
    ::Options o("ST_BasicMove");
    o.put("-nomotionsectors", "don't use motion sectors for collision detection (test all pairs of units)");
  }

  //----------------------------------------------------------------------------------
  ST_BasicMove::ST_BasicMove(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 /*client_pID_*/, sint4 /*neutral_pID_*/)
    : tiles_x(tiles_x_), tiles_y(tiles_y_), tile_points(tile_points_)
  {
    sint4 seed = 0;
    ::Options::get("-seed", seed);
    rand.srand(seed);

    noMotionSectors = false;
    ::Options::get("-nomotionsectors", noMotionSectors);
  }

  ST_BasicMove::~ST_BasicMove() {

  }

  //----------------------------------------------------------------------------------
  void ST_BasicMove::execute_tasks(ST_Terrain::Tasks &tasks, 
                                   ST_Terrain::Objects &objsToStop,
                                   Vector<MoveCmd> &cmds, 
                                   Vector<StatusMsg> &msgs)
  {
    // cout << "ST_Terrain::execute_tasks" << endl;
    // Execute all active tasks
    /*for (Tasks::iterator j = tasks.begin(); j != tasks.end();) {
	Tasks::iterator i = j++;*/
    FORALL(tasks, i) {

      // Report failure of planning (no path at all)
      if (i->noPathFound ) {
        StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TaskId)&(*i);
        s.obj = *(i->task.objs.begin());
        s.type = StatusMsg::NO_PATH_FAILURE;
        msgs.push_back(s);
        continue;
      }

      // Skip invalid tasks.
      if (!i->isValid() ) { 
        TEST(cout << *i->task.objs.begin() << " task not valid" << endl;);
        continue;
      }

      UnitPath &cur_path = *(i->path);
      const Object *obj = cur_path.obj;

      // Skip tasks that are waiting for a retry
      if (cur_path.sleeping()) {
        TEST(cout << obj << " path sleeping" << endl;);
        continue;
      }

      // Skip objects for which we gave a (move) action that is still pending
      /*
      if (obj->is_pending_action()) {
      TEST(cout << obj << " pending action" << endl;)
      continue;
      }
      */

      bool atGoal = is_at_goal(obj, i->task, i);

      // Look ahead: if the object will get to a waypoint in the next tick,
      // give the next action now the prevent it from wasting a tick on stopping.
      // This will cut a single tick at each waypoint, though it will really prevent 
      // the object from stopping only if there's no (network or processing) lag
      // between this client and the server.
      // atWaypoint can be true only if the current waypoint is an intermediate and
      // not the *final* waypoint/goal.
      bool atWaypoint = cur_path.current_target > 0 &&  // reverse order, goal==0
        is_at_location(obj, cur_path.path.at(cur_path.current_target));

      // Skip a moving object which is currently not at a waypoint/goal
      if ((obj->get_speed() > 0 || obj->get_moving() != 0) && !atWaypoint && !atGoal ) {
        TEST(cout << obj << " atWaypoint=" << atWaypoint << " speed=" << obj->get_speed() << " is_moving=" << obj->get_moving() << endl;);
        continue;
      }

      // ok, so it needs attention:
      // 1. maybe it had reached the goal
      // 2. maybe it's at a waypoint (with the lookahead)
      // 3. maybe it needs the first orders
      // 4. maybe it unexpectedly collided with something

      if (atGoal ) {

        TEST(cout << obj << " arrived" << endl;)

        StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TaskId)&(*i);
        s.obj = obj; 
        s.type = StatusMsg::ARRIVED;
        msgs.push_back(s);

      } else if( atWaypoint || 
                 cur_path.no_orders_yet || 
                 cur_path.stopped_to_avoid_collision ) {

        // in all these cases we should just issue a move command

        TEST(cout << obj << " atWaypoint=" << atWaypoint
                  << " no_orders_yet=" << cur_path.no_orders_yet
                  << " stopped_to_avoid_collision=" << cur_path.stopped_to_avoid_collision
                  << endl; );

        // if we are at a waypoint, move to next waypoint
        if( atWaypoint )
          cur_path.current_target--;
        cur_path.no_orders_yet = false;
        cur_path.stopped_to_avoid_collision = false;
        add_move_command(cur_path, cmds);

      } else {

        // Unexpected stop. The current policy is:
        // 1-2. First time => wait a little and send the move command again.
        // 3-4. Second stop at the same place => mark as failed and replan.
        // 5. Third stop at the same place => report failure.

	cout << obj << " unexpectedly stopped, collision_step=" << cur_path.collision_step << endl;

        Loc cur_position = get_center_loc(obj);

        // First collision in this position:
        if (cur_path.spot != cur_position ) {
          cur_path.collision_step = 1;
          cur_path.mark_spot(cur_position);
          // sleep for a short random time, 100-599 millisecs
          cur_path.sleep(100 + (rand.rand_uint4() % 500));
        } else {
          cur_path.collision_step++;

          // Just finished the sleeping:
          if (cur_path.collision_step == 2 ) {
            add_move_command(cur_path, cmds);
          }
          // Second collision
          else if (cur_path.collision_step == 3 ) {
            i->isFailed = true;
          }
          // Back from replanning
          else if (cur_path.collision_step == 4 ) {
            add_move_command(cur_path, cmds);
          }
          // Third collision
          else if (cur_path.collision_step == 5 ) {
            StatusMsg s;
            // fixme: this is not robust to future changes in the meaning of TaskId.
            s.task_id = (TaskId)&(*i);
            s.obj = obj;
            s.type = StatusMsg::MOVEMENT_FAILURE;
            msgs.push_back(s);
	    TEST2(cout << "pf movement failure" << endl); 
          }
        }
      }
    }

    // Priority of commands:
    // 1. stops to avoid a collision.
    // 2. moves.
    // 3. other stops (resulting from a cancelled task, etc.).

    // 2 above 3:
    FORALL(cmds, c) {
      objsToStop.erase(c->obj);
    }

    // With all movement commands were given while assuming an ideal-no-collisions
    // world, now check if any of our controlled objects are going to collide
    // (because of an existing movement vector, and also consider commands
    // we just intended to give).
    resolve_friendly_collisions(tasks, objsToStop);

    // 1 above 2:
    for(typeof(cmds.begin()) c = cmds.begin() ; c != cmds.end() ; ) {
      FIND(objsToStop, s, c->obj);
      if( s != objsToStop.end() )
        c = cmds.erase(c);
      else
        c++;
    }

    // Return also the stop commands
    FORALL(objsToStop, oi) {
      const Object *obj = *oi;

      MoveCmd m;
      m.obj = obj;
      //obj->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
      m.speed = 0; // 0 -> stop
      cmds.push_back(m);
    }
    objsToStop.clear();
  }

  //----------------------------------------------------------------------------------
  void ST_BasicMove::add_move_command(const UnitPath& path, Vector<MoveCmd> &cmds)
  {
    MoveCmd m;
    m.obj = path.obj;
    m.next_loc = path.path.at(path.current_target);
    m.speed = path.obj->get_max_speed();
    cmds.push_back(m);
    TEST(cout << m.obj << " sending move command, target=(" << m.next_loc.x << "," << m.next_loc.y << ")" << endl;);
  }

  //----------------------------------------------------------------------------------
  void ST_BasicMove::resolve_friendly_collisions(ST_Terrain::Tasks &tasks, ST_Terrain::Objects &objsToStop)
  {
    // * Identify collisions between pairs of controlled units in the next few ticks.
    // * For each such expected collision, stop one of the objects in this turn (and
    //   then let the usual execution resume its movement later).
    // * While in the process of identifying pairs, remember the objects we already
    //   stopped and consider this in the following iterations.
    // * Assuming only one object per task

    scoped_ptr< SimpleMap<Sector> > sectors;
    if( !noMotionSectors ) {
      // To avoid the full O(N^2) comparisons, we insert objects into motion
      // sectors, and test only pairs that are in the same sector.
      
      // First, compute motion squares for all objects.
      Vector<UnitSquare> motionSquares;
      FORALL(tasks, t) {
        //  cout << "Computing motion square."<< endl;
        motionSquares.push_back(find_motion_square(t));
      }

      // Now, insert motion squares to all overlapping sectors.
      const sint4 motion_sectors_w = (tiles_x + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
      const sint4 motion_sectors_h = (tiles_y + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
      sectors.reset( new SimpleMap<Sector>(motion_sectors_w, motion_sectors_h) );
      insert_into_sectors(motionSquares, *sectors);
    }
    else {
      // Use a crude mechanism (check all-against-all), might be better for few units
      sectors.reset( new SimpleMap<Sector>(1,1) );
      FORALL(tasks, t)
        sectors->at(0).tasks.push_back(&*t);
    }

    // Walk the sectors, and compare pairs of tasks that are in the same sector.
    FORALL(*sectors, h) {

      // cout << "Checking " << active.size() << " sectors for collisions. " << endl;
      // cout << "This sector has " << (*h)->others.size() << " units in it. " << endl;

      // Test all pairs of objects/tasks:
      FORALL(h->tasks, i) {
        const Object *obj_i = *((*i)->task.objs.begin());

        typeof(i) j = i;
        for( j++ ; j != h->tasks.end() ; j++ ) {

          const Object *obj_j = *((*j)->task.objs.begin());
          //  cout << "Checking object: " << obj_i << " against object: " << obj_j << endl;

          if( will_collide(i, j) ) {
	    TEST(cout << "GOING TO COLLIDE: " << obj_i << "@" << get_center_loc(obj_i) << " and " << obj_j << "@" << get_center_loc(obj_j) << endl); 

            // current policy:
            // 1. try to (arbitrarily) stop the second object
            if( !will_collide(i, j, false, true) ) {
              // if stopping worked, j must have been moving, so we can 
              // safely use j->path
              objsToStop.insert(obj_j);
              (*j)->path->stopped_to_avoid_collision = true;
            }
            // 2. if the collision is still expected (eg they were heading 
            //    "into" one another), then stop both objects and 
            //    replan their paths
            else {
              // don't use the i->path, j->path since they might still not exist.
              objsToStop.insert(obj_i);
              objsToStop.insert(obj_j);
              (*i)->isFailed = true;
              (*j)->isFailed = true;
            }
          }
        }
      }
    }
  }

  //----------------------------------------------------------------------------------
  void ST_BasicMove::get_fine_spatial_params(const Object *o, 
                                           ScalarPoint &target,
                                           ScalarPoint *fine_position, 
                                           ScalarPoint *fine_velocity)
  {
    ScalarPoint position = get_center_sp(o);
    ScalarPoint velocity = target;
    //o->get_center(position.x, position.y);
    //velocity = target;
    velocity.sub(position);

    // convert to fine-grid coordinates

    if( fine_position != NULL ) {
      fine_position->x = fine_grid(position.x);
      fine_position->y = fine_grid(position.y);
    }
    if( fine_velocity != NULL ) {
      fine_velocity->x =  fine_grid(velocity.x);
      fine_velocity->y =  fine_grid(velocity.y);
    }
    target.x = fine_grid(target.x);
    target.y = fine_grid(target.y);


    // normalize velocity to unit length and extend according to the object's speed

    if( fine_velocity != NULL && *fine_velocity != ScalarPoint(0,0) ) {
      // fixme: if object has a pending command, before moving, its speed is still 0. Should we use max_speed, or something else?
      if( o->get_speed() == 0 )
        *fine_velocity = ScalarPoint(0,0);
      else {
        real8 squared_norm = fine_velocity->l2_norm_sq();
        real8 inv_norm = 1.0 / sqrt(squared_norm);
        //cout << "squared_norm=" << squared_norm << " inv_norm=" << inv_norm << " fine_grid(o->get_speed())=" << fine_grid(o->get_speed()) << endl;
        fine_velocity->scale(inv_norm * fine_grid(o->get_speed()));
      }
    }
  }

  //----------------------------------------------------------------------------------
  bool ST_BasicMove::will_collide(Vector<ST_Task*>::iterator i, Vector<ST_Task*>::iterator j, bool assume_obj1_stopped, bool assume_obj2_stopped)
  {
    // fixme: more efficent to cache obj1, since we test all pairs in a double loop

    // will these two objects collide in the next few ticks ?
    // look ahead only until the current waypoint for each object.
    // work with a higher accuracy.

    const Object *obj1 = *((*i)->task.objs.begin());
    ScalarPoint position1, velocity1, target1;
    real8 time_to_target1;

    // if it doesn't already have a move executed ==> immobile object
    if( !(*i)->isValid() || (*i)->path->sleeping() || (*i)->path->no_orders_yet || assume_obj1_stopped ) {
      // pass a dummy target and then adjust params
      ScalarPoint nn(0,0);
      get_fine_spatial_params(obj1, nn, &position1, NULL);
      velocity1 = ScalarPoint(0,0);
      time_to_target1 = ANTICIPATE_COLLISION_TICKS; // large enough to be ignored
    } 
    // else, the object is mobile
    else {
      target1 = ScalarPoint((*i)->path->path.at((*i)->path->current_target).x, (*i)->path->path.at((*i)->path->current_target).y);
      get_fine_spatial_params(obj1, target1, &position1, &velocity1);
      time_to_target1 = sqrt( (real8)(position1.l2_dist_sq(target1)) / velocity1.l2_norm_sq() );
    }


    const Object *obj2 = *((*j)->task.objs.begin());
    ScalarPoint position2, velocity2, target2;
    real8 time_to_target2;

    if( !(*j)->isValid() || (*j)->path->sleeping() || (*j)->path->no_orders_yet || assume_obj2_stopped ) {
      ScalarPoint nn(0,0);
      get_fine_spatial_params(obj2, nn, &position2, NULL);
      velocity2 = ScalarPoint(0,0);
      time_to_target2 = ANTICIPATE_COLLISION_TICKS;
    } else {
      target2 = ScalarPoint((*j)->path->path.at((*j)->path->current_target).x, (*j)->path->path.at((*j)->path->current_target).y);
      get_fine_spatial_params(obj2, target2, &position2, &velocity2);
      time_to_target2 = sqrt( (real8)(position2.l2_dist_sq(target2)) / velocity2.l2_norm_sq() );
    }

    // If they move at the same speed and at the same direction they won't collide
    if( velocity1 == velocity2 )
      return false;

    // if they will always to move like this, if and when will they collide?
    sint4 sum_of_radii = fine_grid(obj1->get_radius() + obj2->get_radius());
    real8 squared = (real8)(sum_of_radii) * sum_of_radii;

    /*James:  I found a bug.  When running this code for game2 with -nomotionsectors the following line produces an assert failure.
     It seems that when comparing two objects that are really far away the square function which is called in collision_time_circle_point()
     fails an assert.*/

    real8 collision_time = collision2d::collision_time_circle_point(position1, squared, velocity1, position2, velocity2);

    // will they collide at all, AND within the window of time that interests us,
    // AND before any of them reaches its target
    return collision_time != collision2d::no_collision && 
           collision_time <= ANTICIPATE_COLLISION_TICKS &&
           collision_time <= time_to_target1 &&
           collision_time <= time_to_target2;
  }

  //-------------------------------------------------------------------
  void ST_BasicMove::insert_into_sectors(const Vector<UnitSquare> &motionSquares,
                                       SimpleMap<Sector> &sectors)
  {
    const sint4 motion_sectors_w = (tiles_x + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
    const sint4 motion_sectors_h = (tiles_y + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
    const sint4 x_points_per_sector = (tiles_x * tile_points + motion_sectors_w-1) / motion_sectors_w;
    const sint4 y_points_per_sector = (tiles_y * tile_points + motion_sectors_h-1) / motion_sectors_h;

    FORALL (motionSquares, sqIt) {

      // square world-coords translated to sector-coords
      sint4 x0 = sqIt->topLeft.x / x_points_per_sector;     //(x - r);
      sint4 x1 = sqIt->bottomRight.x / x_points_per_sector; //(x + r);
      sint4 y0 = sqIt->topLeft.y / y_points_per_sector;     //(y - r);
      sint4 y1 = sqIt->bottomRight.y / y_points_per_sector; //(y + r);

      sint4 mw = sectors.get_w();
      sint4 mh = sectors.get_h();

      if (x0 < 0) x0 = 0;
      if (y0 < 0) y0 = 0;
      if (x1 >= mw) x1 = mw-1;
      if (y1 >= mh) y1 = mh-1;

      // insert to each sector
      for (sint4 y = y0; y <= y1; ++y) {
        for (sint4 x = x0; x <= x1; ++x) {
          //          if( x < 0 || x >= sectors.get_w() || y < 0 || y >= sectors.get_h() ){
          //            cout << "x = " << x << " y = " << y << endl;
          //          }
	  sectors(x,y).tasks.push_back(sqIt->task);
        }
      }
    }
  }
  
  //-------------------------------------------------------------------
  UnitSquare ST_BasicMove::find_motion_square(const ST_Terrain::TIter &task)
  {
    // If the object doesn't have a valid path
    if( !task->isValid() ) {
      // it is not mobile so use its current position
      const Object *obj = *task->task.objs.begin();
      Loc position = get_center_loc(obj);
      return UnitSquare(position, position, obj, &*task);
    }
    // A valid path
    else {
      const UnitPath &cur_path = *(task->path);
      const Object *obj = cur_path.obj;
      sint4 curx, cury; 
      obj->get_center(curx,cury);  // Current location

      // Find the position at a close future time  
      Loc future; 
      sint4 waypoints = get_future_position(cur_path, ANTICIPATE_COLLISION_TICKS, future);

      // Create a bounding box, enlarged by the obj's radius.
      std::vector<sint4> x_coords;
      std::vector<sint4> y_coords;

      FORS(i, waypoints+1){
        x_coords.push_back((*(cur_path.path.end()-i-1)).x);
        y_coords.push_back((*(cur_path.path.end()-i-1)).y);
      }    

      x_coords.push_back(future.x);
      y_coords.push_back(future.y);
      x_coords.push_back(curx);
      y_coords.push_back(cury);

       
      // Create a square containing the object and its motion. Out-of-bounds coords are OK.
      sint4 rad = obj->get_radius(); // The object's radius
      sint4 x0, y0, x1, y1;
      x0 = *min_element(x_coords.begin(), x_coords.end()) - rad;
      y0 = *min_element(y_coords.begin(), y_coords.end()) - rad;
      x1 = *max_element(x_coords.begin(), x_coords.end()) + rad;
      y1 = *max_element(y_coords.begin(), y_coords.end()) + rad;

      // constarin to map boundaries
      x0 = max(x0, 0);
      y0 = max(y0, 0);
      x1 = min(x1, tiles_x*tile_points - 1);
      y1 = min(y1, tiles_y*tile_points - 1);


     /*       
      Loc topLeft(x0 , y0);
      Loc bottomRight(x1 , y1);
      
      //  cout << "motion square is from (" << topLeft->x << "," << topLeft->y << ") to (" ;
      //   cout << bottomRight->x << "," << bottomRight->y << ")" << endl;
      */
      return UnitSquare(Loc(x0, y0), Loc(x1, y1), obj, &*task);
    }
  } 

  //-------------------------------------------------------------------
  sint4 ST_BasicMove::get_future_position(const UnitPath &cur_path, sint4 when, Loc &where)
  {
    const Object *obj = cur_path.obj;
    where = get_center_loc(obj);
    
    sint4 v = obj->get_speed();
    if( v == 0 || when <= 0 || cur_path.path.size() == 0 ) {
      return -1;
    }

    real8 remaining_dist = v * when;
    sint4 got_to_waypoint = -1;
    real8 dist_to_next = 1; // =1 to prevent a 'might be used uninitialized' warning.
     
    TEST2(cout << "initially remaining_dist = " << remaining_dist << endl;);

    FORALL_REV_B(cur_path.path, next_waypoint_it) {

      dist_to_next = where.distance(*next_waypoint_it);
      remaining_dist -= dist_to_next;
      if(remaining_dist <= 0)
	break;
      where = *next_waypoint_it;
      got_to_waypoint++;
      TEST2(cout << "got to " << where << " with remaining_dist = " << remaining_dist << endl;);
    }
    
    // if not at end of path
    if( next_waypoint_it != cur_path.path.rend() ) {

      const Loc &next_waypoint = *next_waypoint_it;

      TEST2(cout << "current_pos = " << where << endl;
            cout << "next_waypoint = " << *next_waypoint_it << endl;
            cout << "dist_to_next = " << dist_to_next << endl;
            cout << "remaining_dist = " << remaining_dist << endl; );

      // compute the position inside the final leg
      if(remaining_dist == 0){
        where = next_waypoint;
        return got_to_waypoint+1;
      }

      real8 missing_to_next_wp_ratio = abs(remaining_dist/dist_to_next);
      assert(missing_to_next_wp_ratio != 0 );
      TEST2(cout << "missing_to_next_wp_ratio = " << missing_to_next_wp_ratio << endl;);
      
      sint4 dx = where.x - next_waypoint.x;
      sint4 dy = where.y - next_waypoint.y;

      where = Loc(next_waypoint.x + (sint4)ceil(dx*missing_to_next_wp_ratio - 0.5), 
                  next_waypoint.y + (sint4)ceil(dy*missing_to_next_wp_ratio - 0.5));
      //where = new Loc( next_waypoint.x + (sint4)ceil(x2*rat - 0.5) , next_waypoint.y + (sint4)ceil(y2*rat - 0.5));
    }
    TEST2(cout << "Object will be at " << where << " in " << when << " ticks." << endl;);
    return got_to_waypoint;
  }

  //-------------------------------------------------------------------

  bool ST_BasicMove::is_at_location(const Object* obj, const Loc& target)
  {
    Loc position = get_center_loc(obj);
    // The 1-turn look-ahead causes collisions when navigating difficult paths 
    // (like inside the mineral cluster); I'm temporarily dropping it until
    // further testing: delays at waypoints are better then unexpected collisions.
    // -- ddeutscher
    //return position.distance(target) <= obj->get_speed();
    return position == target;
  }

  //-------------------------------------------------------------------
  
  bool ST_BasicMove::is_near_location(const Object *obj, const Loc& target, sint4 range){
    Loc position = get_center_loc(obj);
    return position.distance(target) <= range;
  }

  //-------------------------------------------------------------------

  // st_task param is only needed for the hack below...
  bool ST_BasicMove::is_at_goal(const Object* obj, const Task& task, const ST_Terrain::TIter &st_task)
  {
    // Getting to the goal depends on the definition of the task:
    if (task.goal.target == Goal::OBJ ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
	return( obj->distance_to(*task.goal.obj) <= 1 );
	// using 1-tick look-ahead:
	//return( obj->distance_to(*task.goal_obj) <= obj->get_speed() );
	break;

      case Goal::VICINITY :
	return (obj->distance_to(*task.goal.obj) <= task.goal.distance);
	break;

      case Goal::ATTACK :
	//return(is_in_range(obj, task.goal.obj));
        assert(0 && "Not implemented yet");
	break;

      default:
	ERR("Unknown mode");
      }
    } 
    else if (task.goal.target == Goal::LOCATION ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
        // Temporary hack, until final approach is corrected for "LOCATION" 
        // targets: Since the planned path might finish at a location which is only
        // near the requested goal, compare to the planned goal instead of the 
        // requested goal. (When fixed, is_at_goal should be static again).
        if( st_task->isValid() ) {
          TEST( if( get_center_loc(obj) == st_task->path->path[0] )
                  cout << obj << " arrived at location " << get_center_loc(obj) << ", compared to the last location of the planned path" << endl; );
          return get_center_loc(obj) == st_task->path->path[0];
        }
        // using 1-tick look-ahead; also suffers from low-resolution accuracy problems
        TEST( if( is_at_location(obj, task.goal.loc) )
                cout << obj << " arrived at location " << get_center_loc(obj) << ", compared to the last location of the planned path" << endl; );       
	return is_at_location(obj, task.goal.loc);
	break;

      case Goal::VICINITY :
	return is_near_location(obj, task.goal.loc, task.goal.distance);
	break;

      case Goal::ATTACK :
	//return(is_in_range(obj, task.goal.loc));
	assert(0 && "Not implemented yet");
	break;

      default:
	ERR("Unknown mode");
      }
    } else {
      ERR("Unknown target type");
    }
    return false;
  }
  //-------------------------------------------------------------------

#if 0 // This needs coupling with GameObj, which we strive to avoid. As I don't
      // need the "ATTACK" tasks for game1, I commented it out. --ddeutscher
  bool ST_BasicMove::is_in_range(const Object *moving_obj, const Loc goal_loc){
    //fixme: does not check if unit can see the target. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_const_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }

    Loc obj_loc = get_center_loc(moving_obj);
   
    return ( (obj_loc.distance(goal_loc) - moving_obj->get_radius()) <= range );

    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }

  //-------------------------------------------------------------------

  bool ST_BasicMove::is_in_range(const Object *moving_obj, const Object *goal_obj){
    //fixme: does not check if unit can see each other. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_const_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }
  
    //if not a GameObj or with no weapon range, should behave exactly
    //like the check for touch.
    return ( moving_obj->distance_to(*goal_obj) <= range );
    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }
#endif

  //-----------------------------------------------------------------------------
  void ST_BasicMove::insert_object(const Object * /*obj*/)
  {
  }

  //-----------------------------------------------------------------------------
  void ST_BasicMove::remove_object(const Object * /*obj*/)
  {
  }

  //-----------------------------------------------------------------------------
  void ST_BasicMove::update_object(const Object * /*obj*/)
  {
  }

  //-----------------------------------------------------------------------------
  void ST_BasicMove::insert_boundary(const Segment & /*l*/)
  {
  }

}   // End of namespace SimpleTerrain
