#ifndef Game1_ST_ForceFieldMove_H
#define Game1_ST_ForceFieldMove_H

#include "Game1_SimpleTerrain.H"

namespace Game1_SimpleTerrain {

  /** Two Dimensional Vector */
  class Vec_2D{
  public:
    sint4 x;/// x-coord
    sint4 y;/// y-coord
    Vec_2D(sint4 x_=0, sint4 y_=0) :  x(x_), y(y_){};
    Vec_2D(const Vec_2D &other) :  x(other.x), y(other.y){};
    Vec_2D& operator=(const Vec_2D &other) { 
      if( this != &other ) {
        x = other.x;
        y = other.y;
      }
      return *this;
    };
    Vec_2D(const Loc &a, const Loc &b) : x(b.x - a.x), y(b.y - a.y) {};
    Vec_2D operator+(const Vec_2D &other) const; ///Vector addition
    Vec_2D operator-(const Vec_2D &other) const; ///Vector subtraction
    Vec_2D project_onto(const Vec_2D &other) const; /// Project the vector onto the line defined by the given vector
    sint4 dot_prod(const Vec_2D &other) const;
    real8 length() const;
    Vec_2D multiply(real8 k) const;
    bool operator==(const Vec_2D &other) const;
    bool operator!= (const Vec_2D &other) const;

    inline static sint4 round(real8 real) {
      return (sint4)floor(real + 0.5);
    }
  };

  /// Simple Movement-execution module
  class ST_ForceFieldMove {
  public:
    ST_ForceFieldMove(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_);
    ~ST_ForceFieldMove();

    /// add command-line options specific to this implementation, using static methods of class Options
    static void add_options(void);

    /// Monitor task execution, return move commands and status messages.
    void execute_tasks(ST_Terrain::Tasks &tasks, 
                       ST_Terrain::Objects &objsToStop,
                       Vector<MoveCmd> &cmds, 
                       Vector<StatusMsg> &msgs);

    //---------------------------------------------------------------------
    /// \name Changes to the world.
    //---------------------------------------------------------------------
    //@{
    void insert_object(const Object *obj);
    void remove_object(const Object *obj);
    void update_object(const Object *obj);
    void insert_boundary(const Segment& line);
    //@}
    //---------------------------------------------------------------------

    // public to allow REGISTER_TYPEOF definitions
    typedef std::map<const Object*, Loc>  Obj2LocMap;
    typedef std::map<Loc, bool> SegmentEndsMap;

  private:

    //---------------------------------------------------------------------
    /// \name Internal accessories
    //---------------------------------------------------------------------
    //@{
    /** Test if the object is at the given location, 
	accounting for a 1-tick look-ahead for moving objects. */
    static bool is_at_location(const Object* obj, const Loc& target);
    /** Test if the distance from the object to the target is within the given range.*/
    static bool is_near_location(const Object* obj, const Loc& target, sint4 range);
    /** Test if the object is at the goal of the given task, implementing 
        (some) of the various definitions of 'goal'. */
    bool is_at_goal(const Object* obj, const Task& task, const ST_Terrain::TIter &st_task);

#if 0
    /** Test if the object is within firing range */
    static bool is_in_range(const Object* moving_obj, const Object *goal_obj);
    /** Test if the location is within firing range */
    static bool is_in_range(const Object* moving_obj, const Loc goal_loc);
#endif 

    /** Add a move command to the Vector of commands, telling the object to 
	move to the target location at maximum speed.
	Also, updates the location associated with the object for forcefield purposes
    */
    void add_move_command(const Object *obj, Loc target, Vector<TerrainBase::MoveCmd> &cmds);
    /** Determines what team an object is on.
	If the team cannot be determined, returns -1 */
    sint4 what_team(const Object *obj);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Force Field methods
    //---------------------------------------------------------------------
    //@{
    /** 
	Calculates the direction that an object should move to, based on the path
	they are following and any obstacles in the vicinity.
	@param path Path data
	@param output reference to a location that in the direction that the object should
	move to.
	fixme: maybe a location should be returned, instead of the pass by reference
    */
    void get_forcefield_target(UnitPath &path, Loc &output);
    
    /**
      @param object The object that is moving
      @param attractor The point that is attracting the object
      @param output reference to a location that is in the direction that the
      object should move to.
    */
    void get_forcefield_target(const Object *obj, const Loc &attractor, Loc &output);

    /** used for pushing, where there is no attractor*/
    void get_push_target(const Object *obj, Loc &output);
    
   
    /** Adjusts the vector so that it is not directed more than 90 degrees on either side
	of the target vector.*/ 
    void restrict_vector(const Vec_2D &target_vec, Vec_2D &vec);
    
    /** Finds the location of point on a line that is closest to the given location*/
    Loc find_point_on_line(const Loc &loc, const Loc &line_start, const Loc &line_end);

    /** Calculates the repulsion caused by a point on an object and returns
	it as a vector*/
    Vec_2D get_point_repulsion(const Object *moving_obj, const Loc &point, sint4 radius = 0);
    /** Calculates the repulsion caused by a line on an object and returns
	it as a vector */
    Vec_2D get_line_repulsion(const Object *moving_obj, const Segment &seg);
    
    /** Calculates the position of the attractor along the path.
	The attractor is positioned a set distance ahead of the point closest to the
	object on the path.
     */
    Loc get_attractor(UnitPath &path);
    //@}
    //---------------------------------------------------------------------


    // prevent copying as we currently don't need it.
    ST_ForceFieldMove(const ST_ForceFieldMove& other);
    ST_ForceFieldMove& operator=(const ST_ForceFieldMove& other);

    Random rand;
    Obj2LocMap      objects;      ///< Used to calculate object repulsion
    Vector<Segment> segments;     ///< Used to calculate wall repulsion
    SegmentEndsMap  segment_ends; ///< Used to calculate repulsion from the ends of walls

    sint4 client_pID;             ///< Our played ID
    bool pushing_on;              ///< use pushing of objects to make way
    bool wp_attract;              ///< units are attracted to the next waypoint instead of following the path itself

    //---------------------------------------------------------------------
    /// \name Constance
    //---------------------------------------------------------------------
    //@{

    /** This multiplied by the max speed of an object gives the range at which objects
	and walls will exert a force on the objec*/
    static const sint4 FIELD_RANGE_MULTIPLIER = 2;

    /** The strength of attractors (does not vary with distance)*/ 
    static const sint4 ATTRACTOR_FORCE = 26;
    
    /** This divided by the distance squared to the moving object gives the strength
	of the repulsors*/
    static const sint4 REPULSOR_BASELINE = 200;
    
    /** No longer used*/
    static const sint4 WAYPOINT_PROX = 2;

    /** this multiplied by the object max speed gives the distance
	ahead of the object on the path that the attractor will be set to*/ 
    static const sint4 PATH_LOOK_AHEAD = 3;
    //@}
    //---------------------------------------------------------------------
  };

} // end of namespace SimpleTerrain

REGISTER_TYPEOF(2601, Game1_SimpleTerrain::ST_ForceFieldMove::Obj2LocMap::iterator);
REGISTER_TYPEOF(2602, Game1_SimpleTerrain::ST_ForceFieldMove::Obj2LocMap::const_iterator);
REGISTER_TYPEOF(2603, Game1_SimpleTerrain::ST_ForceFieldMove::SegmentEndsMap::iterator);
REGISTER_TYPEOF(2604, Game1_SimpleTerrain::ST_ForceFieldMove::SegmentEndsMap::const_iterator);

#endif
