/* Autonomously runs an ORTS tournament
 * (c) Krysta Mirzayans
 * (c) Michael Buro
 *
 * licensed under GPL 3
 */

// $Id: tmanager.c,v 1.1 2007/05/24 04:40:03 orts_mburo Exp $

// to avoid "closed by foreign host" messages from ssh increase
// MaxStartups in /etc/ssh/sshd_config or slow down thread creation by
// decreasing MAX_START or increasing THREAD_START_WAIT

#include "tmanager.h"
#include "TM_Score.H"
#include <unistd.h>
#include <sys/wait.h>
#include <boost/format.hpp>

#define CHECK_IS_UP      0       // check whether computer is up using ruptime
#define PRINT_CMDS       1       // print ssh commands
#define PRINT_DONE       0       // print a message when server/client is done
#define PRINT_SRV_OUTPUT 0       // print server output

const string ERROR = "! ERROR";  // error message in .gam file when number of game attempts exceeds MAX_TRIES
const int MAX_TRIES = 3;         // maximum number of allowable attempts for a game

const int MAX_START = 10;        // maximum number of games started in checkComputers
const int SERVER_START_WAIT = 5; // sleep after starting server
const int CLIENT_START_WAIT = 1; // sleep between clients

const string SSH_PAR = "-Y";     // ssh parameter string
//const string SSH_PAR = "-x -o ConnectionAttempts=10";

// GLOBAL GAME DATA STRUCTURES, LOCK BEFORE USAGE!

Vector<Game>     games;     // list of games to be played
Vector<Computer> computers; // list of computers
Vector<Player>   players;   // list of players

int game_type=0;    // 1..4 (for scoring function)

// depend on game type (see below)
bool single_player;  // false: two-player
bool single_game;    // false: pair-games

score_func_p score_func = 0;

// global lock
pthread_mutex_t mutexGames;   

// lock/unlock global data structures
#define LOCK   pthread_mutex_lock(&mutexGames)
#define UNLOCK pthread_mutex_unlock(&mutexGames)

pthread_attr_t attr; // thread creation attributes

bool opt_v = false;  // true=>verbose version: print messages
                     // throughout to stdout, false=> silent
                     // version: no stdout messages 

bool opt_s = false;  // print scores
bool opt_c = false;  // generate list of computers
bool opt_g = false;  // generate list of games


/**
 * Error reporting function, which displays the given error
 * message to standard out and terminates the program with code 20.
 *
 * @param msg  error message to be displayed
 */
void Error(string msg)
{
  if (msg == "usage") {
    cerr << "\nusage: tmanager -h | [-c] [-v] [-s] [-g] <basename>\n\n\
 -h    : detailed description \n\
 -v    : verbose [not used right now]\n\
 -c    : create computer list\n\
 -s    : print scores\n\
 -g    : print game list\n\
\n\
 tournament configuration: <basename>.cfg \n\
 game results are stored in: <basename>.gam \n\n";
    exit(20);
  }
  cerr << "*** " << msg << endl;
  
  exit(20);
}

/**
 * Helper function to get the current time using gettimeofday.
 *
 * @return timeval type which contains the timing information
 */
timeval get_time()
{
  timeval ti; 
  gettimeofday(&ti ,NULL);
  return ti;
}

/**
 * Helper function to get the number of seconds elapsed
 * since 1/1/1970.
 *
 * @param ti  timeval parameter containing the timing information
 * @return    time in seconds since 1/1/1970
 */
double to_double(const timeval &ti)
{
  return ti.tv_sec + (ti.tv_usec/1000000.0);
}

/**
 * Helper function for parsing the configuration file which
 * skips sections of comments (denoted by #).
 * 
 * @param in  the input stream (configuration file)
 * @return    a bool flag indicating whether or not end-of-file is reached (true - eof, false - otherwise)
 */
bool skipComments(istream &in)
{
  string s;
  char c;
  FOREVER {
    if ((c = in.peek())== EOF) return true;
    if (c == '\n' || c == ' ') { in.get(); continue; }
    if (c == '#') { getline(in, s); continue; }
    break;
  }
  return false;
}

/**
 * Helper function for converting an int to a string.
 *
 * @param i   integer to be converted to a string
 * @return    string version of the integer
 */
string intToString(int i)
{
  ostringstream s;
  s << i;
  return s.str();
}

/**
 * Helper function for converting a string representing
 * a floating point number to a double.
 *
 * @param s  the string to be converted to a double
 * @return   the double value represented by the string parameter
 */
double stringToNum(string s)
{
  int temp = 0;
  int num1 = 0;
  double num2 = 0;
  unsigned int i, j;
  for(i=0; i < s.size(); ++i)
    {
      if (s[i] == '.')
        break;
      num1 *= 10;
      if (s[i]>'9' || s[i]<'0')
	Error("incorrect format of game result string");
      temp = s[i] - '0';
      num1 += temp;
    }

  for(j = s.size()-1; j>=i; --j)
    {
      num2 *= 0.1;
      if (j==i)
        break;
      if (s[j]>'9' || s[j]<'0')
	Error("incorrect format of game result string");

      temp = s[j] - '0';
      num2 += temp;
    }

  return num1+num2;
}

/**
 * Helper function to scan server output string to 
 * check for client disconnect (<newline><id>:<spaces>dead).
 *
 * @param  src  server output string to be scanned
 * @param  id   0 or 1 depending on which player is being checked for disconnect
 * @return      true <=> this player was disconnected  
 */
bool playerDisconnected(string src, int id)
{
  string s = "\n" + intToString(id) + ":";
  unsigned int pos = src.find(s);
  if (pos == string::npos)
    return false;
  s = src.substr(pos+3); 
  pos = 0;
  while(s[pos] == ' ') {
    ++pos;
    if (pos >= s.size())
      return false;
  }
  s = s.substr(pos);
  if (s.substr(0,4) == "dead")
    return true;
    
  return false;
}

/**
 * Displays the help message to standard out when -help is the command line argument.
 * Exits the program with code 10.
 */
void print_help()
{
  cout << "tmanager - an ORTS tournament manager.                       \n";
  cout << "It reads a list of players and a list of games played so far \n\
and schedules and autonomously runs the tournament.                     \n";
  cout << "\nTournament .cfg File Syntax:\n";
  cout << "# comment                                                    \n\
# general section                                                       \n\
<seed>                                                                  \n\
<game type (1-single,2-double,or 3-double)>                             \n\
<game time in minutes + 1 for timer watchdog>                           \n\
s|d # (single: one game per map or double: two games per map)           \n\
<games(single)/game-pairs(double) per player pair>                      \n\
<max number of jobs on each client computer>                            \n\
<max number of jobs on each server computer>                            \n\
<server output subdirectory>                                            \n\
<client output subdirectory>                                            \n\
<replay output subdirectory in the server directory>                    \n\
\n%\n# player section                                                   \n\
server <unix-id (.: none)> <work-directory> <command [with options] (server-start-line>    \n\
<name> <unix-id (.: none)> <work-directory> <command [with options] (client-start-line>    \n\
...                                                                     \n\
# applications are started like this:                                   \n\
# ssh [unixid@]machine \"cd server-workdir; server-start-line+options\" \n\
# ssh [unixid@]machine \"cd client-workdir; client-start-line+options\" \n\
%                                                                       \n\
# machine list                                                          \n\
<computer1>                                                             \n\
<computer2>                                                             \n\
...                                                                     \n\
#end of .cfg file syntax \n";
  cout << "\nSample .gam Game Result File:    \n\
<seed> <player1 name> <[C]onnect|[D]isconnect> <player2 name> \
<[C]onnect|[D]isconnect> <result string> <player1 name> <player1 score> \
<player2 name> <player2 score>                \n\n";
  exit(10);
}



/**
 * Displays all games in the queue. Shows game index, game seed,
 * player1 name, player2 name
 */
void displayGames()
{
  LOCK; // START CRITICAL SECTION

  cout << "All games:" << endl;
  cout << "Maximum allowable game time: " << games[0].maxTime << endl;
  cout << "ID:\t" << "SEED:\t\t" << "PLR1:\t";

  if (!single_player)
    cout << "PLR2:\t";

  cout << "RESULT:" << endl;

  FORALL (games, i) {
    cout << i->id << "\t" << i->seed << "\t";
    cout << players[i->player_list[0]].name;

    if (!single_player)
      cout << "\t" << players[i->player_list[1]].name;

    if (!(i->result[0].empty()))
      cout << "\t" << i->result[0]; 
    cout << endl;
  }
  
  UNLOCK; // END CRITICAL SECTION  
}

/**
 * Displays game statistics for all players.
 *
 */
void displayScores()
{
  LOCK; // START CRITICAL SECTION
  
  cout << "scores:" << endl << endl;

  cout << (boost::format("%|-10| %|-10| %|-10| %|-10| %|-10|\n\n")
	   % "name" 
	   % "score"
	   % "matches"
	   % "ratio"
	   % "errors"
	   ).str();

  FORALL (players, i) {

    cout << (boost::format("%|-10| %-9.1lf  %|-10| %|-10| %|-10|\n")
	     % i->name 
	     % i->score
	     % i->maxPoints
	     % ((i->maxPoints == 0) ? 0 : (double)i->score/i->maxPoints)
	     % i->errors
	     ).str();
  }
  
  UNLOCK; // END CRITICAL SECTION  
}


/**
 * Generates the game queue.
 * 
 * @param games_per_pair    the number of games per player (pair)
 * @param game seed         base map seed found by XORing all players' seeds
 * @param game_length       the expected length of game 
 * @param srvStartLine      startline of the server's executable (game type will be appended to it)
 * @param srvWorkDir        working directory of the server
 * @param srvUsrName        user name of the server
 * @param srvOutDir         output directory of the server
 * @param replayDir         replay directory
 * @param cliOutDir         output directory of the client
 * @return                  total number of games in games
 */
void buildGameList(int games_per_pair, int game_seed,
		   double game_length,
		   string srvStartLine, string srvWorkDir, string srvUsrName,
		   string srvOutDir, string replayDir, string cliOutDir)
{
  int gameCount = 0;
    
  LOCK; // START CRITICAL SECTION
  
  if (game_type == 1) {

    FORS (j, games_per_pair) {
      FORALL (players, i) {
	
	Game g;
	g.id = gameCount;  
	g.progress = queued; 
	g.saved = false;
	g.seed = game_seed + j;
	g.player_list[0] = i->index;
	g.maxTime = game_length;
	g.serverStartLine = srvStartLine;
	g.serverWorkDir = srvWorkDir;
	g.serverUserName = srvUsrName;
	g.serverOutDir = srvOutDir;
	g.replayDir = replayDir;
	g.clientOutDir = cliOutDir;

	games.push_back(g);
	++gameCount;
      }
    }

  } else {

    int matches_per_pair = games_per_pair;

    if (!single_game) //games per pair includes the matches that belong to the same game pair
      matches_per_pair /= 2;  //gets the number of games (not matches) per pair
    
    FORS (k, matches_per_pair) {
      
      FORT (i, players.size()) {
	for (int j=1; j < (int)(players.size()-i); ++j) {
	  Game g;
	  g.id = gameCount; 
	  g.progress = queued; 
	  g.saved = false;
	  g.seed = game_seed;
	  g.maxTime = game_length; 
	  g.serverStartLine = srvStartLine;
	  g.serverWorkDir = srvWorkDir;
	  g.serverUserName = srvUsrName;
	  g.serverOutDir = srvOutDir;
	  g.replayDir = replayDir;
	  g.clientOutDir = cliOutDir;
	  g.player_list[0] = i;
	  g.player_list[1] = j+i;
	  games.push_back(g);
	  ++gameCount;
	  
	  if (!single_game) { //if pair-game, generate the game pair immediately
	    g.id = gameCount; 
	    g.progress = queued; 
	    g.saved = false;
	    g.seed = game_seed;
	    g.maxTime = game_length; 
	    g.serverStartLine = srvStartLine;
	    g.serverWorkDir = srvWorkDir;
	    g.serverUserName = srvUsrName;
	    g.serverOutDir = srvOutDir;
	    g.replayDir = replayDir;
	    g.clientOutDir = cliOutDir;
	    g.player_list[0] = j+i;
	    g.player_list[1] = i;
	    games.push_back(g);
	    ++gameCount;
	  }	  
	  ++game_seed;
	}
      }
      
    }
  }
  
  UNLOCK; // END CRITICAL SECTION
}


/**
 * Makes accessible both the standard input and output of the shell 
 * process it creates. Two pipes are created: one provides standard
 * input for the shell commmand, and the other for passing back its 
 * standard output. The pipe file descriptors for the caller to use 
 * are pointed to by p_fd_in and p_fd_out.
 * 
 * @param  shell_cmd  the shell command to be executed  
 * @param  p_fd_in    the input pipe file descriptor for the caller to 
 *                    use (value of 0 indicates the shell command doesn't
 *                    use stdin)
 * @param  p_fd_out   the output pipe file descriptor for the caller to 
 *                    use (value of 0 indicates the shell command doesn't
 *                    use stdout) 
 * @return            the parent process id
 */      
pid_t popen2(const string &shell_cmd, int *p_fd_in, int *p_fd_out)
{
  // CREATING TWO PIPES:
  int fds_processInput[2];  //pipe for process input
  int fds_processOutput[2]; //pipe for process output
  
  if (pipe(fds_processInput) != 0) {
    // create process input pipe
    Error("pipe (process input) failed");
  }
  
  if (pipe(fds_processOutput) != 0) {
    //create process output pipe
    Error("pipe (process output) failed");
  }
  
  // FORKING A CHILD PROCESS:
  pid_t pid;
  if ((pid = fork()) < 0) {
    Error("fork failed");
  }
  
  // CONNECT THE CORRECT PIPE ENDS IN THE CHILD:

  if (pid == 0) {

    // child process

    // for process input pipe:
    close(fds_processInput[1]);    //close output
    dup2(fds_processInput[0], 0);  //close fd 0, fd 0 = fds_processInput[0]
    
    // for process output pipe:
    close(fds_processOutput[0]);   //close input
    dup2(fds_processOutput[1], 1); //close fd 1, fd 1 = fds_processOutput[1]
    
    execl("/bin/sh", "sh", "-c", shell_cmd.c_str(), NULL ); 
    Error("failed to run shell_cmd");

  } else {

    // parent process

    // for process input pipe:
    close(fds_processInput[0]);    //close input
      
    // for process output pipe:
    close(fds_processOutput[1]);   //close output

    if (p_fd_in == 0)
      close(fds_processInput[1]);
    else
      *p_fd_in = fds_processInput[1];
      
    if (p_fd_out == 0)
      close(fds_processOutput[0]);
    else
      *p_fd_out = fds_processOutput[0];

  }
  return pid; 
}


#if 0

//Test Case 2: ls example
cout << "\n\nls example:\n";
int fd_in1 =0;
int fd_out1;
pid_t pid1 = popen2("ls", &fd_in1, &fd_out1);

if (pid1 <= 0)
{
  cerr << "Unable to exec prog";
}

while(1)  //read process output
{
  char c1;
  if (read(fd_out1, &c1, 1) <= 0)
    break; //no data
  cout << c1; 
}

//Test Case 3: an invalid shell command
cout << "\n\ninvalid shell command example:\n";
int fd_in3 =0;
int fd_out3;
pid_t pid3 = popen2("dksjfhahf", &fd_in3, &fd_out3);

if (pid3 <= 0)
{
  cerr << "Unable to exec prog";
}

while(1)  //read process output
{
  char c3;
  if (read(fd_out3, &c3, 1) <= 0)
    break; //no data
  cout << c3; 
}


return 0;


#endif


/**
 * Reads the output of command in 's' and stores it in s
 *
 * @param s  shell command, will also hold the resulting output
 * @return   true <=> received signal to kill (errno set to EINTR)
 */

#if 1

// popen2 version (supports killing process)

bool system_output(string &s, int &cmdReturnVal)
{
  errno = 0;
  const string cmd = s;
  bool kill = false;
  bool isNum = true;
  int fd_out;

  // cout << "start system_output " << cmd << endl;
  
  pid_t pid = popen2(cmd, 0, &fd_out);
  
  if (pid <= 0)
    Error("unable to exec cmd " + cmd);
  
  // read process output

  ssize_t r;
  
  while (1) {

    char c;

    if ((r=read(fd_out, &c, 1)) <= 0) {
      cout << "r= " << r << " errno=" << errno << endl;
      break; // no data
    }
    s += c; 
    if (c > '9' || c < '0')
      isNum = false;
  }

  close(fd_out);

  if (isNum)
    cmdReturnVal = static_cast<int>(stringToNum(s));
  else
    cmdReturnVal = 0;
  
  if (errno != 0) {

    cout << "read errno: " << errno << endl;

    // kill process
    ::kill(pid, SIGKILL);

    kill = true;

  } else waitpid(pid, 0, 0);

  // cout << "finished system_output " << cmd << endl;

  return kill;
}


#else

bool system_output(string &s)
{
  errno = 0;
  const string cmd = s;
  bool kill = false;
  FILE *fp = popen(s.c_str(), "r");
  // fixme: what if this fails?
  if (errno != 0) { perror("POPEN ERRNO "); }
  
  s = "";
  char c;

  while (1) {

    c = fgetc(fp);

    if (errno == EINTR) {

      cout << "RECEIVED SIGNAL IN FGETC" << endl;

      // stopping doesn't work right now :-(
      // need access to popen fork pid
      // pclose just waits ...

      kill = true;
      break;
    }
    if (c == EOF) 
      break;
    s += c;
  }

  pclose(fp);
  return kill;
}

#endif


/**
 * Sets up the server and clients and spawns the corresponding threads.
 * Updates progress flags for the game & taskCount on the computers 
 *
 * @param arg         pointer to the game that is to be played 
 */
void *playGame(void *arg)
{
  LOCK; // START CRITICAL SECTION

  Game *game = static_cast<Game*>(arg);
  if (game->magic != Game::MAGIC) Error("not a game 1");
      
  const double maxTime = game->maxTime;
  string serverOutDir = game->serverOutDir;
  game->start_time = get_time();
  timeval start = game->start_time;

  game->thread_status[0] = pending; // was done in play* functions, wrong!
  game->thread_status[1] = pending; // because status still was done when
  game->thread_status[2] = pending; // the loop below checks for game end! 
  
  // create the server and client threads

  if (pthread_create(&(game->thread_srv), &attr, playServer, game)) {
    Error("cannot create another thread");
  }

  UNLOCK;
  sleep(SERVER_START_WAIT);
  LOCK;
  
  GameArg gameArg1, gameArg2; // gameArg2 was defined in local block, oops!

  gameArg1.playerIndex = 0;
  gameArg1.game = game;

  if (pthread_create(&(game->threads_ply[0]), &attr, playClient, &gameArg1))
    Error("cannot create another thread");
  
  if (!single_player) {

    gameArg2.playerIndex = 1;
    gameArg2.game = game;

    UNLOCK;
    sleep(CLIENT_START_WAIT);
    LOCK;
    
    if (pthread_create(&(game->threads_ply[1]), &attr, playClient, &gameArg2))
      Error("cannot create another thread");
  }

  UNLOCK; // END CRITICAL SECTION

  // wait for the server & client threads -- in the meantime, make
  // sure they're done & in the right order set up alarm:

  struct sigaction actions;
  memset(&actions, 0, sizeof(actions));
  sigemptyset(&actions.sa_mask);
  actions.sa_flags = 0;
  actions.sa_handler = sighand;
  if (sigaction(SIGALRM, &actions, NULL))
    Error("cannot set up signal alarm");
  
  bool killSrv = false;
  double begin = to_double(start);
  bool record1 = true;
  bool record2 = true;
  while (1) {	

    LOCK; // START CRITICAL SECTION

    if (game->thread_status[0] == done &&
	game->thread_status[1] == done &&
	(single_player || game->thread_status[2] == done)) {
      UNLOCK; // END CRITICAL SECTION
      break;
    }

    double now = to_double(get_time());
      
    if (now-begin > 60*maxTime) {

      cout << "TIMEOUT" << endl;
      
      if (game->thread_status[0] != done) {
	killSrv = true;
	if (pthread_kill(game->thread_srv, SIGALRM))
	  Error("cannot send kill signal to server thread");
      }

      if (game->thread_status[1] != done) {
	if (pthread_kill(game->threads_ply[0], SIGALRM))
	  Error("cannot send kill signal to player 1 thread");
      }

      if (!single_player && game->thread_status[2] != done) {
	if (pthread_kill(game->threads_ply[1], SIGALRM))
	  Error("cannot send kill signal to player 2 thread");
      }
    } else if (record1 &&
	       game->thread_status[0] != done && 
	       game->thread_status[1] == done &&
	       (now - to_double(game->end_times[1])) > 2) {
      //report early client disconnect to its log
      ostringstream oss;  
      oss << "./" << game->clientOutDir << "/" << players[game->player_list[0]].name;
      oss << "/" << game->seed << "_" << players[game->player_list[0]].name;

      if (!single_player)
	oss << "_" << players[game->player_list[1]].name;

      oss << "-" << players[game->player_list[0]].name << "-" << game->tries << ".plr";
      ofstream out((oss.str()).c_str(), ios::app);
       
      if (!out.is_open()) Error("unable to open the disk file for writing " + oss.str());
       
      out << "Client was disconnected at least two seconds earlier than the server!\n";
      record1 = false;
      out.close();

    } else if (record2 &&
	       !single_player &&
	       game->thread_status[0] != done && 
	       game->thread_status[2] == done &&
	       (now - to_double(game->end_times[2])) > 2) {
      //report early client disconnect to its log
      ostringstream oss;  
      oss << "./" << game->clientOutDir << "/" << players[game->player_list[1]].name;
      oss << "/" << game->seed << "_" << players[game->player_list[0]].name;
      oss << "_" << players[game->player_list[1]].name;
      oss << "-" << players[game->player_list[1]].name << "-" << game->tries << ".plr";
      ofstream out((oss.str()).c_str(), ios::app);
      
      if (!out.is_open()) Error("unable to open the disk file for writing " + oss.str());
      
      out << "Client was disconnected at least two seconds earlier than the server!\n";
      record2 = false;
      out.close();

    }

    //      cout << game->id << " isn't done yet!" << endl;
    UNLOCK; // END CRITICAL SECTION

    sleep(1);      
  }

  LOCK; // START CRITICAL SECTION

  cout << "game " << game->id << " done" << endl;
  
  game->end_times[0] = get_time();
  game->server->taskCount--;

  if (game->server->taskCount <= 0)
    game->server->isServer = false; // computer no longer dedicated server

  game->clients[0]->taskCount--;

  if (!single_player)
    game->clients[1]->taskCount--;

  // generate game data to be saved
  
  ostringstream out;
  
  out << "Server: " << game->server->name << endl;

  tm *timeinfo = localtime(&game->start_time.tv_sec);
  out << "start time:\t" << asctime(timeinfo);
  timeinfo = localtime(&game->end_times[0].tv_sec);
  out << "end time:\t" << asctime(timeinfo) << endl; 

  out << "Output:\n" << game->result[0] << endl;

  if (killSrv) {
    ostringstream os;
    os << "*** timeout: game " << game->id << " thread forced to terminate" << endl;
    out << os.str();
    cout << os.str();
  }

  // @ at beginning of line marks game result
  size_t pos = (game->result[0]).find("\n@");
  
  if (pos == string::npos) {

    ostringstream os;
    os << "*** no valid result in server result string - game "
       << game->id
       << " aborted"
       << endl;

    out << os.str();
    cout << os.str();
    
  } else {

    ostringstream os;
    os << "game " << game->id << " ended successfully" << endl;
    out << os.str();
    cout << os.str();
  }

  // save server output
  {
    ostringstream oss;

    oss << game->serverOutDir << "/" << game->seed
	<< "_" << players[game->player_list[0]].name;
    
    if (!single_player)
      oss << "_" << players[game->player_list[1]].name;

    oss << "-" << game->tries <<  ".srv";
    
    // save output to file
    ofstream fout((oss.str()).c_str());
    if (!fout.is_open())
      Error("unable to open the disk file for writing " + oss.str());
    fout << out.str();
  }

  if (opt_v && PRINT_SRV_OUTPUT) 
    cout << out.str();

  // store result
  if (pos != string::npos) {
    string r = (game->result[0]).substr(pos+1);
    size_t pos = r.find("\n");    
    game->result[0] = r.substr(0, pos);
    game->progress = done;
  } else
    game->progress = queued; // retry
  
  UNLOCK; // END CRITICAL SECTION

  pthread_exit(NULL);
}

/**
 * Makes the ssh call to the server and stores the result of this call 
 * as the server's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 * 
 * @param arg  pointer to the game that is being played
 */
void *playServer(void *arg)
{   
  LOCK; // START CRITICAL SECTION

  Game *game = static_cast<Game*>(arg);
  if (game->magic != Game::MAGIC) Error("not a game 2");
  Computer *server   = game->server;
  int player1        = game->player_list[0];
  int player2        = game->player_list[1];  
  string startline   = game->serverStartLine;
  string workDir     = game->serverWorkDir;
  string username    = game->serverUserName;

  string s = "ssh " + SSH_PAR + " ";
  if (username != ".") s += username +"@";
  
  s += server->name + " \"cd " + workDir + "/; " + startline
    + " -ids " + players[player1].name;

  if (!single_player)
    s += "," + players[player2].name;

  s += " -seed " + intToString(game->seed);
  s += " -port " + intToString(game->port);
  s += " -savereplay " + game->replayDir + "/" + intToString(game->seed) + 
    + "_" + players[player1].name;

  if (!single_player)
    s += "_" + players[player2].name;
  
  s += ".rpl";
  s += " 2>&1\"";

  UNLOCK; // END CRITICAL SECTION

  if (PRINT_CMDS)  
    cout << "server command: " << s << endl;
  int retVal = 0;
  system_output(s, retVal);  //ignore the return value here
  
  // lock games to update games info:
  LOCK; // START CRITICAL SECTION

  game->end_times[0] = get_time();
  game->result[0] = s;  
  game->thread_status[0] = done;

  if (retVal != 0) {           //computer can be flagged as nonworking and reported to stdout
    server->isUp = false;
    cout << "server ssh command returned value " << retVal;
    if (retVal == 1)
      cout << ": connection problems";
    else if (retVal == 127)
      cout << ": remote command cannot be executed";
    cout << ", computer " << server->name << " flagged as nonworking." << endl;
  }

  UNLOCK; // END CRITICAL SECTION

  if (PRINT_DONE)
    cout << "SERVER DONE" << endl;
   
  pthread_exit(NULL);
}

/**
 * Makes the ssh call to a client and stores the result of this call 
 * as this client's result for this particular game.
 * Writes all received results to files.
 *
 * @param arg  pointer to the game that is being played
 */
void *playClient(void *arg)
{
  LOCK; // START CRITICAL SECTION

  GameArg *gameArg = static_cast<GameArg*>(arg);
  Game *game = gameArg->game;

  if (game->magic != Game::MAGIC) Error("not a game 2");
  
  int playerIndex = gameArg->playerIndex;

  // get game information:
  Player &player = players[game->player_list[playerIndex]];
  Computer *client = game->clients[playerIndex];

  string s = "ssh " + SSH_PAR + " ";
  if (player.username != ".")
    s += player.username +"@";

  s += client->name + " \"cd ";
  s += player.workDir + "/; ";
  s += player.startLine
    + " -host " + game->server->name
    + " -port " + intToString(game->port) 
    + " -id " + player.name;
  s += " 2>&1\"";

  UNLOCK; // END CRITICAL SECTION

  if (PRINT_CMDS)
    cout << "client command: " << s << endl; 
  int retVal = 0;
  bool killed = system_output(s, retVal);
  
  // lock games to update games info:

  LOCK; // START CRITICAL SECTION

  timeval end = game->end_times[playerIndex+1] = get_time();
  game->result[playerIndex+1] = s;

  if (retVal != 0) {           //computer can be flagged as nonworking and reported to stdout
    client->isUp = false;
    cout << "client ssh command returned value " << retVal;
    if (retVal == 1)
      cout << ": connection problems";
    else if (retVal == 127)
      cout << ": remote command cannot be executed";
    cout << ", computer " << client->name << " flagged as nonworking." << endl;
  }

  ostringstream oss;  
  oss << "./" << game->clientOutDir << "/" << player.name;

  system(("mkdir -p " + oss.str()).c_str());

  oss << "/" << game->seed << "_" << players[game->player_list[0]].name;

  if (!single_player)
    oss << "_" << players[game->player_list[1]].name;

  oss << "-" << player.name << "-" << game->tries << ".plr";
  
  ofstream out((oss.str()).c_str());

  if (!out.is_open()) Error("unable to open the disk file for writing " + oss.str());

  out << "__________________________________________________\n";
  out << "Player: " << player.name << "\n(username: " << player.username << ")\n";
  out << "(computer: " << client->name << ")\n";
  tm *timeinfo = localtime(&game->start_time.tv_sec);
  out << "start time:\t" << asctime(timeinfo);
  timeinfo = localtime(&end.tv_sec);
  out << "end time:\t" << asctime(timeinfo) << endl; 

  if (killed)
    out << "***TIME OUT ERROR: thread aborted.\n";

  if (!s.empty())
    out << "Output:\n" << s << endl;

  out.close();

  game->thread_status[playerIndex+1] = done;

  UNLOCK; // END CRITICAL SECTION

  if (PRINT_DONE)
    cout << "CLIENT DONE" << endl;
  
  pthread_exit(NULL);
}


/**
 * Updates the contents of the .gam file with inorder game results 
 * as they complete.
 *
 * @param filename        the name of the configuration file with .gam suffix appended
 */
void updateGameFile(const string &filename)
{
  bool overWrite = false;
  
  LOCK; // START CRITICAL SECTION

  FORALL (games, i) {
    if ((i->progress == done) && !(i->saved)) {
      overWrite = true;
      break;
    }
  }

  if (!overWrite) {
    UNLOCK;
    return;
  }
  
  ofstream out(filename.c_str());

  if (!out.is_open())
    Error("unable to open the .gam game result file for writing " + filename);
  
  int new_n = 0;

  FORALL (games, i) {

    if (i->progress == done) {

      ++new_n;
      out << i->seed << " " << players[i->player_list[0]].name;
 
      // print 1st player information     
      if (playerDisconnected(i->result[0], 0))
	out << " D";
      else
	out << " C";

      if (single_player) {

	// game type 1: single player, single game matches
	if (!single_game) Error("single player but not single games!");

	out << " " << i->result[0];
	float sc1, dummy;
	if (i->result[0] == ERROR) {
	  sc1 = 0; 
	} else if (score_func(i->result[0], "", sc1, dummy))
	  Error("error in result string -- cannot get score");
	
	out << " " << players[i->player_list[0]].name << " " << sc1 << endl;
	
	if (!(i->saved)) {
	  // only update scores once
	  i->saved = true;

	  players[i->player_list[0]].score += sc1;
	  players[i->player_list[0]].maxPoints++;
	}

      } else {

	// two-player

	// print 2nd player information
	
	out << " " << players[i->player_list[1]].name;

	if (playerDisconnected(i->result[0], 1))
	  out << " D";
	else
	  out << " C";
	
	if (single_game) {

	  // single game
	  
	  float sc1, sc2;
	  
	  if (score_func(i->result[0], "", sc1, sc2))
	    Error("error in result string -- cannot get score");
	  
	  out << " " << players[i->player_list[0]].name << " " << sc1
	      << " " << players[i->player_list[1]].name << " " << sc2
	      << endl;

	  if (!(i->saved)) {

	    // only update scores once
	    i->saved = true;
	  
	    players[i->player_list[0]].score += sc1;
	    players[i->player_list[1]].score += sc2;
	    // assumes highest points possible in a game is 1
	    players[i->player_list[0]].maxPoints++;
	    players[i->player_list[1]].maxPoints++;
	  }

	} else {

	  // pair-game

	  out << " " << i->result[0];
	  
	  if (((i->id) & 1) && (i-1)->saved) { // if your id is odd, check the one before it for saved

	    // append scores for game pair
	    
	    float sc1, sc2;
	    
	    if ((i->result[0] == ERROR) || ((i-1)->result[0] == ERROR)) {
	      
	      sc1 = sc2 = 0;
	      
	    } else {

	      if (score_func(i->result[0], (i-1)->result[0], sc1, sc2))
		Error("error in result string -- cannot get score");
	    }

	    out << " " << players[i->player_list[0]].name << " " << sc1;
	    out << " " << players[i->player_list[1]].name << " " << sc2 << endl;

	    if (!(i->saved)) {
	      // only update scores once
	      i->saved = true;
	      
	      players[i->player_list[0]].score += sc1;
	      players[i->player_list[1]].score += sc2;
	      // assumes highest points possible in a game is 1
	      players[i->player_list[0]].maxPoints++;
	      players[i->player_list[1]].maxPoints++;
	    }

	  } else {

	    out << endl;

	    if (!((i->id) & 1)) 
	      i->saved = true;
	  }
	}
      }
    } 
  }  

  UNLOCK; // END CRITICAL SECTION

  cout << "wrote " << new_n << " game(s)" << endl;

}


/**
 * Displays the summary of game results in matrix form to stdout
 *
 * @param filename        the base name of the file 
 */
void displayMatrix(const string &filename)
{
  LOCK;

  bool valid = false;

  FORALL (games, g) {
    if (g->progress == done) {
      valid = true;
      break;
    }
  }

  if (!valid) {
    cout << "No games have been played yet (no " << filename << ".gam file provided)\n";
    UNLOCK;
    return;
  }

  int num = (int)players.size();
  int rank[num];
  int i, j, k;
  float sc1, sc2;

  FOR (i, num)
    rank[i] = i;
    
  FOR (i, num-1) {
    for(j = i+1; j < num; j++) {
      if (players[rank[j]].score > players[rank[i]].score) {
	k = rank[i];
	rank[i] = rank[j];
	rank[j] = k;
      }
    }
  }

  cout << "summary of results:" << endl << endl;
  cout << "players    ratio (total points/total games played)" << endl;
  cout << "--------------------------------------------------" << endl << endl;
  FOR (i, num)
    cout << boost::format("%-10s %f")
    % players[rank[i]].name
    % (players[rank[i]].maxPoints == 0 ? 0 : (double)players[rank[i]].score/players[rank[i]].maxPoints)
	 << endl;

  valid = true;

  FORALL (games, g)
    if (g->result[0] == ERROR)
      valid = false;

  if (!valid) {
    cout << "At least one game was aborted because more than " << MAX_TRIES;
    cout << " attempts were made to play this game.\n";
    cout << "Consider " << filename << ".gam for detailed results.\n";
    UNLOCK;
    return;
  }

  // now rank contains the indices of the players in decreasing order of score
  // generate matrix before displaying it

  if (!single_player) {

    Summary matrix[num][num];

    FOR (i, num) {
      FOR (j, num) {

	matrix[i][j].player_list[0] = rank[i];
	matrix[i][j].player_list[1] = rank[j];

	if (i != j) {      

	  FORALL (games, g) { 

	    if ((g->player_list[0] == rank[i] || g->player_list[0] == rank[j]) &&
		(g->player_list[1] == rank[i] || g->player_list[1] == rank[j])) {

	      // player names match

	      if (single_game) {

		// single game
		
		if (score_func(g->result[0], "", sc1, sc2))
		  Error("error in result string -- cannot get score");

	      } else {

		// pair-games
		
		if (!((g->id) & 1)) continue; // even id : skip
		
		if (score_func(g->result[0], (g-1)->result[0], sc1, sc2))
		  Error("error in result string -- cannot get score");
	      }
	      
	      // update scores
	      
	      if (g->player_list[0] == rank[i]) {
		matrix[i][j].scores[0] += sc1;
		matrix[i][j].scores[1] += sc2;
	      } else {
		matrix[i][j].scores[0] += sc2;
		matrix[i][j].scores[1] += sc1;
	      }
	      matrix[i][j].matches++;
	    }
	  }
	}
      }
    }

    // now display the matrix

    cout << "\nMatrix Summary of Results: \n\n";
    cout << "Name:\tIndex:";

    FOR(i, num)
      cout << "\t " << i+1;

    cout << "\t Total Score/Total Games" << endl;

    FOR (i, num) {
      cout << players[rank[i]].name << "\t" << i+1; 

      FOR (k, 3) {
	
	if (k != 0)
	  cout << "\t";

	FOR (j, num) {

	  cout << "\t|";

	  if (k==0) {

	    if (i!=j)
	      cout << matrix[i][j].scores[0];
	    else
	      cout << "--";
	    
	    if (j == num-1) {
	      cout << "\t|\t" << players[rank[i]].score << "/" << players[rank[i]].maxPoints
		   << " = "
		   << (players[rank[i]].maxPoints == 0 ? 0 : (double)players[rank[i]].score/players[rank[i]].maxPoints)
		   << endl;
	    }
	  } else if (k == 1) {
	    if (i != j)
	      cout << matrix[i][j].scores[1];
	    else
	      cout << "--";

	    if (j == num-1)
	      cout << "\t|" << endl;

	  } else {
	    if (i != j)
	      cout << matrix[i][j].matches;
	    else
	      cout << "--";

	    if (j == num-1)
	      cout << "\t|" << endl;
	  }
	}
      }
    }

    cout << "\n_____________________________________________________________________\n";
    cout << "Matrix Legend:" << endl;
    cout << "entry (i,j) ->|player i's total score for all games between i and j|\n";
    cout << "              |player j's total score for all games between i and j|\n";
    cout << "              |total number of games between i and j               |\n\n";
  }
  
  UNLOCK;
}

/**
 * Sets up the alarm for the signal handler
 *
 * @param signo
 */
void sighand(int /* signo */)
{ 
  // pthread_t self = pthread_self();  //WHAT DOES THIS DO???
  // cout << "Signal received by thread " << self << endl;
  return;
}


/**
 * Returns the next queued game.
 *
 * @return the index of the next game in games that is queued (-1 <=> no queued game)
 */
int getNextQueuedGame()
{
  LOCK; // START CRITICAL SECTION

  int i;
  FOR (i, (int)games.size()) { 
    if (games[i].progress == queued) { 
      if (games[i].tries < MAX_TRIES)
	break;
      else{//exceeded maximum number of game attempts
	games[i].result[0] = ERROR;
	games[i].progress = done;
      }
    }
  }

  if (i >= (int)games.size()) i = -1;

  UNLOCK; // END CRITICAL SECTION
      
  return i;
}

/**
 * Checks for available computers.
 * If there are available ones, assign them to new games and mark as unavailable.
 * 
 * @param max_client_jobs the max number of jobs allowed per client machine
 * @param max_server_jobs  the max number of jobs allowed per server machine
 *                         if 0: no distinction between server/client machines
 * @return                 Return true if we are done with all games
 */ 
bool checkComputers(int max_client_jobs, int max_server_jobs)
{
  int compsNeeded = 2;

  if (!single_player)
    compsNeeded = 3;

  cout << "." << flush;
  
  FORS (k, MAX_START) {
  
    int nextGame = getNextQueuedGame();

    if (nextGame == -1) {
    
      LOCK; // START CRITICAL SECTION

      bool flag = true;
    
      FORALL (games, i) {
	if (i->progress == pending) {
	  cout << "No game left, waiting for pending game: "
	       << (i-games.begin()) << endl;
	  flag = false;
	  break;
	}
      }

      UNLOCK; // END CRITICAL SECTION
      return flag;
    }

    Vector<Computer*> free_comps;
  
    LOCK; // START CRITICAL SECTION

    Vector<int> task_ids;   // task id on the assigned computers
    Vector<int> compIndex;  // randomized indices into computers array
    
    FORT (j, computers.size())
      compIndex.push_back(j);
    
    random_shuffle(compIndex.begin(), compIndex.end());

    // assign server job
    
    if (max_server_jobs == 0) {

      // servers and clients can run on the same machine

      FORT (i, computers.size()) {
	if (computers[compIndex[i]].isUp && 
	    computers[compIndex[i]].taskCount < max_client_jobs) {
	  //CHECK_IS_UP bit
	  task_ids.push_back(computers[compIndex[i]].taskCount);
	  ++computers[compIndex[i]].taskCount;
	  free_comps.push_back(&computers[compIndex[i]]);
	  break;
	}
      }
      
    } else {

      // servers and clients run on different machines
      
      // search for free slot on server machines

      FORT (i, computers.size()) {
	if (computers[compIndex[i]].isUp && 
	    computers[compIndex[i]].isServer &&
	    computers[compIndex[i]].taskCount < max_server_jobs) {
	  //CHECK_IS_UP bit
	  task_ids.push_back(computers[compIndex[i]].taskCount);
	  ++computers[compIndex[i]].taskCount;
	  free_comps.push_back(&computers[compIndex[i]]);
	  break;
	}
      }
        
      // if can't add another server job to any existing server, find an empty machine

      if (free_comps.size() < 1) {
	FORT (i, computers.size()) {
	  if (computers[compIndex[i]].isUp && computers[compIndex[i]].taskCount==0) {
	    //CHECK_IS_UP bit
	    computers[compIndex[i]].isServer = true;
	    task_ids.push_back(computers[compIndex[i]].taskCount);
	    ++computers[compIndex[i]].taskCount;
	    free_comps.push_back(&computers[compIndex[i]]);
	    break;
	  }
	}
      }
    }
    
    // only assign client jobs if a server was found

    if (free_comps.size() > 0) {

      FORT (i, computers.size()) {

	while ((int)free_comps.size() < compsNeeded && 
	       computers[compIndex[i]].isUp && 
	       computers[compIndex[i]].taskCount < max_client_jobs && 
	       !computers[compIndex[i]].isServer) { 

#if CHECK_IS_UP
	  
	  // check that this computer is up before assigning it:
	  // only works on a ugrad machine
	  
	  s = "ruptime | grep " + computers[compIndex[i]].name+ " | grep up"; 

	  UNLOCK; // unlock during shell command execution
	  system_output(s);
	  LOCK;
  
	  if (s.empty()) { 
	    computers[compIndex[i]].isUp = false;
	    continue;  
	  }   

#endif

	  task_ids.push_back(computers[compIndex[i]].taskCount);
	  ++computers[compIndex[i]].taskCount;
	  free_comps.push_back(&computers[compIndex[i]]);
	}

	if ((int)free_comps.size() >= compsNeeded) break;
      }

      if ((int)free_comps.size() >= compsNeeded) {

	// all machines assigned for this game
	
	assert(free_comps.size() >= 2);

	cout << "start game " << nextGame
	     << " try: " << games[nextGame].tries
	     << " server: "
	     << "(" << (free_comps[0]-&computers[0])
	     << "," << task_ids[0]
	     << ") ";
      
	games[nextGame].server = free_comps[0];
	games[nextGame].server->isServer = true;
    
	cout << " client 1: "
	     << "(" << (free_comps[1]-&computers[0])
	     << "," << task_ids[1]
	     << ")";

	games[nextGame].clients[0] = free_comps[1];
	games[nextGame].clients[0]->isServer = false;
    
	if (free_comps.size() >= 3) {

	  cout << " client 2: "
	       << "(" << (free_comps[2]-&computers[0])
	       << "," << task_ids[2]	    
	       << ")";
	
	  games[nextGame].clients[1] = free_comps[2];
	  games[nextGame].clients[1]->isServer = false;
	}

	cout << endl;
      
	games[nextGame].port = games[nextGame].server->portNum; // pick unique server port
	games[nextGame].server->portNum++;
	games[nextGame].tries++;
	games[nextGame].progress = pending; // was in playGame ...
      
	if (pthread_create(&games[nextGame].thread_game,&attr, playGame, &games[nextGame])) {
	  perror("FOO");
	  Error("Cannot create another thread for game");
	}

      } else {

	// couldn't satisfy game requirements -> undo tasknum ++
	FORALL (free_comps, i) {
	  --(*i)->taskCount;
	}

      }
    }
    UNLOCK; // END CRITICAL SECTION
  }
  return false;
} 

// read and process start line
string getStartLine(istream &in)
{
  string s;
  getline(in, s);
  if (s.empty() || s[0] == '\n') return "";
  // erase leading space
  while (!s.empty() && s[0] == ' ') s.erase(0, 1);
  // erase trailing lf
  size_t n = s.size();
  if (n > 0 && s[n-1] == '\n') s.erase(n-1);
  return s;
}

/**
 * Main function
 */
int main(int argc, char **argv)
{
  int argi;
  string s;
  char c;
  string filename;          // append .cfg for config file, .gam for game result file

  int playerCount = 0;      // total number of players
  int compCount = 0;        // total number of computers
  int matches_per_pair = 0; // how many matches between two players
  int games_per_pair = 0;   // how many games played between each pair of players
  int max_client_jobs = 0;  // how many concurrent client jobs on each computer
  int max_server_jobs = 0;  // how many concurrent server jobs on each computer
  int game_seed = 0;        // base map seed found by XORing all players' seeds
  double game_length = 0;   // max length of a game in minutes       
  string serverStartLine;   // startline of server executable (game type is appended)
  string clientStartLine;   // startline of client executable (game type is appended)
  string serverWorkDir;     // working directory of server
  string serverUsrName;     // username of server  
  string serverOutDir;      // in what subdirectory server output is stored
  string replayDir;         // replays are stored here
  string clientOutDir;      // in what subdirectory client output is stored


#if 0
  // score tests
  score4_test(); exit(0);
  score1_test();  
  score2_test();
  score3_test();  
  exit(0);
#endif  
  
  if (argc < 2) 
    Error("usage");

  argi = 1;   // start with the first command line argument
 
  FOREVER {

    if (argi >= argc) Error("usage");

    string opt = argv[argi];
    
    if (opt == "-h") {
      print_help();
    } else if (opt == "-c") {
      opt_c = true;
      ++argi;
      continue;
    } else if (opt == "-v") {
      opt_v = true;
      ++argi;
      continue;
    } else if (opt == "-g") {
      opt_g = true;
      ++argi;
      continue;
    } else if (opt == "-s") {
      opt_s = true;
      ++argi;
      continue;
    }
    break;
  }

  // if generate computers option, generate the list, append it to the file & quit

  if (opt_c) {  

    // ASSUMED TMANAGER RUNS ON ug00 AND DON'T WANT TO ASSIGN ANYTHING ELSE TO IT!!!
    string file = argv[argi];
    s = "ruptime | grep ^ug | grep -v ug00 | grep up | cut -f1 -d' ' -s | cat >> " + file;
    system(s.c_str());

    ofstream out(file.c_str(), ios::app);
    if (!out.is_open())
      Error("unable to open the tournament file for writing generated computer list " + file);
    
    out << "\n%\n\n";
    out.close();

    cout << "\ngenerated (ug) computer list, appended to configuration file\n" << endl; 

    return 0;
  }   
 
  // parsing the input configuration file
  filename = argv[argi];
  ifstream in((filename+".cfg").c_str());

  if (!in.is_open())
    Error("unable to open the tournament file for reading " + filename+ ".cfg");

  if (skipComments(in))
    Error("end of file?");

  // general parameters

  if (!(in >> game_seed))
    Error("improper configuration file format -- game seed");
  if (skipComments(in))
    Error("end of file?");

  if (!(in >> game_type))
    Error("improper configuration file format -- game type?");
  if (game_type < 1 || game_type > 4)
    Error("invalid game type (1,2,3 or 4)");
  if (skipComments(in))
    Error("end of file?");

  if (!(in >> game_length))
    Error("improper configuration file format -- max game length(in minutes)?");
  if (game_length < 1 || game_length > 1000)
    Error("invalid game length >= 1 <= 1000");
  if (skipComments(in))
    Error("end of file?");

  if (!(in >> matches_per_pair))
    Error("improper configuration file format -- matches per pair?");
  if (matches_per_pair < 1 || matches_per_pair > 1000)
    Error("invalid matches per pair value >= 1 <= 1000");
  if (skipComments(in))
    Error("end of file?");

  if (!(in >> max_client_jobs))
    Error("improper configuration file format -- max jobs?");
  if (max_client_jobs < 1)
    Error("invalid max jobs value >= 1");
  if (skipComments(in))
    Error("end of file?");

  if (!(in >> max_server_jobs))
    Error("improper configuration file format -- max server jobs?");
  if (max_server_jobs < 0)
    Error("invalid max server jobs value >= 0");
  if (skipComments(in))
    Error("end of file?");

  // server output directory
  
  if (!(in >> serverOutDir))
    Error("improper configuration file format -- server output dir");
  if (skipComments(in))
    Error("end of file?");

  {
    ostringstream oss;  
    oss << "./" << serverOutDir;
    system(("mkdir -p " + oss.str()).c_str());
  }
  
  // client output directory
  
  if (!(in >> clientOutDir))
    Error("improper configuration file format -- client output dir");
  if (skipComments(in))
    Error("end of file?");

  {
    ostringstream oss;  
    oss << "./" << clientOutDir;
    system(("mkdir -p " + oss.str()).c_str());
  }
  
  // replay directory

  if (!(in >> replayDir))
    Error("improper configuration file format -- replay dir");
  if (skipComments(in))
    Error("end of file?");
  
  // players and server/client parameters

  getline(in, s);
  if (s != "%")
    Error("improper configuration file format -- % delimiter before server/client params?");

  if (skipComments(in)) Error("end of file?");

  // server
  
  if (!(in >> s))
    Error("improper configuration file format -- server parameters?");

  if (s != "server")
    Error("improper configuration file format -- server parameters?");

  if (!(in >> serverUsrName))
    Error("improper configuration file format -- server user name?");

  if (!(in >> serverWorkDir))
    Error("improper configuration file format -- server working directory?");

  serverStartLine = getStartLine(in);
  if (serverStartLine.empty()) Error("invalid server startline");
  
  if (skipComments(in)) Error("end of file?");

  FOREVER {

    if ((c = in.peek()) == EOF) Error("end of file?");
    in >> s;
    if (!s.empty()) { 
      if (s[0] == '%') break;
      if (s[0] != '#') {

	Player p;

	p.index = playerCount;
	p.name = s;

	if (!(in >> p.username)) Error("player username?");
	if (!(in >> p.workDir)) Error("client working directory?");

	p.startLine = getStartLine(in);
	if (p.startLine.empty()) Error("invalid " + p.name + " startline");
	
	players.push_back(p);
	playerCount++;
	if (skipComments(in)) Error("end of file?");
      }
    }
  }


  // set single_player, single_game dependent on game type

  switch (game_type) {
  case 1: single_player = true;  single_game = true;  score_func = score1; break;
  case 2: single_player = false; single_game = false; score_func = score2; break;
  case 3: single_player = false; single_game = false; score_func = score3; break;
  case 4: single_player = false; single_game = true;  score_func = score4; break;    
  }

  assert(!single_player || single_game);
  
  games_per_pair = matches_per_pair;

  if (!single_game)
    games_per_pair *= 2;

  // print out all the read-in player information  

  cout << "\nsetup:\n\n";
  cout << "seed:            " << game_seed << endl
       << "game-type:       " << game_type << endl
       << "game-time:       " << game_length << endl
       << "player-num:      " << (single_player ? 1 : 2) << endl
       << "games/match:     " << (single_game ? 1 : 2) << endl
       << "matches:         " << matches_per_pair << endl
       << "max-client-jobs: " << max_client_jobs << endl
       << "max-server-jobs: " << max_server_jobs << endl
       << "server-out:      " << serverOutDir << endl
       << "client-out:      " << clientOutDir << endl
       << "replays:         " << replayDir << endl;
  
  cout << endl << "server/players:" << endl << endl;

  cout << "name\tunix-id\twdir\tcommand" << endl << endl;

  cout << "server" << "\t"
       << serverUsrName << "\t"
       << serverWorkDir << "\t"
       << serverStartLine
       << endl;
  
  FORALL (players, i) {
    cout << i->name << "\t"
	 << i->username << "\t"
	 << i->workDir << "\t"
	 << i->startLine
	 << endl;
  }
  
  cout << endl;
    
  // get the list of available computers in the lab:  
  if (skipComments(in)) Error("end of file -- no computer list?");

  FOREVER {
    if (skipComments(in)) break;
    in >> s;
    if (!s.empty()) { 
      if (s[0] != '#') {
	
	Computer comp;
	comp.index = compCount;
	comp.name = s;
	comp.taskCount = 0; 
	comp.isUp = true;   // should check it is up before you try to use it

	computers.push_back(comp);
	++compCount;
      }
    }
  }
  in.close();
      
  // generate the game vector

  buildGameList(games_per_pair, game_seed, game_length, serverStartLine, serverWorkDir, serverUsrName,
		serverOutDir, replayDir, clientOutDir);

  // read game result lines
  ifstream in_gam((filename + ".gam").c_str());

  if (in_gam.is_open()) {
    if (!skipComments(in_gam)) { // have game results

      cout << "read result lines ...";
    
      string s1, s2;
      size_t pos1, pos2;
      int games_found = 0;
      bool found = false;
      FOREVER {

	if ((c=in_gam.peek()) == EOF) break;
      
	if (!(in_gam >> game_seed))
	  Error("provided result section -- seed?");

	if (!(in_gam >> s1))
	  Error("provided result section -- player1 name?");
	
	if (!(in_gam >> c) || ((c!= 'C') && (c!='D')))
	  Error("provided result section -- need connect (C) or disconnect (D) flag after player1's name");
	
	if (!single_player && !(in_gam >> s2))
	  Error("provided result section -- player2 name?");
	
	if (!single_player && (!(in_gam >> c) || ((c!= 'C') && (c!='D'))))
	  Error("provided result section -- need connect (C) or disconnect (D) flag after player2's name");
	
	// search for this game in the games list

	found = false;

	FORALL_B (games, i) {
	
	  if ((i->seed == game_seed) &&
	      (players[i->player_list[0]].name == s1) &&
	      (single_player || (players[i->player_list[1]].name == s2))) {
	 
	    getline(in_gam, s);
	    if (s.empty())
	      Error("provided result section -- result string?");

	    unsigned int j = 0;

	    while(j<s.size() && s[j] == ' ')
	      j++;

	    if (j >= s.size())
	      Error("provided result section -- result string?");

	    if (j != 0)
	      s = s.substr(j);

	    if (s[0] != '!') {
	      if (s[0] != '@')
		Error("provided result section -- result string must start with @");
	      
	      pos1 = s.find("|");
	      if (pos1 == string::npos)
		Error("provided result section -- needs player result string");
	      
	      if (single_player) {
		
		if (pos1 == s.size()-1)
		  Error("provided result section -- no player scores?");
		i->result[0] = s.substr(0,pos1+1);
		s = s.substr(pos1+1);
		
	      } else {

		if (pos1 == s.size()-1)
		  Error("provided result section -- need second player's result string");
		pos2 = (s.substr(pos1+1)).find("|");
		
		if (pos2 == string::npos)
		  Error("provided result section -- need both players' result strings");
		
		if (pos1+pos2+2 == s.size()) {
		  i->result[0] = s;
		  s = "";
		}
		else{
		  i->result[0] = s.substr(0,pos1+pos2+2);
		  s = s.substr(pos1+pos2+2);
		}	  
	      }
	    } else { //fixme: hardcoded somethings about the error string...
	   
	      j = 0;
	      FORS (m, 2) {
		while(j<s.size() && s[j] != ' ')
		  j++;
		if (j >= s.size() && (m==0))
		  Error("1provided result section -- aborted games must be indicated with result string: " + ERROR);
		if (m == 0)
		  j++;
	      }
	      
	      if (s.substr(0,j) != ERROR)
		Error("2provided result section -- aborted games must be indicated with result string: " + ERROR);
	        
	      if (j != 0)
		s = s.substr(j);
	    
	      players[i->player_list[0]].errors++;
	      if (!single_player) 
		players[i->player_list[1]].errors++;

	      i->result[0] = ERROR;
	    }
	   
	    //get the scores
	    int matchingId = i->id + 1;;
	    if (!single_game && ((i->id) & 1))
	      matchingId = i->id - 1;

	    //if first of a game pair, make sure scores aren't provided

	    if (!single_player && !single_game && !games[matchingId].saved) {
	      //get rid of any spaces:
	      j = 0;
	      while(j < s.size() && s[j]== ' ')
		++j;
	      
	      if (j>=s.size() || s.empty()) {
		found = true;
		break;
	      } else
		Error("provided game results -- player scores should only be given for the second match of a game pair");
	    }

	    if (single_player || (!single_game && games[matchingId].saved)) {
	      j = 0;
	      while(j<s.size() && s[j]== ' ')
		++j;
	      if (j>=s.size())
		Error("provided game results -- score string missing?");
	      if (j!=0)
		s = s.substr(j);
	    
	      pos1 = s.find(" ");
	      if (pos1 == string::npos)
		Error("provided game results -- space delimited player scores?");
	      if (pos1 == s.size()-1)
		Error("provided game results -- player 1 score?");
	      if (s1 != s.substr(0,pos1))
		Error("provided game results -- player 1 name for score doesn't match player 1 name of game");
	      s = s.substr(pos1+1);
	    
	      j = 0;
	      while(j<s.size() && s[j]== ' ')
		++j;
	      if (j>=s.size())
		Error("provided game results -- score string missing?");
	      if (j!=0)
		s = s.substr(j);
	    
	      //Player 1 score:
	      if (!single_player) {
		pos1 = s.find(" ");
		if (pos1 == string::npos)
		  Error("provided game results -- no score for player 1?");
		if (pos1 == s.size()-1)
		  Error("provided game results -- no score information for player 2?");
	    
		players[i->player_list[0]].score += stringToNum(s.substr(0,pos1));
		s = s.substr(pos1+1);
	      } else 
		players[i->player_list[0]].score += stringToNum(s);

	      players[i->player_list[0]].maxPoints++;
	    
	      //Player 2 score:
	      if (!single_player) {
		pos1 = s.find(" ");
		if (pos1 == string::npos)
		  Error("provided game results -- space delimited player scores?");
		if (pos1 == s.size()-1)
		  Error("provided game results -- player 2 score?");
		if (s2 != s.substr(0,pos1))
		  Error("provided game results -- player 2 name for score doesn't match player 2 name of game");
		s = s.substr(pos1+1);
	    
		j = 0;
		while(j<s.size() && s[j]== ' ')
		  ++j;
		if (j>=s.size())
		  Error("provided game results -- score string missing?");
		if (j!=0)
		  s = s.substr(j);
		players[i->player_list[1]].score += stringToNum(s);
		players[i->player_list[1]].maxPoints++;
		
	      }
	    }

	    found = true;
	    break;
	  }
	} // FORALL
	if (found) {
	  i->progress = done;
	  i->saved = true;
	  ++games_found;
	  if (skipComments(in_gam)) break;
	  continue;
	}
	if (i == games.end())
	  Error("provided result section -- seed & player names must correspond to a game in this list");
       
	if (skipComments(in_gam)) break;
      } // FOREVER

      cout << " found " << games_found << " game(s)" << endl << endl;;
    }

    in_gam.close();
  }


  if (opt_g) {
    displayGames();
    exit(0);
  }

  if (opt_s) {
    displayScores();
    cout << endl;
    displayMatrix(filename);
    exit(0);
  }

  // run remaining games
  
  pthread_mutex_init(&mutexGames, NULL); //should put this higher up
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  //pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE); // -> out of memory!
  
  // here sleep a second & then scan the list for done games. 
  // write to disk files & grep
  // for res string & put these in the config file

  bool allDone = false;

  while (!allDone) {

    allDone = checkComputers(max_client_jobs, max_server_jobs);

    sleep(3);

    updateGameFile(filename+".gam");
  }		      

  cout << "tournament complete" << endl;

  displayMatrix(filename);
  // ***COME BACK: VALGRIND, STILL MAY LOSE MEM SOMEWHERE b/c exit()
  // doesn't call destructors

  pthread_attr_destroy(&attr); 
  pthread_mutex_destroy(&mutexGames);
  return 0;
}
