#ifndef Action_H
#define Action_H

/** @file Action.H
    Movement and scripted actions.

    $Id: Action.H,v 1.29 2007/04/05 16:06:54 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Action.H,v $
*/

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "SType.H"
#include "Coord.H"
#include "ScriptObj.H"
#include "SaveLoad.H"

class PlayerInfo;
class Game;
class GfxModule;
class GameObj;
class ActionScript;
class Tokenizer;
class ObjAction;

//===================================================================

/** Parameters that are passed to each script action. */

struct SActParam : public SaveLoad
{
  Vector<ScriptObj::Ptr> obj;
  Vector<stype> val;

  void save(std::ostream &os) const;
  void load(std::istream &is);
};

//===================================================================

typedef uint4 ObjId;
typedef stype (*ActionFunction)(const SActParam&);
typedef stype (ScriptObj::*ObjFunction)(const SActParam&);

//===================================================================

/** Stores the current action state of an object.
    Acts mostly as intermediate storage before the action is encoded and sent.
*/

class Action
{
protected:

  sint4 id;    ///< The id of the action within the object. NOOP = no action; (id <= -1) -> stop/move.
  Vector<sint4> params;
  sint4 priority;  // client-side usage for selecting the current action

public:

  mutable uint4 client_action_id; // unessential baggage, only for client
  
  Action()
  {
    clear();
  }

  void set(sint4 id_, const Vector<sint4> &params_, sint4 priority_);
  void clear();

  bool is_set() const { return id != NOOP; }
  sint4 get_id() const { return id; }
  sint4 get_priority() const { return priority; }
  
  /** Convert the action id and parameters to an encoded string. */
  std::string to_string(ObjId cid) const;

private:
  
  enum { NOOP = 1000000 };
};

//===================================================================

/** Describes the name and parameters of an action.
    Created during blueprint parsing.
*/

class ObjActionDefinition
{
public:

  /// An inference about what an action's parameters are supposed to represent.
  enum PARAM_FORM { NONE, OBJ_ID, COORD, OTHER };
  PARAM_FORM format;

  enum ACT_TYPE { SCRIPT, FUNCTION, UNKNOWN };
  ACT_TYPE act_type;

  std::string name;
  bool hidden;    ///< The action is part of the object, but should only be called by the server.
  bool external;  ///< The action is not local to the object referencing it.

  sint4 num_obj_ids;
  sint4 num_params;
  sint4 num_state_params;

  Vector<std::string> obj_id_names;
  Vector<std::string> param_names;
  Vector<std::string> state_param_names;

  std::string script;
  ObjAction *func;

  std::string file_name; ///< The file the script came from.
  sint4 line_num;        ///< Which in the file the action was defined.
  std::string path;      ///< The directory path where the file is located.

  ObjActionDefinition() {
    format = OTHER;
    act_type = UNKNOWN;
    hidden = external = false;
    num_obj_ids = num_params = num_state_params = 0;
    func = 0;
    line_num = 0;
  }

  ObjActionDefinition(std::string n, sint4 num_obj, sint4 num_param, sint4 num_s_param,
                      std::string fname = "", sint4 line = 0) {
    format = OTHER;
    act_type = UNKNOWN;
    hidden = external = false;
    num_obj_ids = num_obj;
    num_params = num_param;
    num_state_params = num_s_param;
    func = 0;
    name = n;
    file_name = fname;
    line_num = line;
  }

  /** Read an action definition, except the name, from a tokenizer.
      @param tokenizer  An initialized tokenizer pointing to the start of an action's
                        parameter list.  I.e. (...;...;...).
  */
  void parse_action_def(Tokenizer &tokenizer);
};

//===================================================================

/** An action (scripted or compiled) that a game object can take. */

class ObjAction {
private:

  /// All instances are owned by the class.  Maintain a list for cleanup.
  /// Also used to get action ids for saving - as long as the order doesn't change.
  static Vector<ObjAction*> all_actions;

  // private functions - no body defined
  ObjAction();
  ObjAction(const ObjAction &);
  ObjAction &operator=(const ObjAction &);

  static std::stack< std::string > dir_path; ///< Location the action runs from.

public:

  uint4 num_obj_ids;
  uint4 num_params;
  uint4 num_state_params;
  bool hidden;
  bool external;

  ActionFunction CFunc; ///< Compiled action.
  ActionScript *Script; ///< Scripted action.
  ObjFunction SFunc;

  enum ACTION_TYPE { COMPILED, SCRIPTED, CLASS_FUNC, UNKNOWN };
  ACTION_TYPE t;
  ObjActionDefinition::PARAM_FORM format;

  ObjAction(ActionFunction action, ObjActionDefinition &def);
  ObjAction(ActionScript *action, ObjActionDefinition &def);
  ObjAction(ObjFunction action, ObjActionDefinition &def);
  ~ObjAction();

  static void clear();
  static void create(ObjActionDefinition &def, PlayerInfo &pi,
                     Game &game, std::string &bp_name);

  static sint4 act_index(ObjAction *o);
  static ObjAction* get_action(sint4 i);

  static std::string get_pwd();

  stype Act(const SActParam &p);
};

typedef std::map<std::string, ObjAction*> ObjActionMap;
REGISTER_TYPEOF(1701, ObjActionMap::iterator);
REGISTER_TYPEOF(1702, ObjActionMap::const_iterator);
REGISTER_TYPEOF(1703, Vector<ObjAction*>::iterator);
REGISTER_TYPEOF(1704, Vector<ObjAction*>::const_iterator);

//===================================================================
//===================================================================

// example of a compiled object action
stype CompiledAction1(const SActParam &p);

//===================================================================
//===================================================================

class BaseActions {
public:

  BaseActions() {}
  virtual ~BaseActions() {}

  virtual void Act() = 0; ///< Execute all valid actions queued.
};

//===================================================================

// Actions taken by script objects.

class SimpleActions : public BaseActions, public SaveLoad {

public:

  struct saction {
    std::string name;
    uint4 zero_tick; ///< The tick at which the action should be executed.
    uint4 sec_key;   ///< Secondary key used to sort (randomize) actions with the same zero_tick.

    ObjAction *a;
    sint4 obj_act_id; ///< -1 iff the action is an extension to the game definition
    bool active;     ///< Whether the action is still valid or has been cancelled.

    SActParam p;     ///< Action parameters.

    bool obj_ref;    ///< Whether the object still maintains a pointer to this action.
                     ///< This will be invalidated if the object creates a newer action.
    bool queue_ref;  ///< If the action is still in (referenced by) the queue.
  };

private:

  struct SortComp {
    // true iff x should be executed after y
    bool operator()(saction *x, saction *y) {
      if (x->zero_tick > y->zero_tick) return true;
      if (x->zero_tick < y->zero_tick) return false;

      return x->sec_key > y->sec_key; // check secondary key (randomizes order)
    }
  };

  SimpleActions(const SimpleActions &) : BaseActions(), SaveLoad() { ERR("what"); }
  SimpleActions &operator=(const SimpleActions &) { ERR("what"); }

  Game *game;

  std::priority_queue<saction*, Vector<saction*>, SortComp > actions;

public:

  SimpleActions(Game *g=0) : game(g) {}
  ~SimpleActions();

  // assume the first num_obj_ids parameters are client object ids
  void add(ScriptObj *x, sint4 act_id, Vector<stype> &param, uint4 delay);

  // assumes state params are included; pointers as passed, not ids
  void spawn(saction *q);
  stype spawn_now(saction *q);

  void save(std::ostream &os) const;
  void load(std::istream &is);

  void Act();
};

//===================================================================

class MoveActions : public BaseActions {

private:

  MoveActions(const MoveActions &) : BaseActions() { ERR("what"); }
  MoveActions &operator=(const MoveActions &) { ERR("what"); }  

  Game *game;

public:

  struct maction {
    GameObj *a;    ///< The object to start moving.
    ScalarPoint t; ///< The location to move to (straight line).
    sint4 speed;

    maction() : a(0) { speed = 0; }
  };

  Vector<maction> moves;

  MoveActions(Game *g=0) : game(g) {}
  ~MoveActions() {}

  void add (GameObj *a, sint4 x, sint4 y, sint4 speed) {
    maction q;
    q.a = a;
    q.t.x = x;
    q.t.y = y;
    q.speed = speed;
    moves.push_back(q);
  }

  /// Start all of the queued objects moving, but does not affect position.
  void Act();
};

//===================================================================

#endif
