/** @file Tile.C
    @see Tile.H

    $Id: Tile.C,v 1.14 2007/03/21 17:16:02 orts_mburo Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Tile.C,v $
*/
// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Tile.H"
#include "ByteData.H"

using namespace std;

// read/write all tile data
#define FULL_DATA
 
//===================================================================

void Tile::read_b(ByteStream &s)
{
#ifdef FULL_DATA
  h[0]  = s.read_b_int4();
  h[1]  = s.read_b_int4();
  h[2]  = s.read_b_int4();
  h[3]  = s.read_b_int4();
  h[4]  = s.read_b_int4();
  h[5]  = s.read_b_int4();
  h[6]  = s.read_b_int4();
  h[7]  = s.read_b_int4();

  split = s.read_b_int4();
  typeW = s.read_b_int4();
  typeE = s.read_b_int4();

  max_h = s.read_b_int4();
  min_h = s.read_b_int4();

#else
  data[0] = read_b_char();
  data[1] = read_b_char();
  if (data[0] != 0) data[2] = read_b_char();
  set_corners();
#endif
}

void Tile::write(string &s)
{
#ifdef FULL_DATA
  write_b_int4(s, h[0]);
  write_b_int4(s, h[1]);
  write_b_int4(s, h[2]);
  write_b_int4(s, h[3]);
  write_b_int4(s, h[4]);
  write_b_int4(s, h[5]);
  write_b_int4(s, h[6]);
  write_b_int4(s, h[7]);

  write_b_int4(s, split);
  write_b_int4(s, typeW);
  write_b_int4(s, typeE);

  write_b_int4(s, max_h);
  write_b_int4(s, min_h);

#else
  set_bits();
  s += (char)data[0];
  s += (char)data[1];
  if (data[0] != 0) s += (char)data[2];
#endif
}

void Tile::save(ostream &os) const
{
  FORU (i, 8) OS_WRITE(h[i]);
  OS_WRITE(split);
  OS_WRITE(typeW);
  OS_WRITE(typeE);
  OS_WRITE(max_h);
  OS_WRITE(min_h);
}

void Tile::load(istream &is)
{
  FORU (i, 8) IS_READ(h[i]);
  IS_READ(split);
  IS_READ(typeW);
  IS_READ(typeE);
  IS_READ(max_h);
  IS_READ(min_h);
}

void Tile::clear()
{
  memset(data.begin(), 0, data.end()-data.begin()); // array, not pointer!
  memset(h.begin(), 0, h.end()-h.begin());          // array, not pointer!
  min_h = max_h = 0;
  split = NO_SPLIT;
  flat = true;
  half = false;
  typeW = typeE = UNKNOWN;
}


real4 Tile::height_at(real4 x, real4 y) const
{
  assert(0 <= x && x <= 1.0);
  assert(0 <= y && y <= 1.0);  
  
  switch (split) {

  case NO_SPLIT:
    {
      real4 xh1 = x*h[4] + (1.0-x)*h[5];
      real4 xh2 = x*h[1] + (1.0-x)*h[0];
      return y*xh1 + (1.0-y)*xh2;
    }
    
  case TB_SPLIT:
    {
      if (y < x) {

	if (y == 1.0) return h[3];
	real4 xy = (x-y)/(1.0-y);
	return (h[0]*xy + h[1]*(1.0-xy))*(1.0-y) + h[3]*(y);

      } else {

	if (y == 0.0) return h[7];
	real4 xy = (y-x)/y;
	return (h[5]*xy + h[4]*(1.0-xy))*(y) + h[7]*(1.0-y);
      }
    }
    
  case BT_SPLIT:
    {
      if (y > 1.0-x) {

	if (y == 0.0) return h[2];
	real4 xy = (x-1.0+y)/y;
	return (h[4]*xy + h[5]*(1.0-xy))*(y) + h[2]*(1.0-y);

      } else {

	if (y == 1.0) return h[6];
	real4 xy = (1.0-y-x)/(1.0-y);
	return (h[0]*xy + h[1]*(1.0-xy))*(1.0-y) + h[6]*(y);
      }
    }
  }
  ERR("should not get here");
  return 0;
}


/** The type at a tile corner, with each corner possibly being split.
    @param c A tile corner, N side = 0,1; E = 2,3; ...
*/

Tile::Type Tile::get_type_at_corner(sint4 c) const
{
  if (typeW == typeE) return static_cast<Type>(typeW);
  if (split == TB_SPLIT) { // |\|
    return static_cast<Type>(c >= 4 ? typeW : typeE);
  } else { // |/|
    return static_cast<Type>(c >= 2 && c <= 5 ? typeE : typeW);
  }
}


void Tile::get_edge_height(sint4 &h1, sint4 &h2, Compass dir) const
{
  assert(dir & 1); // only N,E,S,W allowed

  h1 = h[dir-1];
  h2 = h[dir];
  return;

#if 0
  // was:
  switch (dir) {
  case N:  h1 = h[0];  h2 = h[1]; return;
  case E:  h1 = h[2];  h2 = h[3]; return;
  case S:  h1 = h[4];  h2 = h[5]; return;
  case W:  h1 = h[6];  h2 = h[7]; return;
  default:
    ERR("illegal dir");
    return;
  }
#endif
}


void Tile::set_flat(sint4 height)
{
  assert(height >= 0 && height <= MAX_HEIGHT);
  split = 0;
  flat = true;
  half = false;

  min_h = max_h = height;
  FORU (i, 8) h[i] = height;
}


void Tile::set_ramp(sint4 start_height, sint4 end_height, Compass dir)
{
  assert(start_height >= 0 && start_height <= MAX_HEIGHT);
  assert(end_height >= 0 && end_height <= MAX_HEIGHT);  

  half = false;
  flat = start_height == end_height;

  if (start_height > end_height) {
    swap(start_height, end_height);
    if (dir > 4) dir = static_cast<Compass>(dir - 4);
    else         dir = static_cast<Compass>(dir + 4);
  }

  min_h = min(start_height, end_height);
  max_h = max(start_height, end_height);

  sint4 new_split = 0;
  if (dir == NE || dir == SW) new_split = TB_SPLIT;
  if (dir == NW || dir == SE) new_split = BT_SPLIT;

  if (new_split != split) {
    if (new_split != 0 || typeE == typeW) {
      typeE = typeW;
      split = new_split;
    }
  }

  sint4 mid_h = (start_height + end_height) / 2;
  switch (dir) {
  case N:
    h[3] = h[4] = h[5] = h[6] = start_height;
    h[7] = h[0] = h[1] = h[2] = end_height;
    break;
  case NE:
    h[5] = h[6] = start_height;
    h[3] = h[4] = h[7] = h[0] = mid_h;
    h[1] = h[2] = end_height;
    break;
  case E:
    h[5] = h[6] = h[7] = h[0] = start_height;
    h[1] = h[2] = h[3] = h[4] = end_height;
    break;
  case SE:
    h[7] = h[0] = start_height;
    h[1] = h[2] = h[5] = h[6] = mid_h;
    h[3] = h[4] = end_height;
    break;
  case S:
    h[7] = h[0] = h[1] = h[2] = start_height;
    h[3] = h[4] = h[5] = h[6] = end_height;
    break;
  case SW:
    h[1] = h[2] = start_height;
    h[3] = h[4] = h[7] = h[0] = mid_h;
    h[5] = h[6] = end_height;
    break;
  case W:
    h[1] = h[2] = h[3] = h[4] = start_height;
    h[5] = h[6] = h[7] = h[0] = end_height;
    break;
  case NW:
    h[3] = h[4] = start_height;
    h[1] = h[2] = h[5] = h[6] = mid_h;
    h[7] = h[0] = end_height;
    break;
  default:
    ERR("?");
  }
}


void Tile::set_ramp(sint4 start_height, sint4 mid_height, sint4 end_height, Compass dir)
{
  assert(start_height >= 0 && start_height <= MAX_HEIGHT);
  assert(end_height >= 0 && end_height <= MAX_HEIGHT);
  assert(mid_height >= 0 && mid_height <= MAX_HEIGHT);
  assert(mid_height >= start_height || mid_height >= end_height);
  assert(mid_height <= start_height || mid_height <= end_height);  
  
  half = false;
  flat = start_height == mid_height && mid_height == end_height;

  if (start_height > end_height) {
    swap(start_height, end_height);
    if (dir > 4) dir = static_cast<Compass>(dir - 4);
    else         dir = static_cast<Compass>(dir + 4);
  }

  sint4 new_split = 0;
  if (dir == NE || dir == SW) new_split = TB_SPLIT;
  if (dir == NW || dir == SE) new_split = BT_SPLIT;

  if (new_split != split) {
    if (new_split != 0 || typeE == typeW) {
      typeE = typeW;
      split = new_split;
    }
  }

  min_h = min(start_height, end_height);
  max_h = max(start_height, end_height);

  if (mid_height < min_h) min_h = mid_height;
  if (mid_height > max_h) max_h = mid_height;

  switch (dir) {
  case N:
    h[3] = h[4] = h[5] = h[6] = start_height;
    h[7] = h[0] = h[1] = h[2] = end_height;
    break;
  case NE:
    h[5] = h[6] = start_height;
    h[3] = h[4] = h[7] = h[0] = mid_height;
    h[1] = h[2] = end_height;
    break;
  case E:
    h[5] = h[6] = h[7] = h[0] = start_height;
    h[1] = h[2] = h[3] = h[4] = end_height;
    break;
  case SE:
    h[7] = h[0] = start_height;
    h[1] = h[2] = h[5] = h[6] = mid_height;
    h[3] = h[4] = end_height;
    break;
  case S:
    h[7] = h[0] = h[1] = h[2] = start_height;
    h[3] = h[4] = h[5] = h[6] = end_height;
    break;
  case SW:
    h[1] = h[2] = start_height;
    h[3] = h[4] = h[7] = h[0] = mid_height;
    h[5] = h[6] = end_height;
    break;
  case W:
    h[1] = h[2] = h[3] = h[4] = start_height;
    h[5] = h[6] = h[7] = h[0] = end_height;
    break;
  case NW:
    h[3] = h[4] = start_height;
    h[1] = h[2] = h[5] = h[6] = mid_height;
    h[7] = h[0] = end_height;
    break;
  default:
    ERR("?");
  }
}

void Tile::set_type(Type tW, Type tE, Split s)
{
  typeW = tW;
  typeE = tE;
  if (tW != tE && s == NO_SPLIT) s = TB_SPLIT;
  split = s;
}

void Tile::set_half_tile(sint4 start_height, sint4 end_height, Compass dir)
{
  assert(start_height >= 0 && start_height <= MAX_HEIGHT);
  assert(end_height >= 0 && end_height <= MAX_HEIGHT);

  if (start_height == end_height) {
    set_flat(start_height);
    return;
  }

  flat = false;
  half = true;

  sint4 new_split;

  switch (dir) {
  case SW: swap(start_height, end_height); 
  case NE: new_split = TB_SPLIT; break;    
  case NW: swap(start_height, end_height); 
  case SE: new_split = BT_SPLIT; break;    
  default: ERR("here?"); new_split = BT_SPLIT;
  }

  if (new_split != split) {
    typeE = typeW;
    split = new_split;
  }

  if (split == TB_SPLIT) {
    h[4] = h[5] = h[6] = h[7] = start_height;
    h[0] = h[1] = h[2] = h[3] = end_height;
  } else {
    h[6] = h[7] = h[0] = h[1] = start_height;
    h[2] = h[3] = h[4] = h[5] = end_height;
  }

  min_h = min(start_height, end_height);
  max_h = max(start_height, end_height);
}


void Tile::set_topo(sint4 hNW, sint4 hNE, sint4 hSE, sint4 hSW, Split split_)
{
  h[7] = h[0] = hNW;
  h[1] = h[2] = hNE;
  h[3] = h[4] = hSE;
  h[5] = h[6] = hSW;

  half = false;
  flat = (hNW == hNE && hNE == hSE && hSE == hSW);

  split = split_;

  min_h = hNW;
  if (hNE < min_h) min_h = hNE;
  if (hSE < min_h) min_h = hSE;
  if (hSW < min_h) min_h = hSW;

  max_h = hNW;
  if (hNE > max_h) max_h = hNE;
  if (hSE > max_h) max_h = hSE;
  if (hSW > max_h) max_h = hSW;
}


void Tile::get_topo(sint4 &hNW, sint4 &hNE, sint4 &hSE, sint4 &hSW, Split &split_) const
{
  hNW = h[0];
  hNE = h[2];
  hSE = h[4];
  hSW = h[6];
  split_ = static_cast<Split>(split);
}

/** Convert the bit-encoded data to corner heights.
    These corner heights implicitly define walls between adjacent tiles.
*/

void Tile::set_corners()
{

  ERR("does not do anything");
  
  /*
  uint1 &topography = data[0];
  uint1 &byte1 = data[1];
  uint1 &byte2 = data[2];
  */
}

/** Encodes the slope direction and height values as bit data. */

void Tile::set_bits()
{
  ERR("does not do anything");
  
  data[0] = 0;
  data[1] = 0;
  data[2] = 0;
}

//===================================================================
