// $Id: OSL.H,v 1.5 2004/10/29 16:58:40 orts_jyip Exp $

// (c) Michael Buro, mburo@cs.ualberta.ca, licensed under the GPL

#ifndef OSL_H
#define OSL_H
 
#include "Global.H"
#include "OSL_Opcodes.H"


namespace OSL {

  const sint4 POINTER_SIZE  = sizeof(void*);              // 4 or 8 
  const sint4 POINTER_WORDS = POINTER_SIZE/sizeof(sint4); // 2 or 4
  
  class Type;
  class Object;
  class Context;
  class Program;
  class FuncMember;
  class ClassInstance;
  class ArrayInstance;  
  
  typedef std::string TypeName;
  typedef uint4 CodeAddr;
  typedef sint4 Code;
  typedef uint2 TypeIndex;

  struct Exception : public std::exception {

    std::string msg;

    Exception(const std::string &s="") : msg(s) {}
    ~Exception() throw() { }

    const char *what() const throw() { return msg.c_str(); }
  };


  /** describes types of OSL variables */
  
  class Type
  {
  public:

    enum Category {

      // important: no holes up to TC_cref! (see Program::init_types)
      // also: if changed, adjust sizes[] below!
      
      TC_unknown = 0,
      TC_pc,       // code index where to resume execution in caller
      TC_fp,       // caller's frame pointer (actually index)
      TC_fu,       // caller's pointer to FuncMember 
      TC_void,
      TC_sint1,
      TC_uint1,   
      TC_sint2,   
      TC_uint2,
      TC_sint4,
      TC_uint4,
      TC_sint8,
      TC_uint8,
      TC_real4,   
      TC_real8,
      TC_object,
      TC_END = TC_object, // end of stack variable types (see Program::init_types)
      TC_class,   // class reference 
      TC_array,   // array reference 
      TC_string   // string reference
    };

    Category cat;  // type category - one of the above

    // referenced class or referenced array element type (if class)
    // cat == TC_cref or (cat == TC_aref and T_cat == TC_cref)
    uint4 class_index;

    // cat == TC_aref
    sint4 array_rank;  // number of dimensions
    Category T_cat;    // category of T (can't be aref) where array = T[]...[]
    
    void write(std::ostream &os = std::cout) const;

    static void write_cat(Category cat, std::ostream &os = std::cout);

    // size of variable of this type
    sint4 size() const {
      if (cat >= TC_END) return POINTER_SIZE; else return sizes[cat];
    }

  protected:

    static const sint4 sizes[TC_END];
  };


  /** local variables, parameters, and temporal variables (all located on stack) */
  
  struct Value {

    union {
      CodeAddr       v_pc;
      uint4          v_fp;
      FuncMember    *v_fu;
      sint4          v_sint4;
      uint4          v_uint4;
      sint8          v_sint8;
      uint8          v_uint8;
      real4          v_real4;            
      real8          v_real8;
      Object        *v_object;
    };

    Type::Category cat; // type category (TC_sint1..TC_object)
    // if object, actual type can be deduced

    Value() { cat = Type::TC_unknown; v_uint8 = 0; }
    // fixme: does that initialize pointers and reals, too?

    Type::Category get_cat() const { return static_cast<Type::Category>(cat); }
    void set_cat(Type::Category cat_) { cat = cat_; }
  };

  class Attributes {
    std::string name;
  };

  enum Modifier { PRIVATE, PROTECTED, PUBLIC };


  /** Class data member description */
  
  class DataMember {

  public:
    std::string name;
    //    Type type;
    TypeIndex type;
    uint4 offset; // into data area
    Modifier mod;
    bool is_static;
    //std::vector<string> attributes; // static information (e.g. for modeling sharing)

    DataMember() { offset = 0; is_static = false; }
  };


  /** Class function member description (either virtual or static) */
  
  class FuncMember {

  public:
    
    std::string name;
    TypeIndex ret_tind;           // return value type index
    Modifier mod;
    bool is_static;
    uint4 class_id;               // where defined
    
    std::vector<TypeIndex> par_types;  // parameter type indexes 
    std::vector<TypeIndex> loc_types;  // local variable type indexes
    std::vector<Code> code;       // executable code

    std::vector<Value> loc_init;     // initial image of local variables (for memcopy)
    
    //std::vector<std::string> attributes; // static information (e.g. for modeling sharing)

    FuncMember() { is_static = false; ret_tind = Type::TC_void; }

    // code generation

    // fixme: test 8-byte instructions
    
    void add(sint4 x)          { std::cout << "ADDING " << x << " at " << code.size() << "*******************************************\n"; code.push_back(x); }
    void add_code (sint4 x)    { add(x); }
    void add_addr (CodeAddr x) { add(x); }
    void add_sint4(sint4 x)    { add(x); }
    void add_uint4(uint4 x)    { add(x); }
    void add_sint8(sint8 x)    { add(*((uint4*)&x)); add(*(((uint4*)&x)+1)); }
    void add_uint8(sint8 x)    { add(*((uint4*)&x)); add(*(((uint4*)&x)+1)); }
    void add_real4(real4 x)    { add(*((uint4*)&x)); } 
    void add_real8(real8 x)    { add(*((uint4*)&x)); add(*(((uint4*)&x)+1)); }

    void set_addr(uint4 loc, CodeAddr x) { code[loc] = x; }

    void set_code(uint4 loc, sint4 x) { code[loc] = x; }
    void set_uint4(uint4 loc, uint4 x) { code[loc] = x; }

    CodeAddr get_addr() const { return code.size(); }
  };


  /** Class description */

  class Class {

  public:
    bool defined;
    std::string name;
    TypeIndex ref_tind;  // index of reference type for this class (set from outside)

    sint4 super_class_index;  // < 0: no super class
    // interfaces later: vector<sint4> interface_indexes;
    
    std::vector<DataMember> inst_data;
    std::vector<DataMember> static_data;    

    std::vector<FuncMember> virt_funcs;
    std::vector<FuncMember> static_funcs;

    FuncMember static_init;  // must be static function

    std::vector<Value> static_values;

    __gnu_cxx::hash_map<std::string, uint4> inst_data_ids;
    __gnu_cxx::hash_map<std::string, uint4> static_data_ids;
    __gnu_cxx::hash_map<std::string, uint4> virt_func_ids;
    __gnu_cxx::hash_map<std::string, uint4> static_func_ids;
    
    Class();
    
    /** cache values + init. static variables */
    void init_class() { inst_data_size = inst_data.size() * sizeof(Value); }

    /** clear instance variables */
    void init_instance(ClassInstance *ci);

    void inherit(Class &super_class) {
      uint4 end = super_class.inst_data.size();
      FORU(i, end) {
	if (super_class.inst_data[i].mod == PRIVATE)
	  continue;

	inst_data.push_back(super_class.inst_data[i]); 
	inst_data_ids[inst_data[i].name] = i;
	if (super_class.inst_data[i].mod == PROTECTED)
	  inst_data[i].mod = PRIVATE;
      }

      end = super_class.static_data.size();
      FORU(i, end) {
	if (super_class.static_data[i].mod == PRIVATE)
	  continue;

	static_data.push_back(super_class.static_data[i]);   
	static_data_ids[static_data[i].name] = i;
	if (super_class.static_data[i].mod == PROTECTED)
	  static_data[i].mod = PRIVATE;
      }

      end = super_class.virt_funcs.size();
      FORU(i, end) {
	if (super_class.virt_funcs[i].mod == PRIVATE)
	  continue;

	virt_funcs.push_back(super_class.virt_funcs[i]);
	virt_func_ids[virt_funcs[i].name] = i;
	if (super_class.virt_funcs[i].mod == PROTECTED)
	  virt_funcs[i].mod = PRIVATE;
      }

      end = super_class.static_funcs.size();
      FORU(i, end) {
	if (super_class.static_funcs[i].mod == PRIVATE)
	  continue;

	static_funcs.push_back(super_class.static_funcs[i]);
	static_func_ids[static_funcs[i].name] = i;
	if (super_class.static_funcs[i].mod == PROTECTED)
	  static_funcs[i].mod = PRIVATE;
      }
    }
    
    void inherit2(Class &super_class) {
      uint4 end = super_class.virt_funcs.size();
      FORU(i, end) {
	if (super_class.virt_funcs[i].mod == PRIVATE)
	  continue;
 
	virt_funcs[i] = super_class.virt_funcs[i];
	virt_func_ids[virt_funcs[i].name] = i;
	std::cout << "^^^^^^^^^HOW COME " << virt_funcs.size()-1 << std::endl;
	if (super_class.virt_funcs[i].mod == PROTECTED)
	  virt_funcs[i].mod = PRIVATE;
      }
      
      end = super_class.static_funcs.size();
      FORU(i, end) {
	if (super_class.static_funcs[i].mod == PRIVATE)
	  continue;
	
	static_funcs[i] = super_class.static_funcs[i];
	static_func_ids[static_funcs[i].name] = i;
	std::cout << "^^^^^^^^HOW COME " << static_funcs.size()-1 << std::endl;
	if (super_class.static_funcs[i].mod == PROTECTED)
	  static_funcs[i].mod = PRIVATE;
      }
    }

    /** total data size in bytes (including padded space) */
    uint4 inst_data_size; // computed by init_class()

#if 0
    Value get(const std::string &x) const;
    void  set(const std::string &x, Value &v);
  
    sint4 get_sint4(const std::string &x);
    sint4 set_sint4(const std::string &x, sint4 v);
    real4 get_real4(const std::string &x);
    real4 set_real4(const std::string &x, real4 v);

    sint4 get_sint4_arr(sint4 index);
    sint4 set_sint4_arr(sint4 index, sint4 v);  

    sint4 call_sint4(const std::string &x, Context &cont);
    real4 call_real4(const std::string &x, Context &cont);
#endif
  };

  /** garbage collected heap object (either class or array instance) */
  
  struct Object {

    TypeIndex tind;
    uint2 flags;

  };
  

  /** this is how class instances are stored on the heap */
  
  struct ClassInstance : public Object {

    Value data[]; // data follows here (can't use new for ClassInstance!)

    ClassInstance() { tind = Type::TC_unknown; }
  };


  /** how array instances are stored on the heap */
  
  struct ArrayInstance : public Object {
  
    uint4 n; // number of elements

    Value data[]; // data follows here (can't use new for ArrayInstance!)

    ArrayInstance() { tind = Type::TC_unknown; n = 0; }
  };
  

  /** External function description */
  
  class ExtFunc {

  public:
    
    std::string name;
    TypeIndex ret_tind;                // return value type index
    std::vector<TypeIndex> par_types;  // parameter type indexes 

    typedef Value (*C_Func)(std::vector<Value> &params);

    C_Func c_func;
    
    ExtFunc() { ret_tind = Type::TC_void; c_func = 0; }

    Value call(std::vector<Value> &params) { return (*c_func)(params); }
  };

  /** Heap */
  class Heap {
  public:
    typedef uint1 mem_unit;

    Heap(Program *prog_);

    mem_unit* alloc(uint4 size);
  private:
    static const uint4 init_heap_size = 1024;

    mem_unit* heap;
    uint4 heap_index;
    uint4 heap_size;

    Program *prog;

    void GC(uint4 size);
  };

  /** execution context */
  
  class Context {
  public:

    enum StepResult { END, CONT, YIELD };

    uint4 line_num;  // last line number
    
    Context(Program &prog_, sint4 sl=1000);
    virtual ~Context();

    /** prepare variables and stack for run */
    void init_run(FuncMember *new_fu, std::vector<Value> &par);

    /** execute one step */
    StepResult step();

    /** execute sequence
	n<0: only stop at END
	n=0: also stop at YIELD
	n>0: run n steps, stop at END and YIELD
    */
    
    StepResult run(sint4 n=-1);

    /** return result of execution */
    Value result() const;

    /** print stack contents to stream */
    void write_stack(std::ostream &os = std::cout) const;

    /** memory problem */
    struct MemException : public Exception {
      MemException(const std::string &s) throw() : Exception(s) { }
    };

    /** parameter problem */    
    struct ParException : public Exception {
      ParException(const std::string &s) throw() : Exception(s) { }
    };

    /** value out of range */    
    struct OutOfRangeException : public Exception {
      OutOfRangeException(const std::string &s) throw() : Exception(s) { }
    };

    /** type problem */    
    struct TypeException : public Exception {
      TypeException(const std::string &s) throw() : Exception(s) { }
    };

    friend class Heap;

  protected:

    Program &prog;     // access to program environment
    Value   *stack;    // execution stack
    uint4   stack_len; // its length
    
    CodeAddr    pc;    // program counter (index)
    uint4       sp;    // stack pointer (actually an index)
    uint4       fp;    // frame pointer (actually an index)
    FuncMember *fu;    // current function, call set_fu() to change

    /** cached fu variables */
    const Code *code;     // pointing to fu.code[0]
    uint4       code_len; // fu.code.size()

    /** prepare new frame (assumes all parameters have been pushed) */
    void init_frame(FuncMember *new_fu);

    /** set fu and cached variables */
    void set_fu(FuncMember *new_fu);

    /** check type of topmost variables on stack */
    void check_type(Type::Category t1);
    void check_types(Type::Category t1, Type::Category t2);

    /** assert(sp >= n) */
    void check_sp(uint4 n);

    /** assert(sp + n <= stack_len) */
    void check_ssize(uint4 n);

    /** push values on stack */

    void push(const Value &v);
    
    void push_sint4(sint4 v);
    void push_uint4(uint4 v);
    void push_sint8(sint8 v);
    void push_uint8(uint8 v);
    void push_real4(real4 v);
    void push_real8(real8 v);
    void push_null();

    /** indicates where runtime error happened  */
    std::string where() const;

    /** return previous fu */
    FuncMember *remove_frame(); 

    /** get the class top() points to */
    Class *get_class(const std::string &op);

    /** get stack index of parameter */
    sint4 get_par_ind(sint4 index) const { assert(index >= 0); return fp - 4 - index; }

    /** get stack index of local variable */
    sint4 get_loc_ind(sint4 index) const { assert(index >= 0); return fp + index; }
  };


  /** program environment */
  
  class Program {
  public:

    std::vector<Type>    types;
    std::vector<Class>   classes;
    std::vector<FuncMember> funcs;
    std::vector<ExtFunc> extfs;
    std::vector<Context*> contexts;

    Heap *heap;

    __gnu_cxx::hash_map<std::string, uint4> type_ids;
    __gnu_cxx::hash_map<std::string, uint4> class_ids;
    __gnu_cxx::hash_map<std::string, uint4> func_ids;
    __gnu_cxx::hash_map<std::string, uint4> extf_ids;


    Program();
    virtual ~Program() {}

    /** add class to class vector, also adds new class reference type */
    Class* add_class(const std::string& name);

    /** add array to type vector */
    void add_array(TypeIndex t, sint4 dimensions);

    /** add type to type vector (mainly arrays) */
    //    void add_type(const Type &type);

    // add external function to extfs vector
    void add_ext(ExtFunc::C_Func c_func, const std::string& name, std::vector<TypeIndex> par_types);
    
    // call static function
    Value run(const std::string &cl, const std::string &func, std::vector<Value> &params);
    Value run(const uint4 class_index, const uint4 func_index, std::vector<Value> &params);    

    // call virtual function
    Value run(ClassInstance *ci, const std::string &func, std::vector<Value> &params);
    Value run(ClassInstance *ci, uint4 func_index, std::vector<Value> &params);

    // to be implemented:
    //void create_object(Object *&p, const std::string &class_name, std::vector<Value> &params);
    //void release_object(Object *&p);

    /** write value to stream including type information */
    static void write_value(const Value &v, std::ostream &os = std::cout);

    /** write type information to stream */
    void write_type(const Type &type, std::ostream &os = std::cout) const;

    /** disassemble code */
    void write_code(const FuncMember &fm, std::ostream &os = std::cout) const;

    /** return total size of array data */
    uint4 array_data_size(ArrayInstance *a) {
      return a->n * types[a->tind].size();
    }

    /** compute address of array element */
    void *array_elem_addr(ArrayInstance *a, uint4 index) {
      if (index >= a->n) ERR("array index out of range");
      return ((uint1*)a) + index * types[a->tind].size();
    }
    /*
      void *array_elem_addr(ArrayInstance *a, sint4 index) {
      if (index < 0 || index >= a->n) ERR("array index out of range");
      return ((uint1*)a) + index * types[a->tind].size();
      }
    */

    /** get starting address of data */
    uint4 get_class_data_offset() const { return class_data_offset; }
    uint4 get_array_data_offset() const { return array_data_offset; }    

    
  protected:

    /** offsets of data in ClassInstance and ArrayInstance */
    uint4 class_data_offset, array_data_offset;

    /** initialize type vector with basic types */
    void init_types();
    
  };


}

#if 0

for now only Java arrays (1-dimensional)

Later C# Arrays:

int[,,] a = new int[2,3,4];

int[][,,] a = new int[3][3,4,5];

alias int[3,4,5] A;

A[3] b;

Type name?: T[2[3[4[2  (numbers are #dimensions!)
		   
#endif

#endif
