/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* $header$
   $log$
*/

#ifndef Coord_H
#define Coord_H

#include "Global.H"
#include "String.H"

#define USE_LRINT          0
#define USE_ROUNDING_MACRO 1


typedef sint4 Scalar;  // fine grid

class FixPoint {

public:

  inline static Scalar add(const Scalar &x, const Scalar &y) { return x+y; }
  inline static Scalar sub(const Scalar &x, const Scalar &y) { return x-y; }
  inline static Scalar abs(const Scalar &x) { return ::abs(x); }
  inline static sint8  to_sint8(const Scalar &x) { return sint8(x); }
  inline static sint8  square(const Scalar &x) { return sint8(x)*sint8(x); }
  
  static sint8 slow_from_real8(const real8 &f) {

    Scalar v;
    if      (f > 0.0) v = sint4(floor(f+0.5));
    else if (f < 0.0) v = sint4(ceil(f-0.5));
    else v = 0;
    return v;

  }

  static Scalar from_real8(const real8 &f) {

#if USE_LRINT    
    
    extern sint8 lrint(double);
    return lrint(f);

#elif USE_ROUNDING_MACRO

#define FLOAT_TO_INT(in,out)  \
    __asm__ __volatile__ ("fistpl %0" : "=m" (out) : "t" (in) : "st") ;

    Scalar v;
    FLOAT_TO_INT(f, v);

#if 0  
    cout << f << " -> " << v << endl;
    if (v != slow_from_real8(f)) {
      cout << slow_from_real8(f) << endl;
      exit(1);
    }
#endif

    return v;

#else    
    
    Scalar v;
    if      (f > 0.0) v = sint4(floor(f+0.5));
    else if (f < 0.0) v = sint4(ceil(f-0.5));
    else v = 0;
    return v;

#endif

  }

  static Scalar scale(const Scalar &x, const real8 &f) {
    return from_real8(x*f);
  }
  
  static Scalar round(const Scalar &x, sint4 bits) {

    // untested
    
    Scalar y = x;
    sint4 mask = 1 << bits;
    if      (y > 0) y += mask >> 1;
    else if (y < 0) y -= mask >> 1;
    y &= ~(mask-1);
    return y;
  }

  static String to_string(const Scalar &x) {
    String s;
    s.form("%d", x);
    return s;
  }

};


class Pos {

public:
  
  Scalar x, y, z;

  Pos() { x = 0; y = 0; z = 0; }

  Pos(Scalar x_, Scalar y_, Scalar z_) {
    x = x_;
    y = y_;
    z = z_;
  }

  Pos(Scalar x_, Scalar y_) {
    x = x_;
    y = y_;
    z = 0;
  }

  bool bad() const {

    return (x == min_sint4 || abs(x) >= 500000000 ||
	    y == min_sint4 || abs(y) >= 500000000 ||
	    z == min_sint4 || abs(z) >= 500000000);
  }
  
  Scalar l2_norm() const {

    if (bad()) {
      write(cerr);
      ERR("scalar out of range");
    }

    return FixPoint::from_real8(sqrt(real8(FixPoint::square(x) +
					   FixPoint::square(y) +
					   FixPoint::square(z))));
  }

  sint8 l2_norm_sq() const {

    if (bad()) {
      write(cerr);
      ERR("scalar out of range");
    }

    return FixPoint::square(x) + FixPoint::square(y) + FixPoint::square(z);
  }

  Scalar max_norm() const {

    Scalar m = FixPoint::abs(x);
    m = ::max(m, FixPoint::abs(y));
    return ::max(m, FixPoint::abs(z));    
  }

  Scalar l2_dist(const Pos &y) const {
    Pos d = *this;
    d.sub(y);
    return d.l2_norm();
  }

  sint8 l2_dist_sq(const Pos &y) const {
    Pos d = *this;
    d.sub(y);
    return d.l2_norm_sq();
  }

  Scalar max_dist(const Pos &y) const {
    Pos d = *this;
    d.sub(y);
    return d.max_norm();
  }

  void add(const Pos &p) {
    x = FixPoint::add(x, p.x);
    y = FixPoint::add(y, p.y);
    z = FixPoint::add(z, p.z);    
  }

  void sub(const Pos &p) {
    x = FixPoint::sub(x, p.x);
    y = FixPoint::sub(y, p.y);
    z = FixPoint::sub(z, p.z);    
  }  

  void scale(real8 f) {
    x = FixPoint::scale(x, f);
    y = FixPoint::scale(y, f);
    z = FixPoint::scale(z, f);
  }

  void round(sint4 bits) {
    x = FixPoint::round(x, bits);
    y = FixPoint::round(y, bits);
    z = FixPoint::round(z, bits);
  }

  void write(ostream &os = cout) const {
    os << "(" << x << "," << y << "," << z << ")";
  }
  
};


inline bool operator==(const Pos &c1, const Pos &c2) {
  return c1.x == c2.x && c1.y == c2.y && c1.z == c2.z;
}

#endif
