#include "Global.H"
#include <vector>

const sint4 PLAYER_MAX_NUM = 8;

class Pair {

public:

  sint4 x, y;
};

class Player {

public:

  Player() {}
  virtual ~Player() {}
};


class Tile {

public:

  sint1 level;
  
  Tile() { level = 0; }
  virtual ~Tile() {}
};


class TileInfo {

public:

  sint2 vis_counters[PLAYER_MAX_NUM];

  void reset() {
    memset(vis_counters, 0, PLAYER_MAX_NUM*sizeof(vis_counters[0]));
  }
  
  TileInfo() {
    reset();
  }
  
  virtual ~TileInfo() {}
};


class Object {

public:

  sint4 tile_x, tile_y;
  sint4 dx, dy;         // relative to tile
  sint1 heading;
  sint4 owner_set;

  sint4 height, width;
  sint1 sight_range;
  sint1 speed;
  
  Object() {
    tile_x = 0;
    tile_y = 0;
    dx = 0;
    dy = 0;
    heading = 0;
    owner_set = 0;
    height = 0;
    width = 0;
    sight_range = 0;
    speed = 0;
  }
  
  virtual ~Object() {}
};


class Map {

public:

  sint4 tile_w, tile_h;   // # of tiles on the map
  sint4 tile_res;         // # points on each dimension (power of 2)
  
  vector<Tile> tiles;     // mailboxed array of tiles
  vector<Object*> objects;
  
  // scratch space

  sint4 tile_num;         // tile_w * tile_h
  vector<TileInfo> tile_infos;

  sint4 xy2ind(sint4 x, sint4 y) {
    assert(x >= 0 && x < tile_w);
    assert(y >= 0 && y < tile_h);    
    return y * tile_w + x;
  }
  
  Map() {

    tile_w = 64; tile_h = 32;
    tile_num = tile_w * tile_h;
    tile_res = 128;

    tiles.reserve(tile_num);

    Tile tile;
    TileInfo tinfo;
    
    FORS (i, tile_num) {
      tiles.push_back(tile);
      tile_infos.push_back(tinfo);
    }

    tiles[xy2ind(5,5)].level = 1;
    //tiles[xy2ind(5,6)].level = 1;
    tiles[xy2ind(5,7)].level = 1;
    //tiles[xy2ind(5,8)].level = 1;
    tiles[xy2ind(5,9)].level = 1;    
    
    assert((sint4)tiles.size() == tile_num);
  }

  virtual ~Map() {
    FORU (i, objects.size()) { delete objects[i]; objects[i] = 0; }
  }

  void print_tiles(ostream &os = cout) {
    FORS (y, tile_h) {
      FORS (x, tile_w) {
	os << (sint4)tiles[xy2ind(x,y)].level;
      }
      os << endl;
    }
  }

  void print_view(sint4 player_index, ostream &os = cout) {
    const sint4 OBJS_SIZE = tile_h*tile_w;
    sint1 objs[OBJS_SIZE];
    memset(objs, 0, sizeof(objs[0])*OBJS_SIZE);

    FORALL (vector<Object*>, objects, i) {
      objs[xy2ind((*i)->tile_x, (*i)->tile_y)] = 1;
    }
    
    FORS (y, tile_h) {
      FORS (x, tile_w) {
	sint4 index = xy2ind(x, y);
	if (tile_infos[index].vis_counters[player_index] == 0)
	  os << "??";
	else {
	  os << (sint4)tiles[index].level;
	  if (objs[index]) os << "#"; else os << " ";
	}
      }
      os << endl;
    }
  }

  sint4 ok_dist2(sint4 k)
  {
    return ((10*k+7)*(10*k+7))/100;  // ~ (k+sqrt(2)/2)^2
  }


  void reset_tile_info() {
    FORALL (vector<TileInfo>, tile_infos, i) i->reset();
  }
  
  void generate_view_template(sint4 k, vector<sint4> &view)
  {
    view.clear();
    
    sint4 xm = k;
    sint4 ym = k;    

    sint4 d2 = ok_dist2(k);
    char c;
    
    FORS (y, 2*k+1) {
      FORS (x, 2*k+1) {
	if ((x-xm)*(x-xm) + (y-ym)*(y-ym) <= d2) {
	  if (x == xm && y == ym) c = 'o'; else c = '+'; 
	} else {
	  c = ' ';
	}
	view.push_back(c);
      }
    }
  }


  // computes a vector of points at distance at most k that are occluded by
  // point (idx,idy) in view of (0,0)
  // occlusion: square mid-point not in epsilon sight (epsilon>0)
  
  void generate_occlusions(sint4 k, sint4 idx, sint4 idy, vector<Pair> &occ)
  {
    /*
            viewer             obstacle
           o-----o           a o------o
           |     |             |      |       
           |  X----------------|---*-----------* mid-point
           |(0,0)|             |(dx,dy)          (ndx, ndy) 
	   o-----o	     b o------o


	   mid-point (ndx, ndy) invisible from viewer :<=>

	     [(0,0),(ndx,dny)] intersects [a,b]
    */

    occ.clear();

    real8 dx = idx, dy = idy;
    real8 ax, ay, bx, by;
    
    if (dx == 0) {

      ax = dx - 0.5;
      ay = dy;
      bx = dx + 0.5;
      by = dy;
      
    } else {
      
      if (dy == 0) {

	ax = dx;
	ay = dy-0.5;
	bx = dx;
	by = dy+0.5;

      } else {

	// dx != 0 && dy != 0

	if ((dx > 0) != (dy > 0)) {

	  ax = dx - 0.5;
	  ay = dy - 0.5;
	  bx = dx + 0.5;
	  by = dy + 0.5;

	} else {

	  ax = dx - 0.5;
	  ay = dy + 0.5;
	  bx = dx + 0.5;
	  by = dy - 0.5;
	  
	}
      }
    }

    // cout << ax << " " << ay << endl;
    // cout << bx << " " << by << endl;    

    sint4 xm = k, ym = k;
    sint4 d2 = ok_dist2(k);

    // test occlusion for any point in sight
    
    FORS (x, 2*k+1) {
      FORS (y, 2*k+1) {
	if ((x-xm)*(x-xm) + (y-ym)*(y-ym) > d2) continue;
	// (x,y) in sight
	real8 ndx = x-k, ndy = y-k;
	if (ndx == dx && ndy == dy) continue; // the obstacle itself

	// compute line intersection
	// U * (ndx,ndy) = a + (b-a)* V
	// 0 < U < 1 & 0 < V < 1
	
	real8 d3 = (ndx * (ay-by) - ndy * (ax-bx));

	if (d3 != 0) {

	  real8 d1 = ax*(ay-by)-ay*(ax-bx);
	  real8 d2 = ndx * ay - ndy * ax;

	  // U = d1/d3, V = d2/d3
	  
	  if (sign(d1) != sign(d3) || sign(d2) != sign(d3)) continue;

	  //cout << p1 << " " << p2 << " " << d3 << endl;

	  d3 = abs(d3);
	  
	  if (abs(d1) < d3 && abs(d2) < d3) {
	    Pair p;
	    p.x = x-k;
	    p.y = y-k;
	    occ.push_back(p);
	    //cout << ndx << "," << ndy << endl;
	  }
	}
      }
    }
  }

  void add_view(sint4 sight, sint4 x, sint4 y, sint4 owner_set)
  {
    // object sits on (x0,y0) and looks around
    // take occlusions into account

    vector<sint4> owners;

    FORU (i, sizeof(owner_set)*8) {
      if (owner_set & (1 << i)) owners.push_back(i);
    }
    
    sint4 d2 = ok_dist2(sight);

    // mark occluded squares

    sint1 occluded[tile_num];
    memset(occluded, 0, sizeof(occluded[0])*tile_num);
      
    for (sint4 dy=-sight; dy <= sight; dy++) {
      sint4 ny = y+dy;
      if (ny < 0 || ny >= tile_h) continue;    // y outside
      for (sint4 dx=-sight; dx <= sight; dx++) {
	sint4 nx = x+dx;
	if (nx < 0 || nx >= tile_w) continue;      // x outside
	if (dx*dx+dy*dy > d2) continue;            // out of sight
	
	if (tiles[xy2ind(nx, ny)].level == 0) continue;

	// blocking square at (nx, ny)
	// find occluded squares

	vector<Pair> occ_coord;
	generate_occlusions(sight, dx, dy, occ_coord);

	cout << dx << " " << dy << " " << occ_coord.size() << endl;
	
	FORALL (vector<Pair>, occ_coord, i) {
	  sint4 occx = x + i->x;
	  sint4 occy = y + i->y;
	  if (occx < 0 || occx >= tile_w) continue;
	  if (occy < 0 || occy >= tile_h) continue;	  
	  occluded[xy2ind(occx, occy)] = 1;
	}
      }
    }
      
    // compute visible squares

    for (sint4 dy=-sight; dy <= sight; dy++) {
      sint4 ny = y+dy;
      if (ny < 0 || ny >= tile_h) continue;    // y outside
      for (sint4 dx=-sight; dx <= sight; dx++) {
	sint4 nx = x+dx;
	if (nx < 0 || nx >= tile_w) continue;  // x outside
	if (dx*dx+dy*dy > d2) continue;        // out of sight

	sint4 index = xy2ind(nx, ny);
	
	if (!occluded[index]) {

	  FORALL (vector<sint4>, owners, i) {
	    ++tile_infos[index].vis_counters[*i];
	  }
	}
      }
    }
  }

  
#if 0  
  void generate_view(sint4 k, sint4 x, sint4 y, vector<sint4> &view)
  {
    
    generate_view_template(k, view);

  }
#endif
  
};


class World {

public:

  Map map;
  vector<Player> players;
  
  World() {}
  
  virtual ~World() {
  }

  void create_views() {

    map.reset_tile_info();
    
    FORALL (vector<Object*>, map.objects, i) {
      map.add_view((*i)->sight_range, (*i)->tile_x, (*i)->tile_y, (*i)->owner_set);
    }

  }

};


int main()
{
  World w;
  vector<sint4> view;
  vector<Pair> occ;

#if 1

  Object *obj = new Object;

  obj->tile_x = 10;
  obj->tile_y = 10;  
  obj->sight_range = 10;
  obj->owner_set = 1;
  
  w.map.objects.push_back(obj);

  w.create_views();
  
  w.map.print_view(0);

#else
  
  sint4 k = 10;
  sint4 odx = -5;
  sint4 ody = -5;  
  
  cout << "k=" << k << endl;

  w.map.generate_view_template(k, view);

  FORS (y, 2*k+1) {
    FORS (x, 2*k+1) {
      cout << char(view[y *(2*k+1) + x]);
    }
    cout << endl;
  }

  w.map.generate_occlusions(k, odx, ody, occ);

  view[(ody+k)*(2*k+1)+odx+k] = 'O';
  
  FORALL (vector<Pair>, occ, i) view[(k+i->y) *(2*k+1) + (k+i->x)] = '*';

  FORS (y, 2*k+1) {
    FORS (x, 2*k+1) {
      cout << char(view[y *(2*k+1) + x]);
    }
    cout << endl;
  }

#endif
  
  return 0;
}
