// $Id: AI_Client.C,v 1.2 2004/10/14 23:43:36 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#include "Global.H"

#include "AI_Client.H"
#include "AI_Common.H"
#include "AI_Plan.H"
#include "Client.H"
#include "Options.H"
#include "GUI.H"
#include "GameChanges.H"
#include "GameObj.H"
#include "Game.H"
#include "GfxClient.H"
#include "PlayerInfo.H"

#include "util.h"  // from pathfinding library

using namespace std;

/**
 TODO LIST (Michael Chung)
 
 Possible Improvements and New Features:
   - better planning for opponent
   - add some more plans
     - group plan (not used in simulation currently)
     - retreat
   - more terrain
   - bias towards keeping the same plan
   - check for conditions for replanning... possibilities:
     - plan partially completed (key parts completed)
     - plan cannot be completed (e.g. key unit dies)
     - time-out period expires
     - previously non-fighting unit enters a fight
     - flag is picked up (either enemy or friendly)
   - better debugging tools - show status, plans selected, etc.
   - performance issues (multi-threading, background planning...)

   o 'randsim' random plans + simulation evaluation AI
   o better random plans (equal length?)
   
 Recently Completed:
   x randomize planning decisions inside eval_mc
   x check why is_plan_complete() isn't working properly - fixed
   x planning for opponent
   x added plan ID field
   x show plan selected by planner
   x code for plans re-organized (based on plan ID's) to make it easier 
      to manage
   x added defend flag plan
   x replan when plan is completed
   
   x add random evaluation option
*/

// for debugging

//#define DISABLE_OPPONENT_PLANNING

//#define DISABLE_ATTACKS

//#define DISABLE_MOVEMENT

// debug flag...
//#define MAX_8_RANDSIM

// action ID (must match blueprints)
#define UNIT_ATTACK_ID (1)
#define UNIT_PLANT_FLAG_ID (3)

#define POST_PICK_UP_FLAG_ID (0)

// flag pick-up range - should match blueprint's defined value
#define FLAG_RANGE (40)

// flag range in tiles
#define FLAG_RANGE_TILES (FLAG_RANGE / AREA_SIZE)

// (chasing disabled for now)
#define CHASE_RANGE (128)

// debugging flags - the debug command-line option might need to be 
// true for some of these to work
#define DISPLAY_BOARD

//#define REPLAN_INTERVAL  (40)
#define REPLAN_INTERVAL  (120)

// penalty is applied if there is no fighting and we are still
#define STANDING_STILL_PENALTY  (0.001)

// how long (in frames) do we wait before we consider it an unexpected 
// collision (rather than network latency)
#define AI_COLLISION_WAIT_TIME  (4)
//#define AI_COLLISION_WAIT_TIME  (40)


//---------------------------------------------------------------------
// Eval option Identifiers
//---------------------------------------------------------------------

#define EVAL_MODE_UNDEFINED (0)
#define EVAL_MODE_AVG (1)
#define EVAL_MODE_MIN (2)
#define EVAL_MODE_MAX (3)

static const struct {
  char* eval_name;
  sint4 eval_id;
} eval_mode_map[] = {
  {"avg", EVAL_MODE_AVG},
  {"min", EVAL_MODE_MIN},
  {"max", EVAL_MODE_MAX}
};

string get_str_from_eval_mode(sint4 id)
{
  for (uint4 i = 0; i < sizeof(eval_mode_map) / sizeof(eval_mode_map[0]); i++) {
    if (id == eval_mode_map[i].eval_id) {
      return eval_mode_map[i].eval_name;
    }
  }
  
  return "";
}

sint4 get_eval_mode_from_str(const string &s)
{
  for (uint4 i = 0; i < sizeof(eval_mode_map) / sizeof(eval_mode_map[0]); i++) {
    if (s == eval_mode_map[i].eval_name) {
      return eval_mode_map[i].eval_id;
    }
  }
  
  return EVAL_MODE_UNDEFINED;
}


//----------------------------------------------------------------------
// AI_Client implementation
//----------------------------------------------------------------------

AI_Client::AI_Client(Options &opt) :

  GfxClient(opt),

  m_board(),

  m_b_disp_gui(true),
  m_b_debug(false),
  m_b_timing(false),

  m_objects(),

  m_plan(),

  m_b_game_over(false),

  m_ai_opt(0),
  m_eval_opt(0),
  m_eval_mode_opt(0),
  
  m_stat_run_time(),
  
  m_mc_runs(0),
  m_mc_steps(0),
  
  m_rm_runs(0),
  m_rm_dist(0),
  
  m_b_replan_now(true),
  m_b_replan_opt(false),
  
  m_n_enemies_sighted(0)
{
}

AI_Client::~AI_Client()
{
  // empty for now
}

void AI_Client::run_game()
{
  load_options();
  
  if (m_b_disp_gui) {
    GfxClient::run_game();
  } else {
    run_without_gui();
  }
}

// virtual function - executed on game view change/update
void AI_Client::on_game_update()
{
  // update internal board
  m_board.update_board(game);
  
#ifdef DISPLAY_BOARD
  if (m_b_debug) {
    // display board after each update for debugging
    m_board.print();
  }
#endif
  
  if (m_board.is_game_won()) {
    cout << "Game Over - WON" << endl;
    m_b_game_over = true;
    
    cout << "Winning time = " << get_current_time() << endl;
    
  } else if (m_board.is_game_lost()) {
    cout << "Game Over - LOST" << endl;
    m_b_game_over = true;
  }

  // check for new enemies sighted / enemies disappeared
  sint4 enemies_sighted = m_board.enemy_count();
  if (enemies_sighted != m_n_enemies_sighted) {
    
    if (m_b_replan_opt && (enemies_sighted > m_n_enemies_sighted)) {
      cout << "Additional enemy sighted - replanning..." << endl;
      
      // signal replan
      m_b_replan_now = true;
    }
    
    // update member variable with new value
    m_n_enemies_sighted = enemies_sighted;
  }
  
  // generate actions even if game is over - do victory dance, etc.
  gen_ai_client_actions();  
}

void AI_Client::run_without_gui()
{
  FOREVER {

    handle_events();
    if (quit) break;

    Event ev = get_event();

    if (ev != EVT_NONE) {

      if (ev == EVT_READ_ERROR) return;
      if (ev == EVT_STOP) return;

      if (ev == EVT_NEW_VIEW) {

        //process_changes();

        //Vector<string> temp;
        //FORS(i, game->get_player_num()) temp.push_back("");
        //game->exec_actions(temp);

        // call virtual function
        on_game_update();
        
        send_actions();

        sint4 vf = game->get_view_frame();
        sint4 ai = game->get_action_frame();
        sint4 sk = game->get_skipped_actions();
        sint4 md = game->get_merged_actions();	
	
        if (vf != ai || sk || md > 1) {
          
          cout << "frame " << vf;
          if (ai < 0) cout << " [no action]";
          else if (vf != ai) cout << " [behind by " << (vf - ai) << " frame(s)]";
          if (sk) cout << " [skipped " << sk << "]";
          if (md > 1) cout << " [merged " << md << "]";	  
          cout << endl;
        }	
      }
      
    } else {

      SDL_Delay(1);

    }
  }
}

void AI_Client::load_options()
{
  opt.get("-disp", m_b_disp_gui);

  // read command-line options into member variables
  opt.get("-debug", m_b_debug);
  opt.get("-timing", m_b_timing);

  string s_ai_opt;
  opt.get("-ai", s_ai_opt);
  m_ai_opt = AI_Plan::get_id_from_str(s_ai_opt);
  if (AI_PLAN_UNKNOWN != m_ai_opt) {
    cout << "AI setting: " << s_ai_opt.c_str() << endl;  
  }
  
  if ((AI_PLAN_MC == m_ai_opt) || (AI_PLAN_RANDSIM == m_ai_opt)) {
    opt.get("-mcruns", m_mc_runs);
    cout << "-mcruns = " << m_mc_runs << endl;
    
    opt.get("-mcsteps", m_mc_steps);
    cout << "-mcsteps = " << m_mc_steps << endl;
  }
  
  if ((AI_PLAN_RANDOM == m_ai_opt) || (AI_PLAN_RANDSIM == m_ai_opt)) {
    opt.get("-rmruns", m_rm_runs);
    cout << "-rmruns = " << m_rm_runs << endl;
    
    opt.get("-rmdist", m_rm_dist);
    cout << "-rmdist = " << m_rm_dist << endl;
  }

  string s_eval_opt;
  opt.get("-eval", s_eval_opt);
  m_eval_opt = AI_Board::get_eval_opt_from_str(s_eval_opt);
  if (EVAL_UNDEFINED != m_eval_opt) {
    cout << "Eval setting: " << s_eval_opt.c_str() << endl;
  }
  
  opt.get("-replan", m_b_replan_opt);
  
  string s_eval_mode;
  opt.get("-evalmode", s_eval_mode);
  m_eval_mode_opt = get_eval_mode_from_str(s_eval_mode);
  if (EVAL_MODE_UNDEFINED != m_eval_mode_opt) {
    cout << "Eval mode setting: " << s_eval_mode.c_str() << endl;
  }
}

void AI_Client::gen_ai_client_actions()
{
  // generate client actions...

  Game *game = game;
  
  sint4 p = game->get_client_player();

  sint4 tile_points = game->get_tile_points();
  
  sint4 player_n = game->get_player_num();
  
  // allow separate mode of behaviour when we are in battle
  m_b_fighting = false;
  
  const Game::ObjCont &obj_cont = game->get_objs(p);
  FORU (i, obj_cont.size()) {
    GameObj &gob = *obj_cont[i];
    if (TYPE_POST == get_obj_type(gob)) {
      // pickup enemy flag
      
      // todo: improve this code (make it more readable, shorter)
      FORS (q, player_n) {
        if (p == q)
          continue;
        
        const Game::ObjCont &enemy = game->get_objs(q);
        FORU (j, enemy.size()) {
          GameObj &enemyobj = *enemy[j];
          if (TYPE_FLAG != get_obj_type(enemyobj))
            continue;
          
          // flag found
          if (calc_distance(gob, enemyobj) <= FLAG_RANGE) {
            pick_up_flag(gob, enemyobj);
            break;
          }
        }
      }
      
      continue;
    }
    
    if (!is_unit(gob)) {
      // not a unit (so it's a building or other object)
      continue;
    }
    
    sint4 obj_id = get_obj_id(gob);
    
    AI_Object &obj = m_objects[obj_id];
    obj.m_obj_id = obj_id;

    // if already attacking, continue attacking
    if (gob.get_int("is_attacking")) {
      m_b_fighting = true;
      continue;
    }

    // look for something to kill
    double min_dist = GameConst::VIEW_MAX * tile_points;
    GameObj *target = NULL;  // pointer to target object
  
    FORS (q, player_n) {
      if (p == q)
        continue;
      
      const Game::ObjCont &enemy = game->get_objs(q);
      FORU (j, enemy.size()) {
        GameObj &enemyobj = *enemy[j];
        
        switch (get_obj_type(enemyobj)) {
        case TYPE_UNIT:
        case TYPE_FLAG:
          {
            double dist = calc_distance(gob, enemyobj);
            if (dist < min_dist) {
              min_dist = dist;
              target = &enemyobj;
            }
          }
          break;
        }
      }
    }

    // note: AI does not run and shoot at the same time.  In some games
    // this is not allowed, anyway.

    if (target != NULL) {

      // target acquired
      
      const sint4 error_margin = 8;  // max error in the exact range
      if (min_dist <= ATTACK_RANGE + error_margin) {

        m_b_fighting = true;
        
        // target is close enough - shoot at it
        attack_target(gob, *target);
        continue;
      }
    }

    if (has_flag(gob)) {
      // check if we are in range of our base - if so, plant the flag
      FORU (k, obj_cont.size()) {
        GameObj &gob_base = *obj_cont[k];
        if (TYPE_POST != get_obj_type(gob_base))
          continue;  // wrong object
        
        // found our base - check range
        if (calc_distance(gob_base, gob) <= FLAG_RANGE) {
          // drop the flag and hopefully win the game
          plant_flag(gob);
        }
      }
    }
  
  } // end of loop

  // check if we should replan...
  
  sint4 current_time = get_current_time();

  if (m_plan.is_completed(m_board)) {
    
    cout << "plan is complete - replan" << endl;
    m_b_replan_now = true;
    
  } else if (current_time > m_plan.get_time() + REPLAN_INTERVAL) {
    
    cout << "replan interval expired - replan" << endl;
    m_b_replan_now = true;
  }
  
  if (m_b_replan_now) {
    
    replan();
      
    // mark the time the plan was generated
    m_plan.set_time(current_time);
    
    m_b_replan_now = false;
  }

  execute_current_plan();
}

void AI_Client::replan()
{
  // start timer
  PathFind::Clock clock;

  m_plan.clear();

  // different code for how to generate plans
  switch (m_ai_opt) {
    
  case AI_PLAN_RANDOM:
    // evaluate some randomly generated plans and pick the best
    find_best_random_plan();
    break;
  
  case AI_PLAN_RANDSIM:
    // evaluate some randomly generated plans using simulation
    // and pick the best
    find_best_randsim_plan();
    break;
  
  case AI_PLAN_STAND:
    // plan: just stand there - achieved by doing nothing
    break;
  
  case AI_PLAN_FLAG:
    // plan: capture flag, bring it back
    m_plan.generate_flag_plan(m_board, FRIENDLY_SIDE);
    break;
  
  case AI_PLAN_EXPLORE:
    // plan: explore the unknown
    m_plan.generate_explore_plan(m_board);
    break;
  
  case AI_PLAN_GROUP:
    // plan: group up (around a certain point...)
    find_best_group_plan();
    break;
  
  case AI_PLAN_ALL:
    // try all plans: pick the best based on static evaluation...
    find_best_plan();
    break;
  
  case AI_PLAN_MC:
    // try all plans: pick the best based on monte-carlo evaluation...
    find_best_mc_plan();
    break;
  
  case AI_PLAN_DEFEND:
    // defend our flag
    m_plan.generate_defend_plan(m_board, FRIENDLY_SIDE);
    break;
  
  default:
    ERR("invalid plan setting");
    break;
  }
  
  if (m_b_timing)
  {
    // stop timer
    double n_time_elapsed = clock.get();
  
    cout << "Planning time: " << n_time_elapsed << endl;
  
    m_stat_run_time.add(n_time_elapsed);
  }
}
  
void AI_Client::attack_target(GameObj &gob, GameObj &target)
{
#ifndef DISABLE_ATTACKS
  
  if (!is_unit(gob)) ERR("non-unit trying to attack");

  ObjId id = get_obj_id(target);
  //cout << "target id = " << id << endl;
  
  Vector<sint4> params;
  params.push_back(id);
  
  if (!gob.set_action(UNIT_ATTACK_ID, params)) {
    ERR("failed to set attack action");
  }
  
#endif //DISABLE_ATTACKS
}

void AI_Client::pick_up_flag(GameObj &gob, GameObj &target)
{
  if (TYPE_FLAG == get_obj_type(gob)) ERR("non-flag expected");
  if (TYPE_FLAG != get_obj_type(target)) ERR("flag expected");
  
  ObjId id = get_obj_id(target);
  Vector<sint4> params;
  params.push_back(id);
  
  if (!gob.set_action(POST_PICK_UP_FLAG_ID, params)) {
    ERR("failed to set pick up flag action");
  }
}

void AI_Client::plant_flag(GameObj &gob)
{
  if (TYPE_FLAG == get_obj_type(gob)) ERR("non-flag expected");
  if (!has_flag(gob)) ERR("flag not found");
  
  Vector<sint4> params; // no params
  
  if (!gob.set_action(UNIT_PLANT_FLAG_ID, params)) {
    ERR("failed to set plant flag action");
  }  
}

void AI_Client::chase_unit(GameObj &gob, GameObj &target)
{
  Point ptA;
  ptA.x = gob.get_int("x");
  ptA.y = gob.get_int("y");

  Point ptC;
  ptC.x = target.get_int("x");
  ptC.y = target.get_int("y");
  
  Point ptB;
  ptB.x = (ptA.x + ptC.x) / (Scalar)2;
  ptB.y = (ptA.y + ptC.y) / (Scalar)2;
  
  internal_move_to_point(gob, ptB);
}

void AI_Client::move_randomly(GameObj &gob)
{
  Point pt;
  pt.x = gob.get_int("x");
  pt.y = gob.get_int("y");

  const sint4 d = AREA_SIZE;
  
  switch (Rand.rand_uint4() % 4) {
  case 0: pt.y -= d; break;
  case 1: pt.y += d; break;
  case 2: pt.x -= d; break;
  case 3: pt.x += d; break;
  }
  
  internal_move_to_point(gob, pt);
}

void AI_Client::find_best_random_plan()
{
  //cout << "current eval: " << eval_board(m_board) << endl;
  
  // should we check if our previous plan was completed successfully?
  // if not we might want to try to complete it...? we'll see...
  
  uint4 n_plans = m_rm_runs;  // number of plans to test
  
  double best_val = -1000000;  // negative infinity
  
  FORU (i, n_plans) {
    
    AI_Plan plan;
    plan.generate_random_plan(m_board, m_rm_dist, FRIENDLY_SIDE);

    double val = evaluate_plan(plan, m_board);
    
    if (val > best_val) {
      // set this as our active plan and execute it
      m_plan = plan;
      best_val = val;
    }
    
  }
  
  //cout << "best plan's evaluation: " << best_val << endl;
}

void AI_Client::find_best_randsim_plan()
{
  //cout << "current eval: " << eval_board(m_board) << endl;
  
  // should we check if our previous plan was completed successfully?
  // if not we might want to try to complete it...? we'll see...
  
  uint4 n_plans = m_rm_runs;  // number of plans to test
  
  double best_val = -1000000;  // negative infinity
  
  FORU (i, n_plans) {
    
    AI_Plan plan;
    plan.generate_random_plan(m_board, m_rm_dist, FRIENDLY_SIDE);

    // use monte-carlo simulation-based evaluator...
    double val = evaluate_plan_mc(plan, m_board);
    
    if (val > best_val) {
      // set this as our active plan and execute it
      m_plan = plan;
      best_val = val;
    }
    
  }
  
  //cout << "best plan's evaluation: " << best_val << endl;
}

void AI_Client::find_best_group_plan()
{
  // for now just find one plan (best by default)
  m_plan.generate_group_plan(m_board);
}

void AI_Client::find_best_plan()
{
  get_best_plan(m_plan, m_board);
}

void AI_Client::get_best_plan(AI_Plan &plan_out, const AI_Board &bd) const
{
  double best_val = -1000000;  // negative infinity
  
  // first, consider the "stand" plan
  plan_out.clear();
  best_val = evaluate_plan(plan_out, bd);
  
  //cout << "stand eval: " << best_val << endl;
  
  AI_Plan plan;
  double val = 0;
  
  // consider the "flag" plan
  plan.generate_flag_plan(bd, FRIENDLY_SIDE);
  val = evaluate_plan(plan, bd);
  
  //cout << "flag eval: " << val << endl;
  
  if (val > best_val) {
    // set this as our active plan, to be executed
    plan_out = plan;
    best_val = val;
  }
  
  // consider the "explore" plan
  plan.generate_explore_plan(bd);
  val = evaluate_plan(plan, bd);
  
  //cout << "explore eval: " << val << endl;
  
  if (val > best_val) {
    // set this as our active plan, to be executed
    plan_out = plan;
    best_val = val;
  }
}

void AI_Client::get_randomly_selected_plan(
  AI_Plan &plan_out, 
  const AI_Board &bd) const
{
  // list of possible plans
  const sint4 plan_id_arr[] = {
    AI_PLAN_STAND,
    AI_PLAN_FLAG,
    AI_PLAN_EXPLORE,
    AI_PLAN_DEFEND
  };
  
  const uint4 num_plans = sizeof(plan_id_arr) / sizeof(plan_id_arr[0]);
  
  uint4 n = Rand.rand_uint4() % num_plans;

  sint4 plan_id = plan_id_arr[n];
  plan_out.generate_plan(plan_id, bd, FRIENDLY_SIDE);

  // note: no need to evaluate the plan - simply return it
}

void AI_Client::get_opponent_plan(AI_Plan &plan_out, const AI_Board &bd) const
{
  // list of possible plans
  const sint4 plan_id_arr[] = {
    AI_PLAN_STAND,
    AI_PLAN_FLAG,
    AI_PLAN_FLAG,  // double chance of selecting flag plan
    AI_PLAN_DEFEND
  };
  
  const uint4 num_plans = sizeof(plan_id_arr) / sizeof(plan_id_arr[0]);
  
  uint4 n = Rand.rand_uint4() % num_plans;
  sint4 plan_id = plan_id_arr[n];
  
  plan_out.generate_plan(plan_id, bd, ENEMY_SIDE);
  
  // note: no need to evaluate the plan - simply return it
}

void AI_Client::get_random_opponent_plan(
  AI_Plan &plan_out, 
  const AI_Board &bd) const
{
  plan_out.generate_random_plan(bd, m_rm_dist, ENEMY_SIDE);
  
  // note: no need to evaluate the plan - simply return it
}

void AI_Client::find_best_mc_plan()
{
  double best_val = -1000000;  // negative infinity
  
  // first, consider the "stand" plan
  m_plan.generate_stand_plan();
  best_val = evaluate_plan_mc(m_plan, m_board);
  
  if (!m_b_fighting)
  {
    // apply the standing-still penalty
    best_val -= STANDING_STILL_PENALTY;
  }
  
  AI_Plan plan;
  double val = 0;
  
  // list of plans to consider
  const sint4 plan_id_arr[] = {
    AI_PLAN_FLAG,
    AI_PLAN_EXPLORE,
    AI_PLAN_DEFEND
  };
  
  const uint4 num_plans = sizeof(plan_id_arr) / sizeof(plan_id_arr[0]);
  FORU(i, num_plans) {
    
    sint4 plan_id = plan_id_arr[i];
    
    plan.generate_plan(plan_id, m_board, FRIENDLY_SIDE);
    val = evaluate_plan_mc(plan, m_board);
  
    if (val > best_val) {
      // set this as our active plan, to be executed
      m_plan = plan;
      best_val = val;
    }
  }

  string s_plan = AI_Plan::get_str_from_id(m_plan.get_plan_id());
  cout << "plan selected: " << s_plan.c_str() << endl;
}

void AI_Client::execute_current_plan()
{
  //cout << "execute_current_plan..." << endl;
  
  sint4 p = game->get_client_player(); // our player index  
  const Game::ObjCont &obj = game->get_objs(p);

  if (m_plan.is_completed(m_board)) {
    // plan is finished
    //cout << "plan finished" << endl;
    return;
  }

  FORU(gob_ind, obj.size()) {
   
    GameObj &gob = *obj[gob_ind];
    if (!is_unit(gob))
      continue;
    
    if (is_moving(gob))
      continue;  // let the unit continue moving
    
    sint4 obj_id = get_obj_id(gob);
    
    if (!m_plan.contains(obj_id))
      continue;  // this object is not considered in this plan...
    
    AI_Path *path = m_plan.get_path(obj_id);
    if (NULL == path)
      continue;  // no path...
    
    Point ptCurr = get_board_position(gob);
    //cout << "curr: " << FORMAT_PT(ptCurr) << endl;
    
    Point pt;
    if (!path->get_next_waypoint(pt))
      continue;
    
    if (is_equal(pt, ptCurr)) {
      
      //cout << "waypoint reached" << endl;
      
      path->pop_next_waypoint();
      
      if (!path->get_next_waypoint(pt))
        continue;
    }
    
    // note: invalid waypoints cause server to crash!
    if (m_board.invalid_point(pt)) ERR("invalid next waypoint");
    
    //cout << "next: " << FORMAT_PT(pt) << endl;
    
    if (is_equal(pt, ptCurr)) {
      ERR("unexpected case");
      continue;
    }
    
    // check for interruption (obstacle) in path...
    if (!m_board.area_empty(pt)) {
      
      /*
      //cout << "replanning path..." << endl;
      
      Point pt_goal = m_plan.get_dest(obj_id);

      // replan the path...
      m_plan.plan_path(obj_id, m_board, ptCurr, pt_goal);
      
      // we could execute the new plan this turn, but for now
      // lets just stay still for a frame...
      */
      
      // instead of just replanning this one path, replan completely
      cout << "new obstacle found - replan" << endl;
      m_b_replan_now = true;
      
      continue;
    }
        
    Point ptEnv = conv_to_env_point(pt);
    move_to_point(gob, ptEnv);
  }
}

void AI_Client::move_to_point(GameObj &gob, const Point &pt)
{
  if (!is_unit(gob)) ERR("not a unit - can't move");
  
  Point ptCurrPos = get_position(gob);

  // snap the destination point to the grid
  Point ptDest = snap_point_to_grid(pt);

  sint4 obj_id = get_obj_id(gob);
  AI_Move_Command &prev_move = m_moves[obj_id];
  
  Point ptOldOrig = prev_move.m_move_origin;
  Point ptOldDest = prev_move.m_move_destination;
  sint4 ai_move_time = prev_move.m_move_time;
  
  sint4 current_time = get_current_time();
  
  if (is_moving(gob)) {
	// repeated move commands may prevent unit from attacking
	// however we will later want to allow overriding moves
	return;
  }

  if (ai_move_time == current_time) {
	// one move command per frame is enough
	ERR("!");
	return;
  }
  
  if (ai_move_time >= 0) {
	// a previous move command may have resulted in a collision.
	// confirm this by checking the position vs destination
	
	if (!is_equal(ptOldDest, ptCurrPos)) {

      //cout << "unit did not successfully complete its last move" << endl;

	  if ((current_time > ai_move_time + AI_COLLISION_WAIT_TIME) &&
		(is_equal(ptCurrPos, ptOldOrig))) {

		//cout << "unit is stuck - moving randomly" << endl;

        // try to get out of this sticky situation by moving randomly
        move_randomly(gob);

		// need to modify this to avoid the duplicate order check
		prev_move.m_move_time = -1;

		return;
      }
	  
      if (is_equal(ptOldDest, ptDest)) {
        // duplicate order...
        //cout << "destination is same as before - ignored" << endl;
        return;
      }
	}
  }
  
  if (is_equal(ptDest, ptCurrPos)) {
     //cout << "Warning: trying to send a null move" << endl;
     return;
  }
  
  internal_move_to_point(gob, pt);
  
  // remember the origin and destination
  prev_move.m_move_origin = ptCurrPos;
  prev_move.m_move_destination = ptDest;
  
  // remember the time (frame) the move command was sent  
  prev_move.m_move_time = get_current_time();
}

void AI_Client::internal_move_to_point(GameObj &gob, const Point &pt)
{
#ifndef DISABLE_MOVEMENT
  
  Vector<sint4> params;
  params.push_back(pt.x);
  params.push_back(pt.y);
  
  gob.set_action("move", params);
  
#endif //DISABLE_MOVEMENT
}

sint4 AI_Client::get_obj_id(GameObj &gob) const
{
  return ::get_obj_id(gob, game);
}

double AI_Client::evaluate_plan(
  const AI_Plan &plan, 
  const AI_Board &board) const
{
  //cout << "evaluate plan" << endl;
  
  // currently only handles complete plans (all units have destinations)
  // as well as completely empty plan
  if (plan.is_empty()) {
    //cout << "plan is empty" << endl;
    return eval_board(board);
  }

  AI_Board bd_think = board;  // the current board  
  
  const object_map &objs = bd_think.get_friendly_objs();
  object_map::const_iterator iter = objs.begin();
  for (; iter != objs.end(); iter++) {
  
    const AI_Object &obj = iter->second;
    if (!obj.is_unit())
      continue;
    
    sint4 obj_id = obj.id();
    Point ptDest = plan.get_dest(obj_id);
    
    // todo: perform legality checks and possibly disqualify the plan
    // this would require our plan generator be more careful.
    // for now, simply try to apply or skip the invalid move.
    
    if (bd_think.invalid_point(ptDest))
      continue;  // no destination for this unit
    
    //cout << "performing movement" << endl;
    
	bd_think.move(obj_id, ptDest);
  }
  
  // perform attacks (kind of randomly for now)
  bd_think.simulate_random_attacks();
  
  //cout << "random attacks simulated" << endl;
  
  return eval_board(bd_think);
}

double AI_Client::evaluate_plan_mc(
  const AI_Plan &plan, 
  const AI_Board &bd) const
{
  double acc = 0.0;
  double min = 0.0;
  double max = 0.0;
  
  for (sint4 i = 0; i < m_mc_runs; i++) {
    
    double val = evaluate_plan_mc_run(plan, bd);
    acc += val;
    
    if (0 == i) {
      min = val;
      max = val;
    } else {
      if (val < min) min = val;
      if (val > max) max = val;
    }
  }
  
  switch (m_eval_mode_opt) {
    case EVAL_MODE_AVG:
      return acc / (double)m_mc_runs;
    
    case EVAL_MODE_MIN:
      return min;
    
    case EVAL_MODE_MAX:
      return max;
    
    default:
      break;
  }
 
  ERR("unexpected eval mode");
  return 0.0;
}

// perform a single run
double AI_Client::evaluate_plan_mc_run(
  const AI_Plan &plan,
  /*const AI_Plan &plan_opponent_in,*/
  const AI_Board &bd) const
{
  //sint4 debug_count = 0;
  
  // simulate part of the plan... then evaluate the board
  
  AI_Board bd_think = bd;  // the current board  
  AI_Plan plan_think = plan;  // remaining part of plan
  
  AI_Plan plan_opponent /*= plan_opponent_in*/;  // the opponent's plan
  
  // maximum time steps to simulate ahead
  sint4 time_left = m_mc_steps;
  while (true) {
    
    // simulate a time-step of plan execution
    simulate_plan_step(plan_think, plan_opponent, bd_think);

    if (--time_left <= 0)
      break;
    
    
    if (plan_think.is_completed(bd_think)) {
      
      if (AI_PLAN_RANDSIM == m_ai_opt)
        break;  // for randsim, evaluation is complete
      
      // for debugging
      //debug_count++;
      
      // replan... make somewhat random decisions...
      get_randomly_selected_plan(plan_think, bd_think);
    }
    
    if (plan_opponent.is_completed(bd_think)) {
      get_random_opponent_plan(plan_opponent, bd_think);
    }
    
    // interactions ignored for now...
    // if (interaction detected) then replan...
  }

  //cout << "number of replans: " << debug_count << endl;
  
  // return the evaluation of the resulting board
  return eval_board(bd_think);
}

double AI_Client::eval_board(const AI_Board &board) const
{
  return board.eval(m_eval_opt);
}

sint4 AI_Client::get_current_time()
{
  return game->get_view_frame();
}

void AI_Client::simulate_plan_step(
  AI_Plan &plan_friendly, 
  AI_Plan &plan_opponent,
  AI_Board &bd) const
{
  FORS(side, 2) {
    
#ifdef DISABLE_OPPONENT_PLANNING
    if (ENEMY_SIDE == side)
      continue;  // skip opponent planning for now
#endif //DISABLE_OPPONENT_PLANNING
    
    // select plan to simulate - do each side one at a time
    AI_Plan &plan = (FRIENDLY_SIDE == side) ? plan_friendly : plan_opponent;
    
    const object_map &objs = bd.get_objs(side);
    object_map::const_iterator iter = objs.begin();
    for (; iter != objs.end(); iter++) {
    
      const AI_Object &obj = iter->second;
      if (!obj.is_unit())
        continue;
      
      sint4 obj_id = obj.id();
      if (!plan.contains(obj_id))
        continue;  // this object is not considered in this plan...
      
      AI_Path *path = plan.get_path(obj_id);
      if (NULL == path)
        continue;  // no path...
      
      Point pt;
      if (!path->get_next_waypoint(pt))
        continue;
      
      if (is_equal(pt, obj.location())) {
        path->pop_next_waypoint();
        if (!path->get_next_waypoint(pt))
          continue;
      }
      
      // note: invalid waypoints cause server to crash!
      if (bd.invalid_point(pt)) ERR("invalid next waypoint");
      
      if (is_equal(pt, obj.location())) {
        ERR("unexpected case");
        continue;
      }
  
      // move obj to next point
      bd.move(obj_id, pt);
    }
  }  
  
  // perform attacks (kind of randomly for now)
  bd.simulate_random_attacks();
}


// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
